#ifndef STRUCTS_H
#define STRUCTS_H

#include <features.h>
__BEGIN_DECLS

#include "defs.h"

/* The listening port of a connection is not known
 * until a socket is accepted and then a msg is
 * received. Until this msg is received the ip:port
 * of a connection is initialized to this value.
 * NOTE: Not a good idea keeping it 0 as its a
 * possible ip or port number but we save some space. */
#define UNKNOWN_IP 0
#define UNKNOWN_PORT 0
#define IS_UNKNOWN_NADDR(x) ((x).ip==UNKNOWN_IP && (x).port==UNKNOWN_PORT)
#define SET_UNKNOWN_NADDR(x) { (x).ip=UNKNOWN_IP; (x).port=UNKNOWN_PORT; }

/* The struct used to used all network addresses visible to the user.
 * There is a special value for this message:
 * i.e. ip=UNKNOWN_IP port=UNKNOWN_PORT. This implies that the address
 *      is not known and is handled as a special case in the connection
 *      manager. */
typedef struct {
	u_int32_t ip;		/* IP address in network byte order */
	u_int16_t port;		/* Port in network byte order */
} naddr_t;


/* Daemon/VAS IDs: contains 'addr' and 'port' in network byte order */
typedef naddr_t daemon_id;
typedef naddr_t vas_id;


/* A null task_id for error conditions */
#define PID_NULL 0
#define TASK_ID_NULL(tid) ((tid).pid=PID_NULL)
#define TASK_ID_ISNULL(tid) ((tid).pid==PID_NULL)

#define TASK_MAX_FILENAME 128
#define TASK_MAX_STARTSYM 64

#define TASKREQ_CREATE_MASTER 1
#define TASKREQ_CREATE_LOCALTASK 2
#define TASKREQ_CREATE_TASK 3


/* Used to identify a single task anywhere on the network */
typedef struct {
	daemon_id cdaemon;	/* Controlling daemon */
	u_int32_t pid;		/* PID of task in network byte order */
} task_id;

/* This structure is sent in the create_task and create_master
 * requests by the loader. */
typedef struct {
	task_id master;
	task_id parent;
	u_int16_t priority;		/* in network byte order*/
	char file_name[TASK_MAX_FILENAME];
	char start_sym[TASK_MAX_STARTSYM];
} taskreq_t;

/* This struct is passed with the create master request only */
typedef struct {
	u_int32_t argc;			/* Number of args following
					 * this structure. Each arg
					 * should be NULL terminated */
} taskarg_t;

#define TASKRPLY_STATUS_SUCCESS 0
#define TASKRPLY_STATUS_ERR 1

/* This structure is sent as a reply to the task creation request */
typedef struct {
	task_id id;
	task_id master;
	task_id parent;
	int status;			/* in network byte order*/
} taskrply_t;

/* IO redirected data is sent following this struct */
typedef struct {
	task_id id;
	task_id master;
} ioredir_t;


#define VASREG_STATUS_SUCCESS	1
#define VASREG_STATUS_FAILURE	0

/* Stores the status of a VAS registration request */
typedef struct {
	u_int32_t status;
} vasreg_rply_t;

/* The load of a machine. Currently just the number of distributed
 * tasks running on a machine. */
typedef struct {
	u_int32_t numprocs;
} vasload_t;


typedef struct {
	daemon_id dem_id;
	u_int32_t ftpind;
} ftpid_t;

typedef struct {
	ftpid_t ftp_id;
	u_int32_t file_size;
	char md5[MD5_DIGEST_SIZE];
	/* char *file_name; following this struct */
} ftpinit_t;


/* used to identify a non-transaction msg */
#define TRANSID_NONE 0

/* transaction status */
#define TRANS_STATUS_PENDING 1
#define TRANS_STATUS_TIMEOUT 2
#define TRANS_STATUS_SUCCESS 3

/* Stores a transaction item. The _transact_t has only been
 * defined for the hnd_trans parameter. Use transact_t in all
 * other cases.
 * NOTE: The trans_id field in the struct has been defined in terms of size
 * coz whatever data/structs are sent within msgs on the network must be
 * defined in size to keep the code portable. */
typedef struct _transact_t {
	u_int16_t trans_id;	/* id of this transaction */
	int status;		/* status from above defines */
	time_t ttl;		/* time() + delay_in_secs for expiry */

	/* any data in the form of a void* maybe provided when a
	 * new transaction is registered. This would be sent as
	 * the 'data' in the transction failer handler below. */
	void *faildata;		/* dynamically allocated arbitrary
				 * data passed to hnd_fail. The
				 * transaction module is responsible
				 * for freeing this data. */

	void (*hnd_fail)(struct _transact_t *trans, void *data);
} transact_t;


#define RPC_STATUS_FAILURE 0
#define RPC_STATUS_SUCCESS 1
#define RPC_STATUS_PENDING 2
#define RPC_STATUS_NOBLOCK 3

/* The data required to perform an RPC/transaction from one
 * user task to another. The call is initiated by a task by
 * sending this struct in a msg to the controlling daemon.
 * The sender then continues to wait for the call's reply
 * while the daemon fwds the call to the destination machine
 * whose daemon in turn fwds it to the task if/when the
 * task issues an accept msg. The data returned by the RPC
 * also uses this same struct as a header. */
typedef struct {
	u_int32_t status;
	task_id src;
	task_id dest;
	u_int32_t func_id;
} rpcinfo_t;

/* This struct is passed with the RPC accept message */
typedef struct {
	u_int32_t func_id;
	u_int32_t is_blocking;
} rpcaccept_t;


/* Used to pass variable sized arbitrary data via RPCs.
 * The data transfer is non-portable so this should be
 * restricted to register size and byte order independent
 * data. */
typedef struct {
	u_int32_t size;
	char *data;
} rawdata_t;


/* Passed with MSG_TERMINATE and MSG_ABORT to terminate
 * subtasks or compelete tasks. The tid field may be left
 * blank if the message is sent locally. Otherwise it
 * must contain the ID of either the local or master task. */
typedef struct {
	task_id tid;
	u_int32_t status;
} exitinfo_t;


/* Passed with MSG_ISTERMINATE and MSG_DOTERMINATE. This
 * data type contains a variable list of daemons. A minimum
 * of one daemon is always present in the list and 'num'
 * specifies how many daemon_ids are present. If more than
 * one they should be present right after this data struct. */
typedef struct {
	task_id master;		/* ID of a master task */
	u_int16_t num;		/* Number of following daemons */
	daemon_id daemons[1];	/* List of atleast one daemons */
} taskpoll_t;


#ifdef NCDAEMON
typedef struct {
	/* items read from the config file */
	int conn_ttl;
	int log_types;
	int listen_port;
	clist *vas_list;

	/* the following items arnt read from the config file */
	naddr_t vas_cur;	/* one of the vas_list servers */
	vasload_t load;		/* current load of this machine */
	daemon_id self_id;	/* our own network address */
	char *homedir;		/* homedir of the effect uid */
	char *binary_dir;	/* path to search for task bins */
	char *log_file;		/* optional file to log to */
} dconf_t;
extern dconf_t dconf;
#endif

#ifdef NCVAS
typedef struct {
	/* items read from the config file */
	int conn_ttl;
	int ping_interval;
	int ping_timeout;
	int log_types;
	int listen_port;
	char *log_file;

	/* following items arnt read from the config file */
	char *homedir;
	daemon_id self_id;
} vconf_t;
extern vconf_t vconf;
#endif

__END_DECLS

#endif /* STRUCTS_H */
