/*
 * Screen Handling
 *
 * Written by Fakhir Ansari <ansari89@yahoo.com>
 *
 * Module - The screen class.
 *          An interface between the program and the terminal.
 *          Modify this class to port your application to any
 *          terminal. All terminal access must be through this
 *          class.
 *
 */

#include <stdio.h>
#include <stdlib.h>


#include "data.h"
#include "screen.h"

#ifdef X_BORLAND
#include <conio.h>
#else
#include <curses.h>
#include <unistd.h>
#endif

/**********************************************************************/

CSubScreen::CSubScreen(int xx1,int yy1,int xx2,int yy2)
{
    x1 = xx1;
    y1 = yy1;
    x2 = xx2;
    y2 = yy2;

#if defined(X_UNIX) || defined(X_DJGPP)
    Data = newwin(y2-y1+1, x2-x1+1, y1, x1);
#else
    Data = new char [(y2-y1+1) * (x2-x1+1) * 2];
#endif
}

CSubScreen::~CSubScreen()
{
#if defined(X_UNIX) || defined(X_DJGPP)
    delwin(Data);
#else
    delete [] Data;
#endif
}


/*******************************************************************/

int MAX_X;
int MAX_Y;

int CScreen::Init = 0;

CScreen::CScreen()
{
    Init = 1;

#if defined(X_UNIX)
    printf("\e(U");
#endif

#if defined(X_UNIX) || defined(X_DJGPP)

    if (!isatty(fileno(stdout))) {
	printf("Output device is not a terminal\n");
	printf("Aborting...\n");
	exit(EXIT_FAILURE);
    }

    initscr();
    cbreak();
    noecho();
    keypad(stdscr,TRUE);
    erase();

    Print = printw;

    if (has_colors()) {
        start_color();

        if (COLOR_PAIRS < 64) {
            endwin();
            printf("ERROR: Your terminal dosnt support enough colors\n");
            printf("to run the interface. Aborting...\n");
            exit(EXIT_FAILURE);
        }
        else
            InitColorPairs();
    }

    max_x = MAX_X = COLS;
    max_y = MAX_Y = LINES;
#else
    max_x = MAX_X = 80;
    max_y = MAX_Y = 25;

    clrscr();
    Print = cprintf;
#endif
}

CScreen::~CScreen()
{
#if defined(X_UNIX) || defined(X_DJGPP)
    endwin();
#endif

#if defined(X_UNIX)
    printf("\e(B");
#endif
}

int CScreen::GetMaxX()
{
    return max_x;
}

int CScreen::GetMaxY()
{
    return max_y;
}

CSubScreen *CScreen::SaveScreen(int xx1,int yy1,int xx2,int yy2)
{
    CSubScreen *ret;

    ret = new CSubScreen(xx1,yy1,xx2,yy2);

#if defined(X_UNIX) || defined(X_DJGPP)
    overwrite(stdscr,ret->Data);
#else
    gettext(xx1,yy1,xx2,yy2,ret->Data);
#endif

    return ret;
}

void CScreen::RestoreScreen(CSubScreen *Scr,int x,int y)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    mvwin(Scr->Data,x,y);
    overwrite(Scr->Data,stdscr);
#else
    int w,h;

    w = Scr->x2 - Scr->x1;
    h = Scr->y2 - Scr->y1;
    puttext(x,y,x+w,y+h,Scr->Data);
#endif

    delete Scr;
}

void CScreen::RestoreScreen(CSubScreen *Scr)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    overwrite(Scr->Data,stdscr);
#else
    puttext(Scr->x1,Scr->y1,Scr->x2,Scr->y2,Scr->Data);
#endif

    delete Scr;
}

void CScreen::Clear()
{
#if defined(X_UNIX) || defined(X_DJGPP)
    erase();
#else
    clrscr();
#endif
}

void CScreen::Clear(int n)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    move(n,0);
    clrtoeol();
#else
    gotoxy(1,n+1);
    clreol();
#endif
}

void CScreen::Clear(int x1,int y1,int x2,int y2)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    for (int j=y1;j<=y2;j++) {
	move(j,x1);
	for (int i=x1;i<=x2;i++) {
	    addch(' ');
	}
    }
#else
    for (int i=y1;i<=y2;i++) {
	for (int j=x1;j<=x2;j++) {
	    gotoxy(j+1,i+1);
	    cprintf(" ");
	}
    }
#endif
}

void CScreen::GetCursor(int &x,int &y)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    getyx(stdscr,y,x);
#else
    x = whichx() - 1;
    y = whichy() - 1;
#endif
}

void CScreen::SetCursor(int x,int y)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    move(y,x);
#else
    gotoxy(x+1,y+1);
#endif
}


void CScreen::Fill(int ch,int x1,int y1,int x2,int y2)
{
    int i,j;

#if defined(X_UNIX) || defined(X_DJGPP)
    for (i=y1;i<=y2;i++) {
	for (j=x1;j<=x2;j++) {
	    mvaddch(i,j,ch);
	}
    }
#else
    for (i=y1;i<=y2;i++) {
        for (j=x1;j<=x2;j++) {
	    gotoxy(j+1,i+1);
            cprintf("%c",ch);
        }
    }
#endif
}


void CScreen::SetColor(int fg,int bg)
{
    int intensity;
    int pair;

#if defined(X_UNIX) || defined(X_DJGPP)
    if (!has_colors())
        return;

    if (fg >= H_INTENSITY) {
	intensity = 1;
	fg -= H_INTENSITY;
    }
    else
	intensity = 0;

    pair = ColorIndex[fg][bg];

    if (intensity)
      attrset(COLOR_PAIR(pair) | A_BOLD);
    else
      attrset(COLOR_PAIR(pair) | A_NORMAL);

#else

#endif
}


void CScreen::InitColorPairs()
{
#if defined(X_UNIX) || defined(X_DJGPP)
    int i,j;
    int count;
    
    count = 1;
    for (i=0;i<8;i++) {
	for (j=0;j<8;j++) {

	    if (count >= COLOR_PAIRS)
		break;
	    
	    init_pair(count,i,j);
	    ColorIndex[i][j] = count++;
	}
    }

    count = ColorIndex[COLOR_WHITE][COLOR_BLACK];
    init_pair(count,COLOR_WHITE,COLOR_WHITE);
    ColorIndex[COLOR_WHITE][COLOR_WHITE] = count;
    ColorIndex[COLOR_WHITE][COLOR_BLACK] = 0;

#else

#endif
}

void CScreen::PrintXY(int x,int y,char *str)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    move(y,x);
    addstr(str);
#else
    gotoxy(x+1,y+1);
    cprintf("%s",str);
#endif
}

void CScreen::PutChar(int ch)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    addch(ch);
#else
    cprintf("%c",ch);
#endif
}

void CScreen::PutCharXY(int x,int y,int ch)
{
#if defined(X_UNIX) || defined(X_DJGPP)
    move(y,x);
    addch(ch);
#else
    gotoxy(x+1,y+1);
    cprintf("%c",ch);
#endif
}


void CScreen::Refresh()
{
#if defined(X_UNIX) || defined(X_DJGPP)
    refresh();
#endif
}

int CScreen::IsInit()
{
    return Init;
}
