/*
 * wcon v0.5 GUI library for the console
 * Written by Fakhir Ansari <ansari89@yahoo.com>
 *
 * Module - The main control of the interface is handled
 *          by this module.
 *
 */

#include <stdlib.h>

#include "data.h"
#include "xinter.h"
#include "xcontrol.h"
#include "screen.h"


/***********************************************************************/

CMessage::CMessage() : List(sizeof(CMessageItem))
{

}

bool CMessage::Add()
{
    List.AddTop(&Message);
}

bool CMessage::Remove()
{
    CMessageItem *m;

    if (List.GetTotal() == 0) {
	return false;
    }

    m = (CMessageItem*) List.GetItem(0);
    Message = *m;
    
    List.Delete(0);

    return true;
}


/***********************************************************************/

CControl::CControl() : WindowList(sizeof(CWindow *))
{
    ActiveWindow = NULL;
}

bool CControl::SetDesktop(char *desktop)
{
    return Desktop.LoadFile(desktop);
}

bool CControl::CreateWindow(CWindow *win)
{
    WindowList.AddBottom(&win);

    win->OnPreCreate();
    win->OnCreate();

    if (win->Property.Visible) {
        if (ActiveWindow)
            ActiveWindow->OnLoseFocus();
        ActiveWindow = win;
    }

    RepaintAll();

    return true;
}

bool CControl::CloseWindow()
{
    CWindow *Points;

    if (WindowList.GetTotal() <= 0)
	return false;

    Points = *(CWindow **)WindowList.GetItem(0);
    WindowList.Delete(0);
   
    if (Points->Property.Response) {
	void *Data[5];
	(Points->Property.Response)->OnSignal(Points->Property.Handle,
					      CSIG_ONCLOSE,
					      Data);
    }
   
    Points->OnClose();
    delete Points;

    ActiveWindow = NULL;

    SwitchWindow(0);

/* if (!WindowList.IsEmpty()) {
	Points = *(CWindow **) WindowList.GetItem(0);
	if (Points->Property.Style & CWIN_DIALOG) 
	    ActiveWindow = Points;
    } */

    return true;
}

bool CControl::HideWindow()
{
    if (ActiveWindow && ActiveWindow->Property.Visible) {
        ActiveWindow->Property.Visible = false;
        ActiveWindow = NULL;
        SwitchWindow(0);
    }

    return true;
}


bool CControl::CloseAll()
{
    int i;
    CWindow *Points;

    if (WindowList.GetTotal() <= 0)
	return true;

    for (i=0;i<WindowList.GetTotal();i++) {
	Points = *(CWindow **) WindowList.GetItem(i);
	Points->OnClose();
	delete Points;
    }

    WindowList.DeleteAll();

    ActiveWindow = NULL;

    return true;
}

bool CControl::SwitchWindow(int n)
{
    CWindow *win;
    int total = WindowList.GetTotal();

    if (total <= 0 || n >= total)
	return false;

    if (ActiveWindow) {
        if (ActiveWindow->Property.Style & CWIN_DIALOG)
            return true;
        ActiveWindow->OnLoseFocus();
    }

    int i=0, index;

    do {
        index = (n<0) ? (abs(total+n-i) % total) : ((n+i) % total);
        win = *(CWindow **)WindowList.GetItem(index);
    } while (!win->Property.Visible && i++ <= total);

    if (i <= total) {
        WindowList.Delete(index);
        WindowList.AddBottom(&win);
        ActiveWindow = win;
    }
    else
        ActiveWindow = NULL;

    return true;
}

bool CControl::RepaintAll()
{
    int i;
    CWindow *Points;

    MyScreen.Clear();

    Desktop.Display();

    for (i=WindowList.GetTotal()-1;i>=0;i--) {
	Points = *(CWindow **) WindowList.GetItem(i);

	Points->OnPaint();
	Points->OnLoseFocus();
    }

    if (ActiveWindow)
	ActiveWindow->OnGetFocus();

    return true;
}

bool CControl::Run()
{
    while (1) {
	GetEvent();
	
	while (MessageQueue.List.GetTotal() > 0)
	    DispatchEvent();

	MyScreen.Refresh();
    }
}

void CControl::GetEvent()
{
    int ch;
    int Status;
    int etype;
    int rtype;
    int n = 0;
    
    Status = 0;

    while (1) {
#if 0
        n++;
	if (n > 10) { re_idle_handler(); n = 0; }
#endif
	re_idle_handler();

	if (Mouse.Event()) {
	    if (etype=Mouse.Get()) {
                Mouse.Display();

		switch (etype) {
		case CMOUSE_MOVE: rtype = RE_MOVE; break;
		case CMOUSE_LEFT_DOWN: rtype = RE_LEFTDOWN; break;
		case CMOUSE_RIGHT_DOWN: rtype = RE_RIGHTDOWN; break;
		case CMOUSE_LEFT_UP: rtype = RE_LEFTUP; break;
		case CMOUSE_RIGHT_UP: rtype = RE_RIGHTUP; break;
		}

		MessageQueue.Message.Type = rtype;
		MessageQueue.Message.p.SetXY(Mouse.x,Mouse.y);
		MessageQueue.Message.Status = 0;

		MessageQueue.Add();

		return;
	    }
	}

	if (Keyboard.KeyPressed()) {
	    ch = Keyboard.Getch();

            if (re_key_handler(ch, 0)) {
                Control.RepaintAll();
            }
            else {
                MessageQueue.Message.Type = RE_KEYDOWN;
                MessageQueue.Message.Key = ch;
                MessageQueue.Message.Status = 0;
                MessageQueue.Add();
            }

	    return;
	}
    }
}

void CControl::DispatchEvent()
{
    CWindow *w;
    int i;

    CPoint Point;
    int Key;
    int Status;
    
    MessageQueue.Remove();

    Point = MessageQueue.Message.p;
    Key = MessageQueue.Message.Key;
    Status = MessageQueue.Message.Status;


    switch (MessageQueue.Message.Type) {
    case RE_MOVE : 
	if (ActiveWindow)
	    ActiveWindow->OnMove(Point,Status);
	break;
    case RE_LEFTUP:
	if (ActiveWindow)
	    ActiveWindow->OnLeftUp(Point,Status);
	break;
    case RE_RIGHTUP:
	if (ActiveWindow)
	    ActiveWindow->OnRightUp(Point,Status);
	break;
    case RE_LEFTDOWN:
	ControlFromPoint(Point);
	if (ActiveWindow)
	    ActiveWindow->OnLeftDown(Point,Status);
	break;
    case RE_RIGHTDOWN:
	ControlFromPoint(Point);
	if (ActiveWindow)
	    ActiveWindow->OnRightDown(Point,Status);
	break;

    case RE_KEYDOWN:
	if (ActiveWindow)
	    ActiveWindow->OnKeyDown(Key,Status);
	break;
    }

}

bool CControl::ControlFromPoint(CPoint p)
{
    CWindow *win;
    int i;

    if (ActiveWindow) {
	if (ActiveWindow->Property.Rect.IsIn(p))
	    return true;

	if (ActiveWindow->Property.Style & CWIN_DIALOG)
	    return true;
    }

    for (i=0;i<WindowList.GetTotal();i++) {
	win = *(CWindow **) WindowList.GetItem(i);

	if (win->Property.Rect.IsIn(p) && win->Property.Visible) {
	    if (ActiveWindow)
		ActiveWindow->OnLoseFocus();

	    ActiveWindow = win;
	    ActiveWindow->OnPaint();
	    ActiveWindow->OnGetFocus();

	    WindowList.Delete(i);
	    WindowList.AddBottom(&win);

	    return true;
	}
    }

    if (ActiveWindow) {
	ActiveWindow->OnLoseFocus();
    }
    
    ActiveWindow = NULL;
    return false;
}

bool CControl::Exit()
{
    MyScreen.SetCursor(0,24);
    MyScreen.Refresh();
    exit(EXIT_SUCCESS);
}

bool CControl::Init()
{
    return true;
}

CWindow *CControl::GetActiveWindow()
{
    return ActiveWindow;
}
