/*
 * mp3enc.c: Distributed MP3 encoder. Uses the lame encoder.
 * (C) Copyright 2003, netcluster
 *
 * Syntax: mp3enc num_of_segments input_file
 *         Make sure the file is given with an absolute path.
 *
 * Description:
 *   The program creates a main task which distributes either
 *   a single wav file to N processes for encoding. Or it
 *   takes N input files and creates N processes, one for
 *   encoding each file. The .MP3 files are generated in the
 *   directory specified on the command line.
 *
 * File distribution written by Ghaus Ali
 * MP3 encoding support added by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <time.h>

#include "libnc.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * APPLICATION FUNCTIONS AND DATA
 * Here are the application dependent functions and
 * data. This section also contains some functions for
 * interfacing with the libmp3lame.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "lame/lame.h"

/* Number of channels in raw data */
#define NUM_CHANNELS 2

/* Size of the blocks which are sent remotely for encoding
 * and buffer sizes for storing these chunks before and
 * after encoding. Is assumed that outbuffer < inbuffer. */
#define CHUNK_SIZE (2304*100)
#define INBUFFER_SIZE CHUNK_SIZE
#define OUTBUFFER_SIZE (CHUNK_SIZE+1024)

#define MAX_PATH_LEN 128

/* Global lame flags */
lame_global_flags *gf;


/* mp3enc_init: Used to initialize the lame library */
int mp3enc_init()
{
	if ((gf = lame_init()) == NULL) {
		fprintf(stderr, "mp3enc: Unable to initialize\n");
		return FALSE;
	}

	lame_set_bWriteVbrTag(gf, 0);

	if (lame_init_params(gf) < 0) {
		fprintf(stderr, "lame_init_params: Unable to initialize\n");
		return FALSE;
	}

	/*
	lame_print_config(gf);
	lame_print_internals(gf);
	*/

	return TRUE;
}

/* mp3enc_deinit: Lame library deinitialized here */
int mp3enc_deinit()
{
	lame_close(gf);
	return TRUE;
}

/* get_audio: Buffer size should be atleast INBUFFER_SIZE */
int get_audio(FILE *infile, short int buffer[INBUFFER_SIZE])
{
	int samples_to_read;
	int samples_read;

	/* samples_to_read = lame_get_framesize(gf); */

	samples_read = fread(buffer, sizeof(short int), INBUFFER_SIZE, infile);
	samples_read /= NUM_CHANNELS;

	/* Shift data to upper 16 bits of int */
	/*
	for (i=0;i<samples_read;i++)
		buffer[i] = buffer16[i] << (8 * sizeof(int) - 16);
	*/

	return samples_read;
}

/* outf_name: Returns the name of the output mp3 file
 * given the name of the raw data file. */
char *outf_name(char *iname)
{
	int len;
	static char oname[MAX_PATH_LEN];

	strcpy(oname, iname);
	len = strlen(oname);

	if (len > 3 && oname[len-4]=='.')
		strcpy(oname+len-4, ".mp3");
	else
		strcat(oname, ".mp3");

	return oname;
}

/*
 * Serialization of jobs for out of order completion
 */
typedef struct _joblist_t {
	int jobno;
	rawdata_t buf;
	struct _joblist_t *next;
} joblist_t;

joblist_t *joblist = NULL;

/* job_write: Either write the specified job to file if it
 * is the next to be written or enque it if its out of
 * order. Returns 1 if job was written or 2 if enqueued. */
int job_write(FILE *outf, int jobno, rawdata_t *buf)
{
	static int next_job = 1;
	joblist_t *newjob;
	joblist_t *cur;
	joblist_t *prev;

	if (jobno == next_job) {
		fwrite(buf->data, 1, buf->size, outf);

		next_job++;
		while (joblist && joblist->jobno == next_job) {
			fwrite(joblist->buf.data, 1, joblist->buf.size, outf);
			next_job++;

			prev = joblist;
			joblist = joblist->next;

			if (prev->buf.data)
				free(prev->buf.data);
			free(prev);
		}

		return 1;
	}

	newjob = malloc(sizeof(joblist_t));
	newjob->jobno = jobno;
	newjob->buf = *buf;

	prev = NULL;
	cur = joblist;
	while (cur && cur->jobno < jobno) {
		prev = cur;
		cur = cur->next;
	}

	if (!prev) {
		newjob->next = joblist;
		joblist = newjob;
	}
	else {
		newjob->next = cur;
		prev->next = newjob;
	}

	return 2;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * USER TASKS
 * The task_main is the task entry point for all binaries.
 * The rest of the subtasks could be arbitratily named but
 * no other subtask takes any arguments or return values
 * except for task_main. The first argument (argv[0]) of
 * task main is the start symbol (i.e. "task_main"). The rest
 * are the arguments given on the command line after the
 * binary name.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define encoder_init 0
#define encoder_jobdata 1
#define distributer_jobdone 2
RPC_FUNC(void encoder_init(task_id distributer))
RPC_FUNC(void encoder_jobdata(int jobno, int samples, rawdata_t raw))
RPC_FUNC(int distributer_jobdone(task_id jobenc, int jobno, rawdata_t raw))

/* The task_main is the raw data distributer */
void task_main(int argc, char *argv[])
{
	FILE *inf;
	FILE *outf;
	int is_eof;
	int num_encoders;
	task_id myself;
	task_id *encoders;
	task_id jobenc;
	rawdata_t raw;
	rawdata_t unraw;
	int samples;
	int out_job;
	int done_job;
	int i;
	time_t timer;

	/* Initialization and argument handling */

	if (argc != 3) {
		fprintf(stderr, "Syntax: mp3enc num_of_segments input_file\n");
		fprintf(stderr, "File should be with an absolute path.\n");
		return;
	}

	num_encoders = atoi(argv[1]);
	if (num_encoders <= 0) {
		fprintf(stderr, "Number of segments must be greater than 0\n");
		return;
	}
	encoders = malloc(sizeof(task_id)*num_encoders);

	inf = fopen(argv[2], "rb");
	outf = fopen(outf_name(argv[2]), "wb+");

	if (!inf || !outf) {
		fprintf(stderr, "Unable to open input<%s> or output<%s> file. Aborting\n",
				argv[2], outf_name(argv[2]));
		return;
	}

	fprintf(stderr, "Loaded main distributer task\n");

	timer = time(NULL);

	/* All the initialization is successful. Now proceed
	 * with the creation of encoders and their init. */

	myself = nc_myid();

	for (i=0;i<num_encoders;i++) {
		encoders[i] = nc_createtask("encoder");
		nc_call(&encoders[i], encoder_init, myself);
	}

	fprintf(stderr, "%d encoders successfully created and initialized\n",
			num_encoders);
	fprintf(stderr, "Sending data...\n");

	/* Send each encoder data to get them started. After
	 * the first assignment of a job, they would be reassigned
	 * a new job on completion of the pervious one. */

	raw.data = malloc(INBUFFER_SIZE*sizeof(short int));
	out_job = 1;

	for (i=0;i<num_encoders;i++) {
		samples = get_audio(inf, (short int*)raw.data);
		if (samples > 0) raw.size = samples*2*sizeof(short int);
		else raw.size = 0;
		nc_call(&encoders[i], encoder_jobdata, out_job, samples, raw);
		out_job++;
	}

	while (num_encoders) {
		is_eof = feof(inf);
#if 0
if (out_job > 7)
	is_eof = 1;
#endif

		if (nc_accept(distributer_jobdone, &jobenc, &done_job, &unraw)) {
			nc_treturn(0, is_eof);

			fprintf(stderr, "Job %d completed\n", done_job);

			if (job_write(outf, done_job, &unraw) == 1) {
				if (unraw.data)
					free(unraw.data);
			}

			if (is_eof)
				num_encoders--;
			else {
				samples = get_audio(inf, (short int*)raw.data);
				if (samples > 0) raw.size = samples*2*sizeof(short int);
				else raw.size = 0;
				nc_call(&jobenc, encoder_jobdata, out_job, samples, raw);
				out_job++;
			}

		}
	}

	timer = time(NULL)-timer;
	fprintf(stderr, "Time taken: %02u:%02u seconds\n", timer/60, timer%60);

	if (raw.data)
		free(raw.data);

	fclose(inf);
	fclose(outf);

	if (encoders)
		free(encoders);
}

SUBTASK(encoder)
{
	int samples;
	int imp3;
	int i;
	int done;
	int jobno;
	task_id distributer;
	task_id myself;
	rawdata_t raw;
	rawdata_t unraw;

	fprintf(stderr, "Encoder task loaded\n");

	if (nc_accept(encoder_init, &distributer)) {
		fprintf(stderr, "Encoder task initialized\n");
		nc_treturn(0);
	}
	else
		fprintf(stderr, "Encoder task NOT\n");

	myself = nc_myid();
	unraw.data = malloc(OUTBUFFER_SIZE*sizeof(short int));

	if (mp3enc_init())
		fprintf(stderr, "Encoder successfully initialized\n");
	else {
		fprintf(stderr, "Encoder initialization failed\n");
		return;
	}

	do {
		if (!nc_accept(encoder_jobdata, &jobno, &samples, &raw))
			continue;
		nc_treturn(0);

		fprintf(stderr, "Received job %d\n", jobno);

		/* Encoder the raw data here */
		imp3 = lame_encode_buffer_interleaved(
				gf,
				(short int*)raw.data,
				samples,
				unraw.data,
				OUTBUFFER_SIZE*sizeof(short int));
		imp3 += lame_encode_flush_nogap(gf, unraw.data+imp3,
				OUTBUFFER_SIZE*sizeof(short int)-imp3);

		/* was our output buffer big enough? */
		if (imp3 < 0) {
			if (imp3 == -1)
				fprintf(stderr, "MP3 buffer is not big enough\n");
			else
				fprintf(stderr, "Internal error: %d\n", imp3);
			return;
		}

		lame_init_bitstream(gf);

		unraw.size = imp3;
		nc_call(&distributer, distributer_jobdone, myself, jobno, unraw, &done);

		fprintf(stderr, "Submitted job %d\n", jobno);

		/* Data must be free'd after every nc_accept */
		free(raw.data);
	} while (!done);

	mp3enc_deinit();

	if (unraw.data)
		free(unraw.data);

	fprintf(stderr, "Encoder deinitialized\n");
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * LIBNC DATA
 * This section is HIDDEN. It is currently manually written but
 * would later be generated by ncpp, the netcluster preprocessor,
 * based on the above source.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* User function definitions indexed by function ID */
nc_rpcdef_t nc_rpcs[] = {
	{ "encoder_init", "%t", "%z" },
	{ "encoder_jobdata", "%d%d%r", "%z" },
	{ "distributer_jobdone", "%t%d%r", "%d" },
	{ NULL, NULL, NULL }
};

/* User task definitions not in any particular order */
nc_taskdef_t nc_tasks[] = {
	{ "encoder", encoder },
	{ NULL, NULL }
};
