/*
 * rdwt.c: Producer/consumer problem with intermediate buffer
 * (C) Copyright 2003, netcluster
 *
 * Syntax: rdwt file_with_abs_path [buffer_size]
 *
 * Description:
 *   The program creates a producer, a consumer and a buffer
 *   process. The producer reads data from the file given on
 *   the command line and sends it character by character to
 *   the buffer process. Each time a newline character is
 *   encountered, a complete line of text is sent to the
 *   consumer process. which then displays the output to its
 *   stdout. The filename passed on the command line should be
 *   an absolute path to a world readable, or atleast readble
 *   by the local daemon uid, file so that the tasks have
 *   access to it.
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

#include "libnc.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * UTILITY FUNCTIONS
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

char *str_extract(char *str, int head, int tail, int size)
{
	int i;
	char *substr;
	int len = 0;

	for (i=tail;i!=head;i=(i+1)%size)
		len++;

	substr = malloc(len+1);
	len = 0;

	for (i=tail;i!=head;i=(i+1)%size)
		substr[len++] = str[i];

	substr[len] = '\0';

	return substr;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * USER TASKS
 * The task_main is the task entry point for all binaries.
 * The rest of the subtasks could be arbitratily named but
 * no other subtask takes any arguments or return values
 * except for task_main. The first argument (argv[0]) of
 * task main is the start symbol (i.e. "task_main"). The rest
 * are the arguments given on the command line after the
 * binary name.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define DEFAULT_BUFFER_SIZE 1024

#define buffer_put 0
#define buffer_get 1
#define buffer_init 2
#define consumer_init 3
RPC_FUNC(void buffer_put(int ch))
RPC_FUNC(int-char * buffer_get())
RPC_FUNC(void buffer_init(int size))
RPC_FUNC(void consumer_init(int size, task_id buffer))

/* The task_main is the producer */
void task_main(int argc, char *argv[])
{
	task_id buffer;
	task_id consumer;
	FILE *infile;
	int buf_size = DEFAULT_BUFFER_SIZE;
	int count = 0;

	if (argc < 2) {
		fprintf(stderr, "Syntax: rdwt filename_with_abs_path [buffer_size]\n");
		fprintf(stderr, "Filename argument missing. Exiting...\n");
		return;
	}

	if (argc == 3) {
		int size = atoi(argv[2]);
		if (size > 0)
			buf_size = size;
	}

	infile = fopen(argv[1], "rt");

	if (!infile) {
		fprintf(stderr, "Unable to read input file. Please specify an"
				"absolute path to a world readable text file\n");
	}

	fprintf(stderr, "Loaded main producer task\n");

	buffer = nc_createtask("buffer");
	fprintf(stderr, "Buffer successfully created\n");

	consumer = nc_createtask("consumer");
	fprintf(stderr, "Consumer successfully created\n");

	/* Initialize the tasks */
	nc_call(&buffer, buffer_init, buf_size);
	nc_call(&consumer, consumer_init, buf_size, buffer);

	fprintf(stderr, "Sending data\n");

	/* Start producing */
	while (!feof(infile)) {
		int ch;

		ch = fgetc(infile);
		nc_call(&buffer, buffer_put, ch);

		count++;
		if (!(count%10)) fputc('.', stderr);
		if (!(count%400)) fputc('\n', stderr);
	}

	fprintf(stderr, "\nFile produce completed. Sent %d bytes.\n", count);
	fclose(infile);
}

SUBTASK(buffer)
{
	int size;
	char *str;
	int ch;
	int iseof;
	int head;
	int tail;
	int flush_no;

	fprintf(stderr, "Buffer task loaded\n");

	if (nc_accept(buffer_init, &size)) {
		fprintf(stderr, "Setting buffer size to: %d\n", size);
		nc_treturn(0);
	}

	str = malloc(size);
	head = tail = 0;
	flush_no = 0;
	iseof = FALSE;

	do {
		if ((head+1)%size != tail) {
			if (nc_accept_noblock(buffer_put, &ch)) {
				if (ch >= 0) {
					str[head] = ch;
					head = (head+1)%size;
				}
				else
					iseof = TRUE;

				nc_treturn(0);
			}
		}
		if (head != tail) {
			char *substr;

			if (nc_accept_noblock(buffer_get)) {
				substr = str_extract(str, head, tail, size);
				fprintf(stderr, "Buffer flushed: %d\n", ++flush_no);
				nc_treturn(0, iseof, substr);
				free(substr);
				tail = head;
			}
			else if (iseof && nc_accept(buffer_get)) {
				substr = str_extract(str, head, tail, size);
				fprintf(stderr, "Buffer flushed: %d\n", ++flush_no);
				nc_treturn(0, iseof, substr);
				free(substr);
				tail = head;
			}
		}
	} while (!iseof);

	fprintf(stderr, "EOF encountered\n");
}

SUBTASK(consumer)
{
	task_id buffer;
	int size;
	char *str;		/* return value of buffer_get */
	int iseof = FALSE;	/* return value of buffer_get */

	fprintf(stderr, "Consumer task loaded. Displaying file...\n");
	fprintf(stderr, "------------ 8< ------------\n");

	if (nc_accept(consumer_init, &size, &buffer))
		nc_treturn(0);

	str = (char*)malloc(size+1);

	while (!iseof) {
		sleep(1);
		nc_call(&buffer, buffer_get, &iseof, str);
		fprintf(stderr, "%s", str);
	}

	fprintf(stderr, "------------ >8 ------------\n");
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * LIBNC DATA
 * This section is HIDDEN. It is currently manually written but
 * would later be generated by ncpp, the netcluster preprocessor,
 * based on the above source.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* User function definitions indexed by function ID */
nc_rpcdef_t nc_rpcs[] = {
	{ "buffer_put", "%d", "%z" },
	{ "buffer_get", "%z", "%d%s" },
	{ "buffer_init", "%d", "%z" },
	{ "consumer_init", "%d%t", "%z" },
	{ NULL, NULL, NULL }
};

/* User task definitions not in any particular order */
nc_taskdef_t nc_tasks[] = {
	{ "buffer", buffer },
	{ "consumer", consumer },
	{ NULL, NULL }
};
