/*
 * chash.c: A hash table implementation.
 * (C) Copyright 2003, netcluster
 *
 * Written by Syed Ghaus Ali
 * Rechecked, tested, fixed and extended - Fakhir Ansari
 *
 * Notes:
 * The interface is quite similar to that of clist but all
 * these functions assume that the htable argument is not
 * NULL. And always initialize using chash_init.
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>

#include "chash.h"
#include "defs.h"
#include "clist.h"


/*
 * chash_init: Initialize the hash table. This must be called the
 * very first time on the list. A hash table of the specified is
 * set along with the hash function. The free_func is a function
 * similar to free() which is used to free a single item of the
 * hash table. It can also be NULL resulting in the hash table
 * items not being free'd by this module.
 * Note that the hash function must return an int from 0 to size-1
 * for any given data.
 * Returns NULL on failure.
 */
chash *chash_init(chash *htable, int size, CHASH_FUNC(hash_func),
		CITEM_FREE(free_func))
{
	int i;

	htable->vector = (clist**)malloc(sizeof(clist*) * size);
	
	for (i=0;i<size;i++)
		htable->vector[i] = NULL;
	
	htable->size = size;
	htable->hash_func = hash_func;
	htable->free_func = free_func;

	return htable;
}

/* All the following functions assume that the table has been init */

/*
 * chash_insert: Inserts a new item to the hash table.
 * Returns the head of list in which the item was inserted.
 */
clist *chash_insert(chash *htable, void *data)
{
	int hash = (*htable->hash_func)(data, htable->size);
	clist *list = htable->vector[hash];
	htable->vector[hash] = clist_append(list, data);
	return htable->vector[hash];
}

/*
 * chash_insert_sorted: Inserts an item and makes sure it is
 * sorted within the list it is being added to. Returns the
 * head of the list the item was added to.
 */
clist *chash_insert_sorted(chash *htable, void *data, CCOMPARE_FUNC(ccfunc))
{
	int hash = (*htable->hash_func)(data, htable->size);
	clist *list = htable->vector[hash];
	htable->vector[hash] = clist_insert_sorted(list, data, ccfunc);
	return htable->vector[hash];
}

/*
 * chash_remove: Removes an item from the hash table and returns
 * the list from which it was removed from.
 */
clist *chash_remove(chash *htable, void *data)
{
	int hash = (*htable->hash_func)(data, htable->size);
	htable->vector[hash] = clist_remove_custfree(htable->vector[hash],
			data, htable->free_func);
	return htable->vector[hash];
}

/*
 * chash_remove_custom: Removes a value from the hash table
 * by checking each value against ccfunc. The ccfunc should
 * return 0 for the item to be deleted. Only the first
 * match is deleted and the function returns NULL if the
 * item was not found and the head of the list from which
 * the item was deleted, incase it was found.
 */
clist *chash_remove_custom(chash *htable, void *user_data, CCOMPARE_FUNC(ccfunc))
{
	int i;
	clist *find = NULL;

	for (i=0;i<htable->size;i++) {
		find = clist_find_custom(htable->vector[i],
				user_data, ccfunc);
		if (find) {
			if (htable->free_func)
				(*htable->free_func)(find->data);
			htable->vector[i] = clist_remove_link(htable->vector[i],
								find);
			return htable->vector[i];
		}
	}

	return NULL;
}

/*
 * chash_free: Frees the whole hash table.
 */
void chash_free(chash *htable)
{
	int i;

	for (i=0;i<htable->size;i++)
		htable->vector[i] = clist_custfree(htable->vector[i],
				htable->free_func);

	free(htable->vector);
	htable->vector = NULL;
	htable->size = 0;
}

/*
 * chash_find: Finds the item containing 'data' in the
 * hash table and returns its clist container. Returns NULL
 * if the item is not found.
 */
clist *chash_find(chash *htable, void *data)
{
	int hash = (*htable->hash_func)(data, htable->size);
	clist *list = htable->vector[hash];
	return clist_find(list, data);
}

/*
 * chash_find_custom: Finds the first match of an item for
 * which the ccfunc returns 0. If not found, returns NULL.
 */
clist *chash_find_custom(chash *htable, void *user_data, CCOMPARE_FUNC(ccfunc))
{
	int i;
	clist *find = NULL;

	for (i=0;i<htable->size;i++) {
		find = clist_find_custom(htable->vector[i],
				user_data, ccfunc);
		if (find)
			break;
	}

	return find;
}

/*
 * chash_find_min_head: Calls ccfunc on the head of each list and
 * compares all the heads using ccfunc. The smallest head in comparison
 * is returned by this function. Is useful it all the items have
 * been added using chash_insert_sorted and the minimum (or max)
 * values item is required thus only the heads need be searched.
 * Returns NULL if the hash table is empty.
 */
clist *chash_find_min_head(chash *htable, CCOMPARE_FUNC(ccfunc))
{
	int i;
	clist *cur;
	clist *min = NULL;

	for (i=0;i<htable->size;i++) {
		if (htable->vector[i]) {
			min = htable->vector[i];
			break;
		}
	}

	for (i++;i<htable->size;i++) {
		cur = htable->vector[i];
		if (cur) {
			if (ccfunc(min->data, cur->data) > 0)
				min = cur;
		}
	}

	return min;
}

/*
 * chash_nth: Returns the nth item's clist container or NULL
 * if the list has less then n items. Item indices are zero based
 */
clist *chash_nth(chash *htable, int n)
{
	int i;
	int cur_count;
	int tot_count = 0;

	for (i=0;i<htable->size;i++) {
		cur_count = clist_length(htable->vector[i]);
		
		if (n < tot_count+cur_count)
			return clist_nth(htable->vector[i], n-tot_count);

		tot_count += cur_count;
	}

	return NULL;
}

/*
 * chash_nth_data: Same as chash_nth but this function returns
 * the actual data of the nth item instead of the clist container.
 */
void *chash_nth_data(chash *htable, int n)
{
	int i;
	int cur_count;
	int tot_count = 0;

	for (i=0;i<htable->size;i++) {
		cur_count = clist_length(htable->vector[i]);
		
		if (n < tot_count+cur_count)
			return clist_nth_data(htable->vector[i], n-tot_count);

		tot_count += cur_count;
	}

	return NULL;
}

/*
 * chash_foreach: Calls cifunc for each item in the hash table. The
 * cifunc is passed two arguments, the data of the item followed by
 * the user supplied user_data.
 */
void chash_foreach(chash *htable, CITEM_FUNC(cifunc), void *user_data)
{
	int i;

	for (i=0;i<htable->size;i++) {
		if (htable->vector[i])
			clist_foreach(htable->vector[i], cifunc, user_data);
	}
}

/*
 * chash_length: Returns the total number of items stored.
 */
int chash_length(chash *htable)
{
	int i;
	int total = 0;

	for (i=0;i<htable->size;i++) {
		if (htable->vector[i])
			total += clist_length(htable->vector[i]);
	}

	return total;
}
