/*
 * clist.c: A generic linked list data structure
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>

#include "defs.h"
#include "clist.h"

/*
 * clist: A double linked list. Check header file for the interface.
 */


/* clist_alloc: Inline function to allocate a single list element */
clist *clist_alloc()
{
	return malloc(sizeof(clist));
}

/* clist_free_1: Inline function to deallocate a single list element */
void clist_free_1(clist *list)
{
    if (list)
        free(list);
}

/* clist_append: Adds a new item to the end of the list. Returns the new
 * head of the list. The head must always be reassigned the returned value */
clist *clist_append(clist *list, void *data)
{
	clist *old_list = list;
	clist *new_list = clist_alloc();

	new_list->data = data;
	new_list->next = NULL;
	new_list->prev = NULL;

	if (!list)
		return new_list;

	while (list->next != NULL)
		list = list->next;

	list->next = new_list;
	new_list->prev = list;

	return old_list;
}

/* clist_prepend: Prepends the given data to the list and returns the new
 * head. The head must always be reassigned the returned value */
clist *clist_prepend(clist *list, void *data)
{
	clist *new_list = clist_alloc();

	new_list->data = data;
	new_list->next = list;
	new_list->prev = NULL;

	if (list)
		list->prev = new_list;

	return new_list;
}

/* clist_insert: Insert an item at the given zero based index
 * in the list. And returns the new head of the list. */
clist *clist_insert(clist *list, void *data, int pos)
{
	int i = 0;
	clist *old_list = list;
	clist *new_list = clist_alloc();
	new_list->data = data;
	new_list->next = new_list->prev = NULL;

	if (!list)
		return new_list;

	while (list && i<pos) {
		list = list->next;
		i++;
	}

	if (!list) {
		clist_free_1(new_list);
		return clist_append(old_list, data);
	}

	new_list->next = list;
	new_list->prev = list->prev;

	if (list->prev)
		list->prev->next = new_list;
	else
		old_list = new_list;

	list->prev = new_list;

	return old_list;
}

/* clist_insert_sorted: Inserts an item in the list at the correct sort
 * position and returns the new head. ccfunc must return 1 if 'a' should
 * come after 'b' */
clist *clist_insert_sorted(clist *list, void *data, CCOMPARE_FUNC(ccfunc))
{
	clist *old_list = list;
	clist *new_list = clist_alloc();
	new_list->data = data;
	new_list->next = new_list->prev = NULL;

	if (!list)
		return new_list;

	while (list) {
		int rel = (*ccfunc)(list->data, data);
		if (rel>=0)
			break;

		list = list->next;
	}

	if (!list) {
		clist_free_1(new_list);
		return clist_append(old_list, data);
	}

	new_list->next = list;
	new_list->prev = list->prev;

	if (list->prev)
		list->prev->next = new_list;
	else
		old_list = new_list;

	list->prev = new_list;

	return old_list;
}

/* clist_remove: Searches the list for the given data with pointer
 * comparisons and removes the item if found. Returns new head of list */
clist *clist_remove(clist *list, void *data)
{
	clist *old_list = list;

	while (list) {
		if (list->data == data) {
			if (list->prev)
				list->prev->next = list->next;
			else
				old_list = list->next;

			if (list->next)
				list->next->prev = list->prev;

			clist_free_1(list);
			break;
		}

		list = list->next;
	}

	return old_list;
}

/* clist_remove_free: Similar to clist_remove but with a custom
 * data free handler. Returns the new head of the list. */
clist *clist_remove_custfree(clist *list, void *data, CITEM_FREE(cfree))
{
	clist *old_list = list;

	while (list) {
		if (list->data == data) {
			if (list->prev)
				list->prev->next = list->next;
			else
				old_list = list->next;

			if (list->next)
				list->next->prev = list->prev;

			if (cfree)
				(*cfree)(list->data);
			clist_free_1(list);
			break;
		}

		list = list->next;
	}

	return old_list;
}

/* clist_remove_link: Removes an item from the list by the specified
 * link. Performs a pointer comparison and returns the new head. */
clist *clist_remove_link(clist *list, clist *link)
{
	clist *old_list = list;

	while (list) {
		if (list == link) {
			if (list->prev)
				list->prev->next = list->next;
			else
				old_list = list->next;

			if (list->next)
				list->next->prev = list->prev;

			break;
		}

		list = list->next;
	}

	return old_list;
}

/* clist_free: Deallocates the whole list and returns the new head
 * which is always NULL ofcourse. */
clist *clist_free(clist *list)
{
	clist *tlist;

	while (list) {
		tlist = list;
		list = list->next;
		clist_free_1(tlist);
	}

	return NULL;
}

/* clist_custfree: Deallocates the whole list and returns the new head
 * which is always NULL ofcourse. Uses a custom free() on each data item. */
clist *clist_custfree(clist *list, CITEM_FREE(cfree))
{
	clist *tlist;

	while (list) {
		tlist = list;
		list = list->next;
		if (cfree)
			(*cfree)(tlist->data);
		clist_free_1(tlist);
	}

	return NULL;
}

/* clist_length: Returns the total items in the list. */
int clist_length(clist *list)
{
	int len = 0;

	while (list) {
		list = list->next;
		len++;
	}

	return len;
}

/* clist_copy: Returns a duplicate of the given list.
 * NOTE: Can be improved coz it keeps calling clist_append
 * for each item and so is inefficient. */
clist *clist_copy(clist *list)
{
	clist *new_list = NULL;

	while (list) {
		new_list = clist_append(new_list, list->data);
		list = list->next;
	}

	return new_list;
}

/* clist_reverse: Reverses the given list and returns the new head. */
clist *clist_reverse(clist *list)
{
	void *tdata;
	clist *old_list = list;

	while (list) {
		tdata = list->next;
		list->next = list->prev;
		list->prev = tdata;

		old_list = list;
		list = list->prev;
	}

	return old_list;
}

/* clist_sort: Sorts the given list according to the compare function.
 * ccfunc must return 1 if 'a' should come after 'b', -1 if 'b' comes
 * after 'a' and 0 if 'a' and 'b' are equal
 * NOTE: Implemented very poorly. Would reimplement later. */
clist *clist_sort(clist *list, CCOMPARE_FUNC(ccfunc))
{
	clist *new_list = NULL;

	while (list) {
		new_list = clist_insert_sorted(new_list, list->data, ccfunc);
		list = list->next;
	}

	clist_free(list);
	return new_list;
}

/* clist_concat: Concatenates list2 to list1 and returns the new head. */
clist *clist_concat(clist *list1, clist *list2)
{
	if (!list1)
		return list2;
	if (!list2)
		return list1;

	while (list1->next)
		list1 = list1->next;

	list1->next = list2;
	list2->prev = list1;

	return list1;
}

/* clist_foreach: Calls the given cifunc for each item in the specified
 * list. Also provides each call of cifunc with the arbitrary user_data. */
void clist_foreach(clist *list, CITEM_FUNC(cifunc), void *user_data)
{
	while (list) {
		(*cifunc)(list->data, user_data);
		list = list->next;
	}
}

/* clist_first: Returns the head of the list which in this implementation
 * also happens to be the list itself. */
clist *clist_first(clist *list)
{
	return list;
}

/* clist_last: Returns the last item of the list. */
clist *clist_last(clist *list)
{
	if (!list)
		return NULL;

	while (list->next)
		list = list->next;

	return list;
}

/* clist_nth: Returns the item at the given zero-based index. */
clist *clist_nth(clist *list, int n)
{
	int i=0;

	while (i<n && list) {
		list = list->next;
		i++;
	}

	return list;
}

/* clist_nth_data: Returns the data at the given zero-based index.
 * Returns NULL if the given index is out of bounds. */
void *clist_nth_data(clist *list, int n)
{
	int i=0;

	if (!list)
		return NULL;

	while (i<n && list) {
		list = list->next;
		i++;
	}

	if (list)
		return list->data;
	else
		return NULL;
}

/* clist_find: Find the given data in the list and return the item
 * which stores this data. Returns NULL if data not found. */
clist *clist_find(clist *list, void *data)
{
	while (list) {
		if (list->data == data)
			break;

		list = list->next;
	}

	return list;
}

/* clist_find_custom: Finds an item from the list by using the user
 * supplied function. ccfunc should return 0 when the item is found. */
clist *clist_find_custom(clist *list, void *user_data, CCOMPARE_FUNC(ccfunc))
{
	while (list) {
		if (!(*ccfunc)(list->data, user_data))
			break;

		list = list->next;
	}

	return list;
}

/* clist_position: returns the position of 'find' in the given list and
 * -1 returned if it is not found in the list. */
int clist_position(clist *list, clist *find)
{
	int i = 0;

	while (list) {
		if (list == find)
			return i;

		i++;
		list = list->next;
	}

	return -1;
}

/* clist_index: Returns the index of the item in which the data is
 * is stored and -1 returned if not found. */
int clist_index(clist *list, void *data)
{
	int i = 0;

	while (list) {
		if (list->data == data)
			return i;

		i++;
		list = list->next;
	}

	return -1;
}
