/*
 * conf.c: A generic config file parser
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>

#include "defs.h"
#include "sock.h"
#include "util.h"
#include "conf.h"
#include "log.h"



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                       T a g   H a n d l e r s
 *
 * These are the tag handlers used to handle different tags in the config
 * files. To define a new handler, add the #define to conf.h, write the
 * tag handler and add it to the conftag_handlers list.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* taghnd_str: Stored as a char* in the list. Must not be qouted unless
 * quotes are part of the string. */
CONFTAG_HANDLER(taghnd_str)
{
	char *str = strdup(targ);
	tinf->values = clist_append(tinf->values, str);
	return TRUE;
}

/* taghnd_qstr: Stored as a char* in the list. This is the quoted
 * version of the string. The quotes are optional but MUST be
 * used if the string contains whitespaces. */
CONFTAG_HANDLER(taghnd_qstr)
{
	char *str;
	char *word;
       
	word = next_arg(targ, &targ);
	
	if (!str_isempty(targ))
		return FALSE;
	
	str = strdup(word);
	tinf->values = clist_append(tinf->values, str);
	return TRUE;
}

/* taghnd_int: A signed integer stored as int* in the list. */
CONFTAG_HANDLER(taghnd_int)
{
	int *iptr;

	if (!str_isinteger(targ))
		return FALSE;
	
	iptr = (int*)malloc(sizeof(int));
	*iptr = atoi(targ);
	tinf->values = clist_append(tinf->values, iptr);
	return TRUE;
}

/* taghnd_float: Stored as a float* in the list. */
CONFTAG_HANDLER(taghnd_float)
{
	float *fptr;

	if (!str_isfloat(targ))
		return FALSE;
	
	fptr = (float*)malloc(sizeof(int));
	*fptr = (float)atof(targ);
	tinf->values = clist_append(tinf->values, fptr);
	return TRUE;
}

/* tanhnd_naddr: Stored as naddr_t* in the list. */
CONFTAG_HANDLER(taghnd_naddr)
{
	naddr_t *na = (naddr_t*)malloc(sizeof(naddr_t));

	if (!str_to_naddr(targ, na))
		return FALSE;

	tinf->values = clist_append(tinf->values, na);
	return TRUE;
}

/* tanhnd_naddr: Stored as int*=1 in the list. Otherwise list
 * size is zero if the tag is not defined. */
CONFTAG_HANDLER(taghnd_null)
{
	int *iptr = (int*)malloc(sizeof(int));
	*iptr = TRUE;
	tinf->values = clist_append(tinf->values, iptr);
	return TRUE;
}


conftag_handler_t conftag_handlers[] = {
	taghnd_str,
	taghnd_qstr,
	taghnd_int,
	taghnd_float,
	taghnd_naddr,
	taghnd_null
};

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *              G e n e r i c   C o n f i g   I n t e r f a c e
 *
 * Below are the functions used to read config files. They are generic
 * and arnt restricted to any particular config format.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/*
 * get_tag: Searches the NULL terminated array of tags for the
 * given tag name. Returns a ptr to the taginfo struct if found
 * and NULL otherwise.
 */
taginfo_t *get_tag(char *tname, taginfo_t tags[])
{
	int i;

	for (i=0;tags[i].tag;i++)
		if (!strcmp(tname, tags[i].tag))
			return &tags[i];

	return NULL;
}
 
/*
 * conf_read: Reads the configuration from the specified file and
 * fills in the taginfo struct. On success returns TRUE and on
 * error, FALSE. An error condition occurs if an unrecognized tag
 * or a tag with incorrect arguments is found in the config file.
 */
int conf_read(char *cfpath, taginfo_t tags[])
{
	char *tagname;
	taginfo_t *taginf;
	int lno = 0;
	char linebuf[MAX_LINE_SIZE];
	FILE *cfile = fopen(cfpath, "rt");

	if (cfile == NULL) {
		yell("Unable to read config file %s", cfpath);
		return FALSE;
	}

	while (fgets(linebuf, sizeof(linebuf), cfile)) {
		char *line = linebuf;
		int wc;

		lno++;

		if (line[strlen(line)-1] == '\n')
			line[strlen(line)-1] = '\0';

		trim_spaces(line);

		if (!(*line) || *line==COMMENT_CHAR)
			continue;

		tagname = next_arg(line, &line);

		/* Fetch the specified tag from the list of tags */
		if (!(taginf = get_tag(tagname, tags))) {
			yell("%s:%d: Unrecognized tag (%s)",
					cfpath, lno, tagname);
			return FALSE;
		}

		/* Check if multiple instances of this tag are allowed */
		if (!taginf->multiple && taginf->values) {
			yell("%s:%d: Multiple values of tag (%s) not allowed",
					cfpath, lno, tagname);
			return FALSE;
		}

		/* Check if the character set for this tag is correct */
		if (taginf->charset && *(taginf->charset)) {
			if (!str_ischarset(line, taginf->charset)) {
				yell("%s:%d: Invalid format for tag (%s)",
						cfpath, lno, tagname);
				return FALSE;
			}
		}

		if (!str_validquotes(line)) {
			yell("%s:%d: Mismatched \" in string", cfpath, lno);
			return FALSE;
		}

		/* Call the format handler for this tag */
		if (taginf->type < NUMBER_OF_TAGHANDLERS) {
			conftag_handler_t *thand;
			thand = &conftag_handlers[taginf->type];
			if (thand->func) {
				wc = word_count(line);
				if (thand->func(taginf, line) == FALSE) {
					yell("%s:%d: Invalid format",
							cfpath, lno);
					return FALSE;
				}
			}	
		}
	}

	return TRUE;
}

/*
 * conf_free: Frees all the data allocated for tags after a call
 * to conf_read. Returns the number of tags free'd.
 */
int conf_free(taginfo_t tags[])
{
	int i;

	for (i=0;tags[i].tag;i++)
		tags[i].values = clist_custfree(tags[i].values, free);

	return i;
}
