/*
 * connman.c: Connection manager responsible for maintaining
 * remote connections. Also handles connection timeouts.
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "defs.h"
#include "sock.h"
#include "io.h"
#include "clist.h"
#include "connman.h"

/* connection time to live in seconds */
int conn_ttl = DEFAULT_CONN_TTL;


typedef struct {
	int sock;
	naddr_t netadd;		/* addr of this sockets other end */
	time_t ttl;		/* connection expiry time */
	int state;
} conn_info;

clist *conn_list = NULL;


/*
 * conn_set_ttl: Sets the time to live of a connection, in
 * seconds. If the connection is idle for the specified ttl
 * it is closed.
 */
int conn_set_ttl(int new_ttl)
{
	conn_ttl = new_ttl;
	return TRUE;
}

/*
 * conn_get_ttl: Returns the current ttl in use.
 */
int conn_get_ttl()
{
	return conn_ttl;
}

/*
 * conn_finder: Utility function used to specify the custom find
 * function to search an address in the connection list. Returns
 * zero if the item is found. Has two parameters:
 * void *a: Ptr to current connection list item data (conn_info*)
 * void *b: Ptr to network address being searched (naddr_t*)
 */
int conn_finder(void *a, void *b)
{
	conn_info *cur = (conn_info *)a;
	naddr_t *find = (naddr_t *)b;
	
	if (cur && find && cur->netadd.ip == find->ip
		       && cur->netadd.port == find->port)
		return 0;
	else
		return -1;
}

/*
 * conn_fd_finder: Same as above, only finds by the actual socket
 * rather then the network address. Has two parameters:
 * void *a: Ptr to current connection list item data (conn_info*)
 * void *b: Ptr to socket being searched (int *)
 */
int conn_fd_finder(void *a, void *b)
{
	conn_info *cur = (conn_info *)a;
	int *find = (int *)b;
	
	if (cur && find && cur->sock == *find)
		return 0;
	else
		return -1;
}

/*
 * conn_fd_open: Registers an already open fd with the connection
 * manager by associating it with a dummy address. Therefore this
 * socket may only be used for receiving data coz the other end's
 * address is not known. The socket could be assigned an address
 * for two way communication with conn_fd_authorize. Returns the
 * socket or -1 on error.
 */
int conn_fd_open(int sock)
{
	conn_info *nconn;

	nconn = (conn_info*)malloc(sizeof(conn_info));

	nconn->sock = sock;
	SET_UNKNOWN_NADDR(nconn->netadd);
	nconn->ttl = time(NULL)+conn_ttl;
	nconn->state = CONNSTATE_DATA_PENDING;
	conn_list = clist_prepend(conn_list, (void*)nconn);

	fd_open(sock, FDTYPE_MSG); /* also register with the I/O module */

	return sock;
}

/*
 * conn_open: Create a new connection and add it to the connection
 * list. The socket to the specified host is returned. Returns -1
 * on error. The connto object passed must have the correct network
 * byte order as no translation is performed.
 */
int conn_open(naddr_t *connto)
{
	int sock;
	struct sockaddr_in saddr;
	conn_info *nconn;

	saddr.sin_family = AF_INET;
	saddr.sin_port = connto->port;
	saddr.sin_addr.s_addr = connto->ip;

	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
		return -1;
	}

	if (connect(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
		close(sock);
		return -1;
	}
	
	nconn = (conn_info*)malloc(sizeof(conn_info));

	nconn->sock = sock;
	nconn->netadd = (*connto);
	nconn->ttl = time(NULL)+conn_ttl;
	nconn->state = CONNSTATE_NO_DATA;
	conn_list = clist_prepend(conn_list, (void*)nconn);

	fd_open(sock, FDTYPE_MSG); /* also register with the I/O module */

	return sock;
}

/*
 * conn_close_all: Close all the open connections and delete all
 * items in the connection list. Used to deinitialize the connection
 * manager.
 */
int conn_close_all()
{
	conn_info *cinfo;
	clist *cur = conn_list;

	while (cur) {
		cinfo = (conn_info*)cur->data;

		if (cinfo) {
			fd_close(cinfo->sock);
			free(cinfo);
		}

		cur = cur->next;
	}

	conn_list = clist_free(conn_list);

	return TRUE;
}

/*
 * conn_fd_close: Close an open connection by its file descriptor. The
 * descriptor is searched in the list of open connections and is closed
 * if found. Returns true if the descriptor was successfully closed and
 * FALSE on error.
 */
int conn_fd_close(int sock)
{
	clist *fitem;
	conn_info *cinfo;

	fitem = clist_find_custom(conn_list, &sock, conn_fd_finder);

	if (fitem == NULL)
		return FALSE;

	fd_close(sock);
	cinfo = (conn_info*)fitem->data;
	if (cinfo) free(cinfo);
	conn_list = clist_remove_link(conn_list, fitem);

	return TRUE;
}

/*
 * conn_fd_unreg: Unregisters a connection without closing it. This
 * is used if the specified socket need not be maintained by the
 * connection manager and would be handled externally. Returns
 * TRUE on success and FALSE on failure.
 */
int conn_fd_unreg(int sock)
{
	clist *fitem;
	conn_info *cinfo;

	fitem = clist_find_custom(conn_list, &sock, conn_fd_finder);

	if (fitem == NULL)
		return FALSE;

	cinfo = (conn_info*)fitem->data;
	if (cinfo) free(cinfo);
	conn_list = clist_remove_link(conn_list, fitem);

	return TRUE;
}

/*
 * conn_close: closes a connection and removes it from the connection
 * list. returns true on successfully closing the connection and false 
 * on error. the find object passed must have the correct network
 * byte order as no translation is performed.
 */
int conn_close(naddr_t *find)
{
	clist *fitem;
	conn_info *cinfo;

	fitem = clist_find_custom(conn_list, find, conn_finder);

	if (fitem == NULL)
		return FALSE;

	cinfo = (conn_info*)fitem->data;
	if (cinfo) {
		fd_close(cinfo->sock);
		free(cinfo);
	}
	conn_list = clist_remove_link(conn_list, fitem);

	return TRUE;
}

/*
 * conn_get: Searches the connection list for the specified connection.
 * If it is found, returns the corresponding socket and otherwise returns
 * -1 on error. Everytime a connection is retreived, its ttl is reset
 * to the default value.
 */
int conn_get(naddr_t *find)
{
	int sock;
	clist *fitem;
	conn_info *cinfo;

	fitem = clist_find_custom(conn_list, find, conn_finder);

	if (fitem == NULL)
		return -1;

	cinfo = (conn_info*)fitem->data;
	cinfo->ttl = time(NULL)+conn_ttl;
	return cinfo->sock;
}

/*
 * conn_authorize: Authorizes a connection so that two way
 * communication may take place on it. Initially a socket can only
 * receieve data coz the other end's listening port is not known. The
 * address of the other end's listening port is specified when the
 * first msg header from the socket is read (the sender field). At
 * that point the connection is authorized and the socket is
 * associated with a valid IP:PORT combo.
 */
int conn_fd_authorize(int sock, naddr_t *authaddr)
{
	clist *fitem;
	conn_info *cinfo;

	fitem = clist_find_custom(conn_list, &sock, conn_fd_finder);

	if (fitem == NULL)
		return FALSE;

	cinfo = (conn_info*)fitem->data;
	cinfo->netadd = (*authaddr);

	fd_authorize(sock);

	return TRUE;
}

/*
 * conn_fd_set_state: Set the state of a connection by its fd.
 * Returns TRUE on success and FALSE on failer.
 */
int conn_fd_set_state(int sock, int state)
{
	clist *fitem;
	conn_info *cinfo;

	fitem = clist_find_custom(conn_list, &sock, conn_fd_finder);

	if (fitem == NULL)
		return FALSE;

	cinfo = (conn_info*)fitem->data;
	cinfo->state = state;

	return TRUE;
}

/*
 * conn_fd_get_state: Returns the state of the specified fd, or
 * returns -1 on failure - if the connection wasnt found.
 */
int conn_fd_get_state(int sock)
{
	clist *fitem;
	conn_info *cinfo;

	fitem = clist_find_custom(conn_list, &sock, conn_fd_finder);

	if (fitem == NULL)
		return -1;

	cinfo = (conn_info*)fitem->data;

	return cinfo->state;
}

/*
 * conn_reset_ttl: Resets the ttl of the specified file descriptor to
 * the initial value. The ttl is reset implicitly by the conn_get
 * function too (for efficiency). Returns TRUE on a successful reset
 * and FALSE on failure.
 */
int conn_reset_ttl(int sock)
{
	clist *fitem;
	conn_info *cinfo;

	fitem = clist_find_custom(conn_list, &sock, conn_fd_finder);

	if (fitem == NULL)
		return FALSE;

	cinfo = (conn_info*)fitem->data;
	cinfo->ttl = time(NULL)+conn_ttl;

	return TRUE;
}

/*
 * conn_dec_ttl: Must be called atleast every second. This function
 * decrements the time to live of each connection by one and closes
 * the connection when it reaches zero. Returns the number of
 * connections closed.
 */
int conn_dec_ttl()
{
	conn_info *cinfo;
	clist *cur = conn_list;
	int nclosed = 0;
	time_t ctime = time(NULL);

	while (cur) {
		cinfo = (conn_info*)cur->data;

		if (ctime>=cinfo->ttl && cinfo->state==CONNSTATE_NO_DATA) {
			fd_close(cinfo->sock);
			free(cinfo);
			conn_list = clist_remove_link(conn_list, cur);
			nclosed++;
#if 0
			fprintf(stderr, "TIMEDOUT connection\n");
#endif
		}

		cur = cur->next;
	}

	return nclosed;
}
