/*
 * ctimer.c: handles the list of functions to be called at various 
 * intervals, no. of times to repeat and their arguments and related
 * functionality. ctimer_list must be set to null.
 * (C) Copyright 2003, netcluster
 *
 * Written by Syed Ghaus Ali
 * Fixed bug in ctimer_remove list head reassignment - Fakhir Ansari
 * Fixed memory leak in ctimer_remove_all - Fakhir Ansari
 * Rewritten - Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdlib.h>

#include "defs.h"
#include "ctimer.h"
#include "clist.h" 

/* the timer list */
clist *ctimer_list = NULL;


/* 
 * ctimer_cmp: Utility function for clist comparison.
 * Returns 1 if a>b, -1 if a<b and 0 if a==b.
 */
int ctimer_cmp(void *a, void *b)
{
	ctimerinf_t *x = (ctimerinf_t*)a;
	ctimerinf_t *y = (ctimerinf_t*)b;

	if (x->id > y->id)
		return 1;
	else if (x->id < y->id)
		return -1;
	else
		return 0;
}

/*
 * ctimer_add: Adds a new timer and schedules it for running
 * after 'delay' seconds. The timer would be deleted after
 * 'repeat' repetitions or will continue to run forever if it
 * is zero. The user_data is an arbitrary argument to the timer
 * function and must be malloc'ed by the user. The timer module
 * is responsible for free'ing this data.
 * Returns the id of the new timer.
 */
int ctimer_add(int delay, int repeat, CTIMER_FUNC_PTR(tfunc), void *user_data)
{
	ctimerinf_t *tinf;
	clist *list = ctimer_list;
	int cur_id = 0;

	while (list) {
		tinf = (ctimerinf_t*)list->data;
		if (tinf->id != cur_id)
			break;
		list = list->next;
		cur_id++;
	}

	tinf = (ctimerinf_t*) malloc(sizeof(ctimerinf_t));
	tinf->id = cur_id;
	tinf->delay = delay;
	tinf->repeat = repeat;
	tinf->last_called = time(NULL);
	tinf->user_data = user_data;
	tinf->tfunc = tfunc;

	ctimer_list = clist_insert_sorted(ctimer_list, tinf, ctimer_cmp);

	return cur_id;
}

/* 
 * ctimer_remove: This will remove the given id form the list of
 * active timers. Returns TRUE on success and FALSE otherwise.
 */
int ctimer_remove(int timer_id)
{
	ctimerinf_t tdummy;
	clist *find;

	tdummy.id = timer_id;
	find = clist_find_custom(ctimer_list, (void*)&tdummy, ctimer_cmp);

	if (find) {
		ctimerinf_t *tinf = (ctimerinf_t*)find->data;
		if (tinf) {
			if (tinf->user_data)
				free(tinf->user_data);
			free(tinf);
		}
		ctimer_list = clist_remove_link(ctimer_list, find);
		return TRUE;
	}

	return FALSE;
}

/* 
 * ctimer_remove_all: Frees all active timers.
 */
void ctimer_remove_all()
{
	ctimerinf_t *tinf;
	clist *list = ctimer_list;
	clist *cur;

	while (list) {
		cur = list;
		list = list->next;
		tinf = (ctimerinf_t*)cur->data;

		if (tinf) {
			if (tinf->user_data)
				free(tinf->user_data);
			free(tinf);
		}

		ctimer_list = clist_remove_link(ctimer_list, cur);
	}

	ctimer_list = NULL;
}

/* 
 * ctimer_run: This should be regularly called by the user because
 * it is responsible for running all the timers which have expired.
 * Returns the number of handlers called. 
 */
int ctimer_run()
{
	ctimerinf_t *tinf;
	clist *list = ctimer_list;
	clist *cur;
	int numcall = 0;

	while (list) {
		cur = list;
		list = list->next;
		tinf = (ctimerinf_t*)cur->data;

		if (tinf->last_called + tinf->delay > time(NULL))
			continue;

		if (tinf->tfunc)
		       (*tinf->tfunc)(tinf->user_data);

		tinf->last_called = time(NULL);

		if (tinf->repeat) {
			tinf->repeat--;

			if(tinf->repeat <= 0) {
				if (tinf) {
					if (tinf->user_data)
						free(tinf->user_data);
					free(tinf);
				}
				ctimer_list = clist_remove_link(ctimer_list, cur);
			}
		}

		numcall++;
	}

	return numcall;
}

/*
 * ctimer_reset: Resets the given timer to the new values. The old
 * values are overwritten. The old 'user_data' is also free'd
 * before setting the new one so make sure the new and old data
 * are separately allocated.
 * Returns TRUE on success and FALSE on failure if timer not found. 
 */
int ctimer_reset(int timer_id, int delay, int repeat,
		CTIMER_FUNC_PTR(tfunc), void *user_data)
{
	ctimerinf_t tdummy;
	clist *find;

	tdummy.id = timer_id;
	find = clist_find_custom(ctimer_list, (void*)&tdummy, ctimer_cmp);

	if (find) {
		ctimerinf_t *tinf = (ctimerinf_t*)find->data;

		tinf->delay = delay;
		tinf->repeat = repeat;
		tinf->user_data = user_data;
		tinf->tfunc = tfunc;

		return TRUE;
	}

	return FALSE;
}
