/*
 * diagnose.c: Various background jobs of diagnostics.
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "defs.h"
#include "structs.h"
#include "clist.h"
#include "ctimer.h"
#include "msgs.h"
#include "util.h"
#include "log.h"
#include "task.h"
#include "diagnose.h"


/* This contains a list of task_ids of masters which are
 * currently being polled for collective termination. */
clist *taskpoll_list = NULL;


/*
 * taskpoll_cancel: Cancels the collective terminate
 * operation for a master task. This function is called when
 * a MSG_NOTERMINATE is received in request of a collective
 * terminate. It removes the specified master from the poll
 * list and returns TRUE if an item from the list was removed
 * and FALSE if the item wasnt in the list already.
 */
int taskpoll_cancel(task_id *master)
{
	task_id *tid;
	clist *cur = taskpoll_list;

	while (cur) {
		tid = (task_id*)cur->data;

		if (!memcmp(tid, master, sizeof(task_id))) {
			free(cur->data);
			taskpoll_list = clist_remove_link(taskpoll_list,
					cur);
			return TRUE;
		}
	}

	return FALSE;
}

/*
 * taskpoll_sendall: Sends the specified msg to all
 * tasks in the taskpoll_list. The possible msgtypes are
 * either MSG_DOTERMINATE or MSG_ISTERMINATE. No other types
 * should be used here. This is a utility function.
 */
int taskpoll_sendall(u_int16_t msg_type)
{
	msghdr_t hdr;
	taskpoll_t data;
	clist *cur = taskpoll_list;

	hdr.sender = dconf.self_id;
	hdr.msg_type = msg_type;
	hdr.priority = 1;
	hdr.data_size = sizeof(data);
	hdr.trans_id = TRANSID_NONE;

	data.num = 1;
	data.daemons[0] = dconf.self_id;
	hton_taskpoll(&data);

	/* Run this loop for each master in the poll list */
	while (cur) {
		childinfo_t *cur_cld;
		childinfo_t *prev_cld;
		task_id *master = (task_id*)cur->data;

		data.master = *master;

		cur_cld = task_remote_child(master);

		/* Run this loop for each child of the
		 * master in the poll list. */
		while (cur_cld) {
			hdr.receiver = cur_cld->child.cdaemon;
			msg_write(&hdr, &data);

			prev_cld = cur_cld;
			cur_cld = cur_cld->next;
			free(prev_cld);
		}

		cur = cur->next;
	}

	return TRUE;
}

/*
 * taskpoll_run: This is a timed function called after
 * every TASKPOLL_INTERVAL seconds. Each time its called it
 * sends a MSG_DOTERMINATE for all the tasks in taskpoll_list
 * and then empties the list. It then adds all tasks with
 * masters running on the local machine to this list and
 * then sends out MSG_ISTERMINATE. Receiving MSG_NOTERMINATE
 * removes the task from taskpoll_list and thus aborts its
 * termination.
 */
CTIMER_FUNC(taskpoll_run)
{
	taskinfo_t *tsk;
	task_id *master;
	clist *cur;

	taskpoll_sendall(MSG_DOTERMINATE);

	/* Now remove all the items from the poll
	 * list to prepare for the next iteration. */

	cur = taskpoll_list;

	while (cur) {
		master = cur->data;
		task_remove_terminated(master);
		
		if (cur->data)
			free(cur->data);
		cur = cur->next;
	}

	taskpoll_list = clist_free(taskpoll_list);

	/* Again add all local master tasks running on
	 * this machine to the poll list. And then send
	 * each of theirs children MSG_ISTERMINATE. */

	tsk = task_list;

	while (tsk) {
		/* Add to list if current task is master with
		 * no un-terminated children on this machine. */

		if (!memcmp(&tsk->id, &tsk->master, sizeof(task_id)) &&
				task_all_terminated(&tsk->master)) {

			task_id *newid = malloc(sizeof(task_id));
			*newid = tsk->id;
			taskpoll_list = clist_prepend(taskpoll_list, newid);
		}

		tsk = tsk->next;
	}

	taskpoll_sendall(MSG_ISTERMINATE);
}

/*
 * sync_tasklist: Synchronizes the task list by checking whether
 * the tasks registered as running are still running. It detects
 * tasks which terminate without informing the daemon. This
 * might happen incase of an illegal operation performed by the
 * task or if it uses some non-conventional method for exit.
 */
CTIMER_FUNC(sync_tasklist)
{
	int status;
	taskinfo_t *tsk = task_list;

	while (tsk) {
		if (tsk->state == TSTATE_EXECUTE ||
				tsk->state == TSTATE_WAIT ||
				tsk->state == TSTATE_SUSPEND) {

			if (waitpid(ntohl(tsk->id.pid), &status,
						WNOHANG|WUNTRACED) > 0) {

				do_log(LOG_WARNING, "sync_tasklist: Abnormal \
					       termination of task. \
					       Daemon not informed");
				task_setstate_terminate(tsk, status);
			}
		}

		tsk = tsk->next;
	}

	if (task_calc_load())
		task_update_load();
}
