/*
 * dmain.c: Main control of the daemon
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#include "defs.h"
#include "structs.h"
#include "clist.h"
#include "ctimer.h"
#include "sock.h"
#include "msgs.h"
#include "io.h"
#include "transact.h"
#include "connman.h"
#include "util.h"
#include "log.h"
#include "conf.h"
#include "fileman.h"
#include "diagnose.h"

#include "task.h"

int ncdaemon_vas_unregister();


/* Places to search for daemon config files */
char *dconf_paths[] = {
	"ncdaemon.conf",
	"../ncdaemon.conf",
	"/etc/ncdaemon.conf",
	"~/ncdaemon.conf"
};

/* Tag specification for daemon config. To rename a tag, modify
 * it in this struct and also in the ncdaemon_readconf function
 * where it is interpretted. */
taginfo_t dconf_tags[] = {
	{ "VAS_SERVER", TAGCSET_ALL, TAGTYPE_NADDR, 1, NULL },
	{ "SELF_ADDR", TAGCSET_ALL, TAGTYPE_NADDR, 0, NULL },
	{ "LISTEN_PORT", TAGCSET_INT, TAGTYPE_INT, 0, NULL },
	{ "CONN_TIMEOUT", TAGCSET_INT, TAGTYPE_INT, 0, NULL },
	{ "BINARY_DIR", TAGCSET_ALL, TAGTYPE_QOTDSTR, 0, NULL },
	{ "LOGGING_FILE", TAGCSET_ALL, TAGTYPE_QOTDSTR, 0, NULL },
	{ "LOG_EMERG", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_ALERT", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_CRIT", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_ERR", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_WARNING", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_NOTICE", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_INFO", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_DEBUG", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ NULL, NULL, 0, 0, NULL }
};

/* Daemon config struct */
dconf_t dconf;

/*
 * ncdaemon_readconf: Reads the config files if found and inits
 * the dconf data to sane values. Returns FALSE on error.
 */
int ncdaemon_readconf()
{
	int i;
	clist *itm;
	int *iptr;
	naddr_t *nptr;

	/* Try all config files one by one */
	for (i=0;i<sizeof(dconf_paths)/sizeof(char*);i++) {
		if (!access(dconf_paths[i], R_OK)) {
			if (!conf_read(dconf_paths[i], dconf_tags))
				return FALSE;
			break;
		}
	}

	if (i==sizeof(dconf_paths)/sizeof(char*))
		do_log(LOG_NOTICE,"Initialization: no config file found");

	dconf.log_types = 0;

	/* Store the values read from the config file into dconf */
	for (i=0;dconf_tags[i].tag;i++) {
		itm = dconf_tags[i].values;

		if (!strcmp(dconf_tags[i].tag, "VAS_SERVER")) {
			if (itm) {
				dconf.vas_list = itm;
				dconf_tags[i].values = NULL;
			}
			else {
				yell("Atleast one VAS server must be specifed "
						"in the config file");
				return FALSE;
			}
		}
		if (!strcmp(dconf_tags[i].tag, "SELF_ADDR")) {
			if (itm && itm->data) {
				dconf.self_id = *((naddr_t*)itm->data);
			}
			else if (!make_self_daemon_id(&(dconf.self_id),
							dconf.listen_port)) {
				do_log(LOG_ERR, "Unable to get local hostname. Specify"
						"SELF_ADDR in the config file.");
				return FALSE;
			}
		}
		else if (!strcmp(dconf_tags[i].tag, "LISTEN_PORT")) {
			if (itm && itm->data)
				dconf.listen_port = *((int*)itm->data);
			else
				dconf.listen_port = DPORT;
		}
		else if (!strcmp(dconf_tags[i].tag, "CONN_TIMEOUT")) {
			if (itm && itm->data)
				dconf.conn_ttl = *((int*)itm->data);
			else
				dconf.conn_ttl = DEFAULT_CONN_TTL;
		}
		else if (!strcmp(dconf_tags[i].tag, "BINARY_DIR")) {
			if (itm && itm->data)
				dconf.binary_dir = strdup((char*)itm->data);
			else
				dconf.binary_dir = strdup(DEFAULT_BINARY_DIR);
		}
		else if (!strcmp(dconf_tags[i].tag, "LOGGING_FILE")) {
			if (itm && itm->data)
				dconf.log_file = strdup((char*)itm->data);
			else
				dconf.log_file = NULL;
		}
		else if (!strcmp(dconf_tags[i].tag, "LOG_EMERG")) {
			if (itm && itm->data)
				dconf.log_types |= LOG_EMERG;
		}
		else if (!strcmp(dconf_tags[i].tag, "LOG_ALERT")) {
			if (itm && itm->data)
				dconf.log_types |= LOG_ALERT;
		}
		else if (!strcmp(dconf_tags[i].tag, "LOG_CRIT")) {
			if (itm && itm->data)
				dconf.log_types |= LOG_CRIT;
		}
		else if (!strcmp(dconf_tags[i].tag, "LOG_ERR")) {
			if (itm && itm->data)
				dconf.log_types |= LOG_ERR;
		}
		else if (!strcmp(dconf_tags[i].tag, "LOG_WARNING")) {
			if (itm && itm->data)
				dconf.log_types |= LOG_WARNING;
		}
		else if (!strcmp(dconf_tags[i].tag, "LOG_NOTICE")) {
			if (itm && itm->data)
				dconf.log_types |= LOG_NOTICE;
		}
		else if (!strcmp(dconf_tags[i].tag, "LOG_INFO")) {
			if (itm && itm->data)
				dconf.log_types |= LOG_INFO;
		}
		else if (!strcmp(dconf_tags[i].tag, "LOG_DEBUG")) {
			if (itm && itm->data)
				dconf.log_types |= LOG_DEBUG;
		}

		/* Free the data read from the config file. Set 'values'
		 * to NULL if it isnt to be free'd as in VAS_SERVER. */
		if (dconf_tags[i].values) {
			dconf_tags[i].values = clist_custfree(
					dconf_tags[i].values, free);
		}
	}

	/* Listen port takes precedence over specified port */
	dconf.self_id.port = htons(dconf.listen_port);

	SET_UNKNOWN_NADDR(dconf.vas_cur);
	dconf.homedir = get_homedir();

	return TRUE;
}

/* terminate_handler: Termination handler for SIGINT, SIGHUP, SIGTERM */
void terminate_handler(int signum)
{
	/* Close all connections */
	fd_close_all();

	/* Close the log file */
	if (log_file)
		fclose(log_file);

	exit(signum);
}

/* ncdaemon_init: Performs all the initialization of the daemon */
int ncdaemon_init()
{
	FILE *lf;
	struct sigaction new_action;

	/* Set up the signal handler */
	new_action.sa_handler = terminate_handler;
	sigemptyset(&new_action.sa_mask);
	new_action.sa_flags = 0;
	sigaction(SIGINT, &new_action, NULL);
	sigaction(SIGHUP, &new_action, NULL);
	sigaction(SIGTERM, &new_action, NULL);

	/* Initialize the open fd to zero */
	FD_ZERO(&fd_ractive);
	FD_ZERO(&fd_wactive);

	/* Read and parse the config file */
	if (!ncdaemon_readconf()) {
		fprintf(stderr, "Initialization: failed\n");
		return FALSE;
	}

	dconf.load.numprocs = 0;

	set_log_types(dconf.log_types);
	set_module_name("netcluster");
	conn_set_ttl(dconf.conn_ttl);

#ifndef DEBUG
	if (dconf.log_file) {
		set_log_file(dconf.log_file);
	}
#endif

	if (!ftp_set_dl_path(dconf.binary_dir))
		yell("Initialization: Unable to set binary path to <%s>",
				dconf.binary_dir);

	/* Register functions to be called by ctimer */
	ctimer_add(1, 0, trans_run, NULL);
	ctimer_add(1, 0, ftp_run, NULL);
	ctimer_add(TASKPOLL_INTERVAL, 0, taskpoll_run, NULL);
	ctimer_add(SYNC_TASKLIST_INTERVAL, 0, sync_tasklist, NULL);

	do_log(LOG_INFO,"Initialization: Success");

	return TRUE;
}

/* ncdaemon_deinit: Performs the deinitialization of the daemon before exit */
int ncdaemon_deinit()
{
	/* Unregister from VAS */
	ncdaemon_vas_unregister();

	/* Close all connections */
	fd_close_all();

	/* Delete all timers */
	ctimer_remove_all();

	if (log_file)
		fclose(log_file);

	return TRUE;
}

/*
 * ncdaemon_idle: This function is continuously called in the idle loop.
 * It performs background tasks such as timing out connections and
 * running timers.
 */
void ncdaemon_idle()
{
	static time_t lasttime;
	
	if (lasttime != time(NULL)) {
		lasttime = time(NULL);

		/* Timeout the expired connections */
		conn_dec_ttl();

		/* Call the timer handler */
		ctimer_run();
	}
}

/*
 * ncdaemon_vas_register: Checks if the current vas in dconf is still
 * valid and if not it goes thru the vas_list and finds a running VAS
 * server.
 * A VAS is not marked as connected until it returns a
 * MSG_VAS_REG_RPLY so this function doesnt ensure that a
 * valid VAS has been found.
 * This is also a transaction failure handler incase the msg delivery
 * fails, it calls itself again.
 * It would keep sending registry msgs and going over and over the
 * vas_list if the msgs are successfully being sent to atleast one of
 * the vas servers. Otherwise it just fails the first time and is
 * never called again. The daemon in such a case runs on its own.
 */
TRANSFAIL_HANDLER(ncdaemon_vas_register)
{
	static clist *cur = NULL;
	msghdr_t reghdr;
	vasload_t regload;
	
	if (!IS_UNKNOWN_NADDR(dconf.vas_cur))
		return;

	if (cur==NULL)
		cur = dconf.vas_list;

	reghdr.sender = dconf.self_id;
	reghdr.msg_type = MSG_VAS_REG;
	reghdr.priority = 1;
	reghdr.data_size = sizeof(vasload_t);
	reghdr.trans_id = TRANSID_NONE;

	regload = dconf.load;

	while (cur) {
		naddr_t *na = (naddr_t*)cur->data;
		reghdr.receiver = *na;
		cur = cur->next;

		hton_vasload(&regload);
		if (trans_write(&reghdr, &regload, ncdaemon_vas_register,
					NULL, VAS_REGISTER_TIMEOUT)) {
			break;
		}
	}

	return;
}

/*
 * ncdaemon_vas_unregister: This function is called before the
 * daemon exits. It unregisters from the VAS with which the
 * daemon is registered (if any). Returns TRUE on successfully
 * sending the unregister msg to the VAS.
 */
int ncdaemon_vas_unregister()
{
	msghdr_t ureghdr;

#ifdef DEBUG
	ureghdr.sender = dconf.self_id;
	ureghdr.receiver = dconf.vas_cur;
	ureghdr.msg_type = MSG_VAS_DISPLAY;
	ureghdr.priority = 1;
	ureghdr.data_size = 0;
	ureghdr.trans_id = TRANSID_NONE;
	msg_write(&ureghdr, NULL);
#endif

	if (! IS_UNKNOWN_NADDR(dconf.vas_cur)) {
		ureghdr.sender = dconf.self_id;
		ureghdr.receiver = dconf.vas_cur;
		ureghdr.msg_type = MSG_VAS_UNREG;
		ureghdr.priority = 1;
		ureghdr.data_size = 0;
		ureghdr.trans_id = TRANSID_NONE;

		msg_write(&ureghdr, NULL);
	}

	return FALSE;
}

#ifdef DEBUG
/* XXX: Test function. Called whenever user presses enter. */
int ncdaemon_mytest()
{
	char buf[128];
	int len;
	daemon_id odem;

	return FALSE;
	fgets(buf, sizeof(buf), stdin);
	len = strlen(buf);
	if (len && buf[len-1]=='\n')
		buf[len-1] = '\0';
	if (!(*buf))
		return FALSE;

	odem = dconf.self_id;
	odem.port = htons(7777);
	ftp_send(&odem, "/home/darkstar/work/ncdebug/mytest.dat", NULL, NULL);

	return TRUE;
}
#endif /* DEBUG */

/*
 * The main control of the daemon subsystem. Its tasks include listening
 * for incoming connections and data on those connections. It calls the
 * communication layer for reading the messages and then dispatches them.
 */
int ncdaemon_main()
{
	int i;
	struct timeval tv;
	fd_set fd_read;
	struct sockaddr_in clientaddr;
	int sock;
	int nmsgs;
	int size;
	int new;

	if ((sock = make_listen_socket(dconf.listen_port)) < 0) {
		fprintf(stderr, "Unable to listen on port %d. Aborting...\n",
				dconf.listen_port);
		return FALSE;
	}
	
	/* Dont register as connection coz its only for listening */
	fd_open(sock, FDTYPE_LISTEN);

#ifdef DEBUG
	fd_open(0, FDTYPE_LOCALDATA);
#endif

	/* Send a registry msg to the VAS and wait for a reply */
	ncdaemon_vas_register(NULL, NULL);
	
	while (1) {
		ncdaemon_idle();

		fd_has_closed = FALSE;

		fd_read = fd_ractive;
		tv.tv_sec = 0;
		tv.tv_usec = 500;

		if (!select(global_max_fd+1, &fd_read, NULL, NULL, &tv))
			continue;

		for (i=global_max_fd;i>=0;i--) {
			if (!FD_ISSET(i, &fd_read))
				continue;
			else if (fd_has_closed)
				break;

			switch (FDTYPE(i)) {
				case FDTYPE_LISTEN:
					/* new connection */
					if (i != sock)
						continue;

					size = sizeof(clientaddr);
					new = accept(sock, &clientaddr, &size);
					if (new < 0) {
						yell("Unable to accept:%s", strerror(errno));
						continue;
					}

					conn_fd_open(new);

					do_log(LOG_NOTICE,"Connect from addr<%s:%d> sock<%d>",
							inet_ntoa(clientaddr.sin_addr),
							ntohs(clientaddr.sin_port), new);
					break;

				case FDTYPE_MSG:
					/* data on one of connected descriptors */

					/* if nmsgs < 0 then either an error
					 * has occured or the connection has
					 * been closed. if zero then a partial
					 * msg has been read and is buffered
					 * otherwise number of msg read is
					 * returned */
					nmsgs = msg_read(i);
					conn_reset_ttl(i);

					if (nmsgs < 0) {
						conn_fd_close(i);
						do_log(LOG_NOTICE,"Session closed: sock<%d>",i);

					}
					else if (nmsgs > 0)
						msg_dispatch();
					break;

				case FDTYPE_LOCALMSG:
					/* data on a local task's descriptor */
					nmsgs = msg_lread(i);

					if (nmsgs < 0) {
						fd_close(i);
						do_log(LOG_NOTICE,"Local session closed: sock<%d>",i);

					}
					else if (nmsgs > 0)
						msg_dispatch();
					break;
					
				case FDTYPE_LOCALDATA:
#ifdef DEBUG
					if (i==0) {
						if (!ncdaemon_mytest())
							return TRUE;
						break;
					}
#endif
					/* data from one of the standard descriptors
					 * of a task. Redirect it to parent. */
					redir_stdfd(i);
					break;

				default:
					do_log(LOG_ERR, "Unknown file des: fd<%d> type<%d>",
							i, FDTYPE(i));
			};
		}
	}
	
	return TRUE;
}

/* main: Program entry point */
int main()
{
	pid_t child_pid;

	if (!ncdaemon_init()) {
		fprintf(stderr, "netcluster: Initialization failed. Aborting...\n");
		exit(EXIT_FAILURE);
	}

#ifndef DEBUG
	child_pid = fork();

	if (child_pid == -1) {
		fprintf(stderr, "netcluster: Cannot startup daemon. Error forking child process!\n");
		exit(EXIT_FAILURE);
	}
	else if (child_pid != 0) {
		fprintf(stderr, "netcluster: Started up successfully\n");
		fprintf(stderr, "pid=%d\n", child_pid);
		exit(EXIT_SUCCESS);
	}

	close(0);
	close(1);
	setsid();
#endif /* DEBUG */

	ncdaemon_main();
	ncdaemon_deinit();

	return 0;
}
