/*
 * fileman.c: File transfer
 * File transfer takes place in blocks each time the run 
 * function is called.
 * (C) Copyright 2003, netcluster
 *
 * Written by Syed Ghaus Ali
 * Modified, Restructured - Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include "defs.h"
#include "clist.h"
#include "util.h"
#include "msgs.h"
#include "log.h"
#include "fileman.h"
#include "md5.h"


/* list of ftpinfo_t for files being sent */
clist *ftp_sends = NULL;
/* list of ftpinfo_t for files being received */
clist *ftp_recvs = NULL;

/* Path for file downloads */
char *ftp_dl_path = NULL;

/*
 * ftp_item_free: A utility function which frees all data
 * associated with an ftpinfo_t struct.
 */
void ftp_item_free(ftpinfo_t *fitem)
{
	if (fitem) { 
		if (fitem->filename) free(fitem->filename);
		if (fitem->user_data) free(fitem->user_data);
		if (fitem->fd >= 0) close(fitem->fd);
		free(fitem);
	}
}

/* 
 * ftp_item_cmp: Compares two ftpinfo_t structs to check
 * if their ftp_ids match. If the daemon_id's dont match the
 * ftpind are compared. Returns values:
 *	1 if a > b
 *	0 if a == b
 *	-1 if a < b
 */
int ftp_item_cmp(void *a, void *b)
{
	ftpinfo_t *x=a;
	ftpinfo_t *y=b;
	
	/* if daemon_ids dont match */
	if (memcmp(&x->ftp_id.dem_id, &y->ftp_id.dem_id, sizeof(daemon_id))) {
		if (x->ftp_id.ftpind > y->ftp_id.ftpind)
			return 1;
		else if (x->ftp_id.ftpind < y->ftp_id.ftpind)
			return -1;
	}

	return 0;
}

/*
 * ftp_default_recv_tout: The default handler for handling FTP
 * receive timeouts.
 */
FTP_HND_EVENT(ftp_default_recv_event)
{
	do_log(LOG_INFO, "FTP recv event: Default handler");
}

/*
 * ftp_default_send_tout: The default handler for handling FTP
 * send timeouts.
 */
FTP_HND_EVENT(ftp_default_send_event)
{
	do_log(LOG_INFO, "FTP send event: Default handler");
}

/*
 * ftp_set_dl_path: Sets the download path. Duplicates the string
 * passed by the user and frees any previous path before setting
 * the new one. Returns TRUE on success of FALSE if no write
 * access to the specified directory.
 * The path would always stored with having a trailing '/' and
 * tilde (~) would be expanded.
 */
int ftp_set_dl_path(char *path)
{
	struct stat statinf;
	int len;
	char *str;
	
	str = expand_tilde(path);

	if (!str || !(*str) || access(str, W_OK|R_OK)) {
		if (str) free(str);
		return FALSE;
	}
	if (stat(str, &statinf) < 0) {
		if (str) free(str);
		return FALSE;
	}
	if (! S_ISDIR(statinf.st_mode)) {
		if (str) free(str);
		return FALSE;
	}

	if (ftp_dl_path)
		free(ftp_dl_path);

	ftp_dl_path = str;
	len = strlen(ftp_dl_path);

	if (ftp_dl_path[len-1] != DIR_SEPARATOR) {
		ftp_dl_path = realloc(ftp_dl_path, len+2);
		ftp_dl_path[len] = DIR_SEPARATOR;
		ftp_dl_path[len+1] = '\0';
	}

	return TRUE;
}

/*
 * ftp_get_send: Returns the ftpinfo_t struct for the given ftp_id.
 * If the given id is not found, it returns NULL.
 */
ftpinfo_t *ftp_get_send(ftpid_t *fid)
{
	ftpinfo_t dummy;
	clist *find;

	dummy.ftp_id = *fid;
	find = clist_find_custom(ftp_sends, &dummy, ftp_item_cmp);
	return (ftpinfo_t*)find->data;
}

/*
 * ftp_get_recv: Returns the ftpinfo_t struct for the given ftp_id.
 * If the given id is not found, it returns NULL.
 */
ftpinfo_t *ftp_get_recv(ftpid_t *fid)
{
	ftpinfo_t dummy;
	clist *find;

	dummy.ftp_id = *fid;
	find = clist_find_custom(ftp_recvs, &dummy, ftp_item_cmp);
	return (ftpinfo_t*)find->data;
}

/*
 * ftp_remove_send: Removes the file send operation.
 */
int ftp_remove_send(ftpid_t *fid)
{
	ftpinfo_t dummy;
	clist *find;

	dummy.ftp_id = *fid;
	find = clist_find_custom(ftp_sends, &dummy, ftp_item_cmp);

	if (find) {
		ftp_item_free(find->data);
		ftp_sends = clist_remove_link(ftp_sends, find);
		return TRUE;
	}

	return FALSE;
}

/*
 * ftp_remove_recv: Removes the file receive operation.
 */
int ftp_remove_recv(ftpid_t *fid)
{
	ftpinfo_t dummy;
	clist *find;

	dummy.ftp_id = *fid;
	find = clist_find_custom(ftp_recvs, &dummy, ftp_item_cmp);

	if (find) {
		ftp_item_free(find->data);
		ftp_recvs = clist_remove_link(ftp_recvs, find);
		return TRUE;
	}

	return FALSE;
}

/*
 * ftp_cancel_send: Cancels the file send operation.
 */
int ftp_cancel_send(ftpid_t *fid)
{
	ftpinfo_t dummy;
	clist *find;

	dummy.ftp_id = *fid;
	find = clist_find_custom(ftp_sends, &dummy, ftp_item_cmp);

	if (find) {
		ftpinfo_t *finf = (ftpinfo_t*)find->data;
		finf->status = FTPSTAT_CANCEL;
		if (finf->hnd_event)
			(*finf->hnd_event)(finf, finf->user_data);
		ftp_item_free(find->data);
		ftp_sends = clist_remove_link(ftp_sends, find);
		return TRUE;
	}

	return FALSE;
}

/*
 * ftp_cancel_recv: Cancels the file receive operation.
 */
int ftp_cancel_recv(ftpid_t *fid)
{
	ftpinfo_t dummy;
	clist *find;

	dummy.ftp_id = *fid;
	find = clist_find_custom(ftp_recvs, &dummy, ftp_item_cmp);

	if (find) {
		ftpinfo_t *finf = (ftpinfo_t*)find->data;
		finf->status = FTPSTAT_CANCEL;
		if (finf->hnd_event)
			(*finf->hnd_event)(finf, finf->user_data);
		ftp_item_free(find->data);
		ftp_recvs = clist_remove_link(ftp_recvs, find);
		return TRUE;
	}

	return FALSE;
}

/*
 * ftp_send_block: Sends a single block of file to the receiver.
 * Returns the number of bytes left to be transferred. 'fd' is the
 * file descriptor to which a reply could possibly be sent (the
 * 'fd' of MSG_HANDLERs) or -1 for creating a new connection. This
 * is called whenever a MSG_FTP_REQ is received.
 * Returns -1 on error.
 */
int ftp_send_block(ftpid_t *fid, int fd)
{
	int nbytes;
	ftpinfo_t *ftpinf;
	msghdr_t msghdr;
	char msgdata[sizeof(ftpid_t)+FTP_SEND_BLOCK_SIZE];
	ftpid_t *ftp_id = (ftpid_t*)msgdata;
	char *file_data = (char*)(msgdata+sizeof(ftpid_t));

	*ftp_id = *fid;

	/* retrieve the ftpinfo_t struct */
	ftpinf = ftp_get_send(ftp_id);
	if (!ftpinf)
		return -1;

	/* read bytes from the file being sent */
	nbytes = read(ftpinf->fd, file_data, FTP_SEND_BLOCK_SIZE);

	if (nbytes < 0)
		return -1;

	/* make the msghdr and data for the transfer */
	msghdr.sender = dconf.self_id;
	msghdr.receiver = ftp_id->dem_id;
	msghdr.msg_type = MSG_FTP_DATA;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(ftpid_t)+nbytes;
	msghdr.trans_id = TRANSID_NONE;

	hton_ftpid(ftp_id);

	if (!msg_reply(&msghdr, msgdata, fd)) {
		free(msgdata);
		return -1;
	}

	ftpinf->cur_size += nbytes;
	ftpinf->blks_trans += 1;
	ftpinf->last = time(NULL);

	return (ftpinf->size - ftpinf->cur_size);
}

/*
 * ftp_recv_block: Recvs a single block of file from the sender.
 * Returns the number of bytes left to be transferred. This is
 * called whenever a MSG_FTP_DATA is received.
 * Returns -1 on error, 0 if file transfer complete or otherwise
 * the number of bytes left to be transfered.
 */
int ftp_recv_block(ftpid_t *fid, void *data, unsigned long size)
{
	msghdr_t msghdr;
	ftpinfo_t *ftpinf;
	int nbytes;

	ftpinf = ftp_get_recv(fid);

	if (!ftpinf || ftpinf->cur_size+size > ftpinf->size) {
		yell("FTP recv_block: Invalid operation or operation not found");
		return -1;
	}

	nbytes = write(ftpinf->fd, data, size);
	if (nbytes != size) {
		yell("FTP recv_block: Unable to write to file<%s>", ftpinf->filename);
		return -1;
	}

	ftpinf->cur_size += nbytes;
	ftpinf->blks_trans += 1;
	ftpinf->last = time(NULL);

	return (ftpinf->size - ftpinf->cur_size);
}

/*
 * ftp_send: This initiates a file send operation. It sends an FTP_INIT
 * message to the receiving daemon. If the message is delivered, it sets
 * up the receiving record in ftp_sends. Returns the ID of the operation.
 * The 'filename' is copied into a dynamically allocated buffer whereas
 * 'user_data' is stored as is. The user_data must be malloc'ed by the
 * called and the ftp module is responsible for freeing it. Filename
 * need not be malloced coz it is copied.
 */
ftpid_t *ftp_send(daemon_id *to, char *filename,
		FTP_HND_EVENT_PTR(fhnd), void *user_data)
{
	msghdr_t msghdr;
	char *msgdata;
	ftpinit_t *ftpinit;
	ftpinfo_t *ftpinf;
	int fd;
	struct stat statinf;
	clist *cur = ftp_sends;
	int cur_ind = 1;

	while (cur) {
		ftpinf = (ftpinfo_t*)cur->data;
		if (ftpinf->ftp_id.ftpind != cur_ind)
			break;
		cur_ind++;
		cur = cur->next;
	}
	
	if ((fd = open(filename, O_RDONLY)) < 0) {
		yell("FTP send: Failed. Unable to open file for reading.");
		return NULL;
	}

	if (fstat(fd, &statinf) < 0) {
		close(fd);
		yell("FTP send: Failed. Unable to retrieve file info.");
		return NULL;
	}
	
	ftpinf = (ftpinfo_t*)malloc(sizeof(ftpinfo_t));
	ftpinf->ftp_id.dem_id = dconf.self_id;
	ftpinf->remote_dem = *to;
	ftpinf->ftp_id.ftpind = cur_ind;
	ftpinf->status = FTPSTAT_PENDING;
	ftpinf->filename = strdup(filename);
	ftpinf->size = statinf.st_size;
	ftpinf->cur_size = 0;
	ftpinf->blks_trans = 0;
	ftpinf->last = time(NULL);
	ftpinf->hnd_event = fhnd ? fhnd : ftp_default_send_event;
	ftpinf->user_data = user_data;
	ftpinf->fd = fd;

	ftp_sends = clist_insert_sorted(ftp_sends, ftpinf, ftp_item_cmp);

	/* Send MSG_FTP_INIT to whom the file is being sent */
	msghdr.sender = dconf.self_id;
	msghdr.receiver = *to;
	msghdr.msg_type = MSG_FTP_INIT;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(ftpinit_t) + strlen(filename) + 1;
	msghdr.trans_id = TRANSID_NONE;

	msgdata = (char*)malloc(msghdr.data_size);
	ftpinit = (ftpinit_t*)msgdata;
	ftpinit->ftp_id = ftpinf->ftp_id;
	ftpinit->file_size = ftpinf->size;
	generate_md5(filename, ftpinit->md5);
	strcpy(msgdata+sizeof(ftpinit_t), filename);

	hton_ftpinit(ftpinit);
	msg_write(&msghdr, msgdata);
	free(msgdata);

	return &ftpinf->ftp_id;
}

/*
 * ftp_recv: Prepares the daemon for receiving a file over the network.
 * Initializes an ftpinfo_t struct and adds it to ftp_recvs. Only the
 * basename of the file is used (path is trimmed) and then the file
 * is written to ftp_dl_path directory. Returns 1 if receiving the
 * file, 0 if file already present or -1 on error.
 */
int ftp_recv(ftpid_t *from, char *filename, unsigned long size, char *md5sum,
		FTP_HND_EVENT_PTR(fhnd), void *user_data)
{
	ftpinfo_t *ftpinf;
	char *basename;
	char *fullname;
	int fd;

	basename = base_name(filename);
	fullname = malloc_strcat(ftp_dl_path ? ftp_dl_path : empty_string,
			basename);
	free(basename);

	if (!access(fullname, R_OK)) {
		char md5cur[MD5_DIGEST_SIZE];
		generate_md5(fullname, md5cur);
		if (!memcmp(md5cur, md5sum, MD5_DIGEST_SIZE))
			return 0;
	}

	if ((fd = open(fullname, O_WRONLY|O_CREAT|O_TRUNC, FTP_DEFAULT_PERM)) < 0) {
		yell("FTP recv: Failed. Unable to open file for write<%s>",
				fullname);
		free(fullname);
		return -1;
	}

	ftpinf = (ftpinfo_t*)malloc(sizeof(ftpinfo_t));
	ftpinf->ftp_id = *from;
	ftpinf->remote_dem = from->dem_id;
	ftpinf->status = FTPSTAT_PENDING;
	ftpinf->filename = fullname;
	ftpinf->size = size;
	ftpinf->cur_size = 0;
	ftpinf->blks_trans = 0;
	ftpinf->last = time(NULL);
	ftpinf->hnd_event = fhnd ? fhnd : ftp_default_recv_event;
	ftpinf->user_data = user_data;
	ftpinf->fd = fd;

	ftp_recvs = clist_insert_sorted(ftp_recvs, ftpinf, ftp_item_cmp);

	return 1;
}

/*
 * ftp_run: This is repeatedly called and should preferable be called
 * every second. It monitors and timeouts sends and recvs if there is
 * no activity for a specific time interval (FTP_TIMEOUT_INTERVAL).
 */
void ftp_run()
{
	ftpinfo_t *ftpinf;
	clist *cur;
	clist *list = ftp_sends;

	while (list) {
		cur = list;
		list = list->next;
		ftpinf = (ftpinfo_t*)cur->data;

		if (time(NULL) > ftpinf->last + FTP_TIMEOUT_INTERVAL) {
			ftpinf->status = FTPSTAT_TIMEOUT;
			if (ftpinf->hnd_event)
				(*ftpinf->hnd_event)(ftpinf, ftpinf->user_data);
			ftp_item_free(ftpinf);
			ftp_sends = clist_remove_link(ftp_sends, cur);
		}
	}

	list = ftp_recvs;

	while (list) {
		cur = list;
		list = list->next;
		ftpinf = (ftpinfo_t*)cur->data;

		if (time(NULL) > ftpinf->last + FTP_TIMEOUT_INTERVAL) {
			ftpinf->status = FTPSTAT_TIMEOUT;
			if (ftpinf->hnd_event)
				(*ftpinf->hnd_event)(ftpinf, ftpinf->user_data);
			ftp_item_free(ftpinf);
			ftp_recvs = clist_remove_link(ftp_recvs, cur);
		}
	}
}
