/*
 * io.c: Low level I/O routines for message passing
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 * Bugfix in msg_enqueue - Ghaus Ali
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "defs.h"
#include "clist.h"
#include "sock.h"
#include "io.h"
#include "msgs.h"
#include "connman.h"
#include "task.h"

/* Flag for a select exception */
int fd_has_closed;

/* Incoming message queue */
clist *in_msgs = NULL;

/* Default size allocated to a message */
#define DEFAULT_IOBUF_SIZE 256

/* MAX sockets to watch */
#define MAX_IOS FD_SETSIZE

typedef struct {
	char *buf;
	int buf_size;
	int read_pos;		/* always points to start of msg if any */
	int write_pos;		/* where next byte read to be written */
	int cvthdr;		/* need to convert header byte order? */
} io_recv;

io_recv **ios = NULL;
char *io_type;


/*
 * init_io: Initialize the global variables for I/O operations. No need
 * to call explicitly coz is called by the fd_open and fd_open_write.
 */
void init_io()
{
	int i;
	static first = 1;

	if (first) {
		first = 0;
		ios = (io_recv **)malloc(sizeof(io_recv *) * MAX_IOS);
		io_type = (char*) malloc(sizeof(char) * MAX_IOS);
		for (i=0;i<MAX_IOS;i++) {
			ios[i] = NULL;
			io_type[i] = FDTYPE_NONE | FDTYPE_NOAUTH;
		}
	}
}

/*
 * msg_read: A low level interface to message passing over the sockets.
 * Reads a message and stores it in the in_msgs queue.
 * incase a partial message is read it is buffered until the complete
 * message arrives.
 *
 * Returns -1 on error if something goes really wrong, 0 if a partial
 * message is read and otherwise the total messages read into queue.
 */
int msg_read(int fd)
{
	io_recv *ioptr;
	int nbytes;
	int nmsgs = 0;

	if (ios[fd] == NULL) {
		ios[fd] = (io_recv *)malloc(sizeof(io_recv));
		ios[fd]->buf = (char *)malloc(DEFAULT_IOBUF_SIZE);
		ios[fd]->buf_size = DEFAULT_IOBUF_SIZE;
		ios[fd]->read_pos = 0;
		ios[fd]->write_pos = 0;
		ios[fd]->cvthdr = 1;
	}

	ioptr = ios[fd];

	if (ioptr->read_pos) {
		int len = ioptr->write_pos - ioptr->read_pos;
		if (len)
			memcpy(ioptr->buf,  ioptr->buf + ioptr->read_pos, len);
		ioptr->read_pos = 0;
		ioptr->write_pos = len;
	}
	
	nbytes = read(fd, ioptr->buf + ioptr->write_pos,
			ioptr->buf_size - ioptr->write_pos);

	if (nbytes <= 0) {
#ifdef DEBUG
		if (nbytes < 0)
			perror("error in read");
#endif
		return -1;
	}
	ioptr->write_pos += nbytes;

	/* atleast a full header is present in the buffer */
	while (ioptr->write_pos - ioptr->read_pos >= sizeof(msghdr_t)) {
		int msg_size;
		msghdr_t *header = (msghdr_t *)&(ioptr->buf[ioptr->read_pos]);
		
		if (ioptr->cvthdr) /* Convert network to host byte order */
			ntoh_msg(header);
		msg_size = MSG_SIZE(header);

		/* authorize the socket as a full header has been read */
		if (!FDAUTH(fd))
			conn_fd_authorize(fd, &(header->sender));

		/* realloc buffer if not enough space in current allocation */
		if (msg_size > ioptr->buf_size) {
			ioptr->buf = realloc(ioptr->buf, msg_size);
			ioptr->buf_size = msg_size;
		}

		/* complete message(s) present in buffer */
		if (ioptr->write_pos - ioptr->read_pos >= msg_size) {
			char *msg = (char*)malloc(msg_size);

			memcpy(msg, ioptr->buf + ioptr->read_pos, msg_size);
			msg_enqueue((msghdr_t *)msg, fd, MSGCLASS_REMOTE);

			ioptr->read_pos += msg_size;
			ioptr->cvthdr = 1;
			nmsgs++;
		}
		else {
			ioptr->cvthdr = 0;
			break;
		}
	}

	/* Set the connection state */
	if (ioptr->write_pos == ioptr->read_pos)
		conn_fd_set_state(fd, CONNSTATE_NO_DATA);
	else
		conn_fd_set_state(fd, CONNSTATE_DATA_PENDING);

	return nmsgs;
}

/*
 * msg_fd_write: A utility function used to write a message to an fd.
 * Its a very low level interface and shouldnt be used directly from
 * outside this module.
 *
 * Returns number of messages sent - 1 on success 0 on failure.
 */
int msg_fd_write(msghdr_t *msg, void *data, int fd)
{
	int nbytes;
	int retval;
	msghdr_t bo_msg;
	naddr_t *recv = &msg->receiver;

	if (fd < 0) {
		yell("msg_fd_write failed: Specified sock is invalid");
		return FALSE;
	}

	/* Copy the header and bring into correct byte order */
	bo_msg = *msg;
	hton_msg(&bo_msg);

	nbytes = 0;
	retval = write(fd, &bo_msg, sizeof(msghdr_t));
	if (retval >= 0) {
		if (retval != sizeof(msghdr_t))
			yell("msg_fd_write: Unable to send complete header");
		nbytes += retval;
	}
	else {
		yell("msg_fd_write: Error in write()");
		return FALSE;
	}

	if (msg->data_size) {
		retval = write(fd, data, msg->data_size);
		if (retval >= 0) {
			if (retval != msg->data_size)
				yell("msg_fd_write: Unable to send complete data");
			nbytes += retval;
		}
		else {
			yell("msg_fd_write: Error in write()");
			return FALSE;
		}
	}

	return (nbytes == MSG_SIZE(msg));
}

/*
 * msg_write: Send a message to the msg destination. If a socket to
 * the destination is already open, it is used, otherwise a new socket
 * with the default timeout is created and the message is sent. Before
 * sending a conversion to network byte order is performed on the msg.
 *
 * Returns number of messages sent - 1 on success 0 on failure.
 */
int msg_write(msghdr_t *msg, void *data)
{
	int sock;
	int nbytes;
	naddr_t *recv = &msg->receiver;

	if ((sock = conn_get(recv)) < 0) {
		if ((sock = conn_open(recv)) < 0)
			return FALSE;
#if 0
		else
			fprintf(stderr, "CONNECTION opened\n");
#endif
	}
#if 0
	else fprintf(stderr, "CONNECTION found\n");
#endif

	return msg_fd_write(msg, data, sock);
}

int msg_reply(msghdr_t *msg, void *data, int fd)
{
	if (fd >= 0)
		return msg_fd_write(msg, data, fd);
#ifdef NCDAEMON
	else if (memcmp(&msg->receiver, &dconf.self_id, sizeof(daemon_id)))
		return msg_write(msg, data);
#endif

	yell("Unable to send reply to LOCAL msg. fd expired.");
	return FALSE;
}

#ifdef NCDAEMON
/*
 * msg_lread: Similar to msg_read but only meant for reading LOCAL msgs.
 * A low level interface to message passing over the sockets. Reads
 * a message and stores it in the in_msgs queue.
 * Incase a partial message is read it is buffered until the complete
 * message arrives.
 *
 * Returns -1 on error if something goes really wrong, 0 if a partial
 * message is read and otherwise the total messages read into queue.
 */
int msg_lread(int fd)
{
	io_recv *ioptr;
	int nbytes;
	int nmsgs = 0;

	if (ios[fd] == NULL) {
		ios[fd] = (io_recv *)malloc(sizeof(io_recv));
		ios[fd]->buf = (char *)malloc(DEFAULT_IOBUF_SIZE);
		ios[fd]->buf_size = DEFAULT_IOBUF_SIZE;
		ios[fd]->read_pos = 0;
		ios[fd]->write_pos = 0;
		ios[fd]->cvthdr = 1;
	}

	ioptr = ios[fd];

	if (ioptr->read_pos) {
		int len = ioptr->write_pos - ioptr->read_pos;
		if (len)
			memcpy(ioptr->buf,  ioptr->buf + ioptr->read_pos, len);
		ioptr->read_pos = 0;
		ioptr->write_pos = len;
	}
	
	nbytes = read(fd, ioptr->buf + ioptr->write_pos,
			ioptr->buf_size - ioptr->write_pos);

	if (nbytes <= 0) {
#ifdef DEBUG
		if (nbytes < 0)
			perror("error in read");
#endif
		return -1;
	}
	ioptr->write_pos += nbytes;

	/* atleast a full header is present in the buffer */
	while (ioptr->write_pos - ioptr->read_pos >= sizeof(msghdr_t)) {
		int msg_size;
		msghdr_t *header = (msghdr_t *)&(ioptr->buf[ioptr->read_pos]);
		
		if (ioptr->cvthdr) /* Convert network to host byte order */
			ntoh_msg(header);
		msg_size = MSG_SIZE(header);

		/* realloc buffer if not enough space in current allocation */
		if (msg_size > ioptr->buf_size) {
			ioptr->buf = realloc(ioptr->buf, msg_size);
			ioptr->buf_size = msg_size;
		}

		/* complete message(s) present in buffer */
		if (ioptr->write_pos - ioptr->read_pos >= msg_size) {
			char *msg = (char*)malloc(msg_size);

			memcpy(msg, ioptr->buf + ioptr->read_pos, msg_size);
			msg_enqueue((msghdr_t *)msg, fd, MSGCLASS_LOCAL);

			ioptr->read_pos += msg_size;
			ioptr->cvthdr = 1;
			nmsgs++;
		}
		else {
			ioptr->cvthdr = 0;
			break;
		}
	}

	return nmsgs;
}
#endif /* NCDAEMON */

#ifdef NCDAEMON
/*
 * msg_lwrite: Send a message to the msg destination. Similar to
 * msg_write but this is a lower level interface which directly writes
 * to the given socket without getting the address from the connection
 * manager. Is useful for sending LOCAL messages. Before sending a
 * conversion to network byte order is performed on the msg.
 *
 * Returns number of messages sent - 1 on success 0 on failure.
 */
int msg_lwrite(msghdr_t *msg, void *data, task_id *tid)
{
	int sock;
	int nbytes;

	sock = task_get_msgio(tid);
	if (sock < 0) {
		yell("msg_lwrite failed: Specified task_id not found");
		return FALSE;
	}
#if 0
	else fprintf(stderr, "LOCAL CONNECTION found\n");
#endif

	return msg_fd_write(msg, data, sock);
}
#endif /* NCDAEMON */


/*
 * fd_authorize: Authorizes an fd so that it can be written back
 * to allowing two way comminication without opening two separate
 * sockets, one for sending and one for receiving. fd_close
 * automatically unauthorizes an fd.
 */
int fd_authorize(int fd)
{
	io_type[fd] |= FDTYPE_AUTH;
	return TRUE;
}

/*
 * fd_unauthorize: The opposite of the above function.
 */
int fd_unauthorize(int fd)
{
	io_type[fd] &= ~(FDTYPE_AUTH);
	io_type[fd] |= FDTYPE_NOAUTH;
	return TRUE;
}

/*
 * fd_set_type: Resets the type of a fd to the specied value
 * and returns the previous type of the fd.
 */
int fd_set_type(int fd, int type)
{
	int prev_type = -1;

	if (FD_ISSET(fd, &fd_ractive)) {
		prev_type = (io_type[fd] & ~(FDTYPE_AUTH));
		io_type[fd] &= FDTYPE_AUTH;
		io_type[fd] |= type;
	}

	return prev_type;
}

/*
 * fd_open: Register an fd with the module to allow I/O operations
 */
int fd_open(int fd, int type)
{
	if (!ios)
		init_io();

	if (fd < 0)
		return fd;

	FD_SET(fd, &fd_ractive);

	if (fd > global_max_fd)
		global_max_fd = fd;

	io_type[fd] = type | FDTYPE_NOAUTH;

	return TRUE;
}

/*
 * fd_close: Unregister an fd from the module.
 */
int fd_close(int fd)
{
	if (ios && ios[fd]) {
		if (ios[fd]->buf)
			free(ios[fd]->buf);
		free(ios[fd]);
		ios[fd] = NULL;
	}

	if (FD_ISSET(fd, &fd_ractive)) {
		FD_CLR(fd, &fd_ractive);
		close(fd);
		fd_has_closed = TRUE;
	}

	if (fd == global_max_fd) {
		do {
		       	global_max_fd--;
			if (global_max_fd < 0)
				break;
		} while(!FD_ISSET(global_max_fd, &fd_ractive));
	}

	io_type[fd] = FDTYPE_NONE;

	return TRUE;
}

/*
 * fd_close_all: Unregister all fd from the module.
 */
int fd_close_all()
{
	int i;

	for (i=0;i<global_max_fd;i++) {
		if (FD_ISSET(i, &fd_ractive)) {
			FD_CLR(i, &fd_ractive);
			fd_close(i);
		}
	}

	fd_has_closed = TRUE;

	return TRUE;
}

/*
 * fd_open_write: Register a writable fd with the module to allow
 * I/O operations
 */
int fd_open_write(int fd, int type)
{
	if (!ios)
		init_io();

	if (fd < 0)
		return fd;

	FD_SET(fd, &fd_wactive);

	if (fd > global_max_fd)
		global_max_fd = fd;

	io_type[fd] = type | FDTYPE_NOAUTH;

	return TRUE;
}

/*
 * fd_close_write: Unregister a writable fd from the module
 */
int fd_close_write(int fd)
{
	if (ios && ios[fd]) {
		if (ios[fd]->buf)
			free(ios[fd]->buf);
		free(ios[fd]);
		ios[fd] = NULL;
	}

	FD_CLR(fd, &fd_wactive);
	close(fd);

	if (fd == global_max_fd) {
		do {
			global_max_fd--;
			if (global_max_fd < 0)
				break;
		} while(!FD_ISSET(global_max_fd, &fd_ractive));
	}
	return TRUE;
}
