/*
 * libnc.c: Libnc functions which are linked with user tasks.
 * (C) Copyright 2003, netcluster
 *
 * Written by Kashif Jamal, Ghaus Ali, Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

/* for ntoh and hton functions */
#include "util.h"

/* for msghdr and MSG type defines */
#include "msghnd.h"
#include "libnc.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * LIBNC UTILITY FUNCTIONS
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Stores info of the last RPC call. Used by nc_treturn. */
rpcinfo_t nc_lastrpc;

/*
 * msg_fd_write: A utility function used to write a message to an fd.
 * Its a very low level interface and shouldnt be used directly from
 * outside this module.
 *
 * Returns number of messages sent - 1 on success 0 on failure.
 */
int msg_fd_write(msghdr_t *msg, void *data, int fd)
{
	int nbytes;
	int retval;
	msghdr_t bo_msg;
	naddr_t *recv = &msg->receiver;

	if (fd < 0) {
		return FALSE;
	}

	/* Copy the header and bring into correct byte order */
	bo_msg = *msg;
	hton_msg(&bo_msg);

	nbytes = 0;
	retval = write(fd, &bo_msg, sizeof(msghdr_t));
	if (retval >= 0) {
		if (retval != sizeof(msghdr_t))
			fprintf(stderr, "msg_fd_write: Unable to send complete header");
		nbytes += retval;
	}
	else {
		fprintf(stderr, "msg_fd_write: Error in write()");
		return FALSE;
	}

	if (msg->data_size) {
		retval = write(fd, data, msg->data_size);
		if (retval >= 0) {
			if (retval != msg->data_size)
				fprintf(stderr, "msg_fd_write: Unable to send complete data");
			nbytes += retval;
		}
		else {
			fprintf(stderr, "msg_fd_write: Error in write()");
			return FALSE;
		}
	}

	return (nbytes == MSG_SIZE(msg));
}

/*
 * read_nbytes: A low level interface to reading a fixed number of
 * bytes from a given file descriptor. read() is repeatedly called
 * until exactly n bytes are read. Returns TRUE on success and FALSE
 * on failure.
 */
int read_nbytes(int fd, char *buffer, int n)
{
	int retval;
	int nbytes = 0;

	while (nbytes < n) {
		retval = read(fd, buffer+nbytes, n-nbytes);
		
		if (retval < 0) {
			fprintf(stderr, "Error in read(). Aborting.\n");
			return FALSE;
		}

		nbytes += retval;
	}

	return TRUE;
}

/*
 * msg_fd_read: This too is a utility function which is the msg
 * reading counterpart to msg_fd_write. It is a blocking function
 * which does not return until it reads a complete msg from the
 * given file descriptor. Returns TRUE on success.
 */
int msg_fd_read(msghdr_t *hdr, void *data, int fd)
{
	read_nbytes(fd, (char*)hdr, sizeof(msghdr_t));
	ntoh_msg(hdr);
	if (hdr->data_size)
		read_nbytes(fd, (char*)data, hdr->data_size);

	return TRUE;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * LIBNC INTERFACE FUNCTIONS
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * nc_terminate: Terminates the current process. This function
 * never returns and only terminates the calling task with
 * the specified exit status. The rest of the tasks may continue
 * running. This call is also implicitly made if the main
 * function of the task returns to the caller (LIBNC).
 */
void nc_terminate(int status)
{
	msghdr_t msghdr;
	exitinfo_t msgdata;

	printf("Task terminated.\n");
	fflush(stdout);

	/* These are used as wait states to flush all
	 * unprocessed msgs to the daemon. */
	nc_myid();
	nc_myid();

	msghdr.msg_type = MSG_TERMINATE;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(exitinfo_t);
	msghdr.trans_id = TRANSID_NONE;

	msgdata.status = status;

	hton_exitinfo(&msgdata);
	msg_write(&msghdr, &msgdata);

	exit(status);
}

/*
 * nc_abort: This call results in the abortion of the whole
 * user application including all tasks. The function call
 * never returns. The status parameter could be passed to
 * give an error code to the user. This is an abnormal way
 * to terminate a task so must be avoided.
 */
void nc_abort(int status)
{
	msghdr_t msghdr;
	exitinfo_t msgdata;

	printf("Task aborted.\n");
	fflush(stdout);

	/* These are used as wait states to flush all
	 * unprocessed msgs to the daemon. */
	nc_myid();
	nc_myid();

	msghdr.msg_type = MSG_ABORT;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(exitinfo_t);
	msghdr.trans_id = TRANSID_NONE;

	msgdata.status = status;

	hton_exitinfo(&msgdata);
	msg_write(&msghdr, &msgdata);

	exit(status);
}

/*
 * nc_myid: Returns the task_id of the calling task. This
 * function blocks until the myid reply is received from the
 * controlling daemon.
 */
task_id nc_myid()
{
	msghdr_t msghdr;
	msghdr_t rplyhdr;
	task_id rplydata;

	msghdr.msg_type = MSG_MY_TASKID;
	msghdr.priority = 1;
	msghdr.data_size = 0;
	msghdr.trans_id = TRANSID_NONE;

	msg_write(&msghdr, NULL);

	/* Must wait for a myid reply here with a msgtype of
         * MSG_MY_TASKID. Its reply data need not be converted
	 * bcoz task_id is always in network byte order. */
	msg_read(&rplyhdr, &rplydata);

	return rplydata;
}

/*
 * nc_createtask: Creates a new task, either remotely or locally.
 * The function blocks until either the task is successfully
 * or task creation fails. It returns the task_id of the newly
 * created task or TASK_ID_NULL on failure.
 */
task_id nc_createtask(char *start_sym)
{
	msghdr_t msghdr;
	taskreq_t msgdata;
	msghdr_t rplyhdr;
	taskrply_t rplydata;

	msghdr.msg_type = MSG_CREATE_TASK;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(taskreq_t);
	msghdr.trans_id = 1;
	msghdr.trans_isrply = 0;

	strcpy(msgdata.start_sym, start_sym);
	msgdata.priority = 1;

	hton_taskreq(&msgdata);
	msg_write(&msghdr, &msgdata);

	/* Must wait for a confirmation here since
	 * MSG_CREATE_TASK is a transaction. A MSG_CREATE_TASK_RPLY
	 * would arrive at FD_MSG which must be read. */
	msg_read(&rplyhdr, &rplydata);
	ntoh_taskrply(&rplydata);

	if (rplydata.status == TASKRPLY_STATUS_SUCCESS)
		return rplydata.id;

	TASK_ID_NULL(rplydata.id);
	return rplydata.id;
}

/*
 * nc_call: Is equivalent of a remote procedure call. The parameters
 * are as follows:
 * tid: the task_id of the task being called
 * funcid: The user function ID of the function being called, which
 *     could be any numeric ID assigned by the user and which is
 *     unique for a particular function of this user application.
 * The following list of variable parameters are those taken by the
 * called funcid. At the end of this list, one additional parameter
 * is passed by reference, which takes the return value of the
 * function, if any. If the return type if void, this last parameter
 * need not be passed.
 *
 * Returns TRUE on success and FALSE otherwise.
 */
int nc_call(task_id *tid, u_int32_t funcid, ...)
{
	va_list ap;
	int len;
	msghdr_t msghdr;
	char msgdata[sizeof(rpcinfo_t)+MAX_RPCARG_SIZE];
	rpcinfo_t *rpc_data = (rpcinfo_t*)msgdata;
	char *prm_data = msgdata+sizeof(rpcinfo_t);

	va_start(ap, funcid);
	len = put_param(prm_data, MAX_RPCARG_SIZE, nc_rpcs[funcid].fmt_param, ap);
	va_end(ap);

	if (len < 0) {
		fprintf(stderr, "NCLIB_ERR: nc_call<funcid=%d> args exceed max size\n", funcid);
		return FALSE;
	}

	msghdr.msg_type = MSG_RPC_CALL;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(rpcinfo_t) + len;
	msghdr.trans_id = TRANSID_NONE;

	/* rpc_data->src would be filled in by the daemon */
	rpc_data->status = RPC_STATUS_PENDING;
	rpc_data->dest = *tid;
	rpc_data->func_id = funcid;

	hton_rpcinfo(rpc_data);
	msg_write(&msghdr, msgdata);

	/* Now wait for MSG_RPC_RPLY which would contain the
	 * reply data of the RPC call if the type isnt void. */
	msg_read(&msghdr, msgdata);
	ntoh_rpcinfo(rpc_data);

	if (rpc_data->status != RPC_STATUS_SUCCESS)
		return FALSE;

	/* Now read the return value if any */
	if (strcmp(nc_rpcs[funcid].fmt_ret, "%z")) {
		va_start(ap, funcid);
		len = get_sparam(prm_data, nc_rpcs[funcid].fmt_param, nc_rpcs[funcid].fmt_ret, ap);
		va_end(ap);

		if (len < 0) {
			fprintf(stderr, "NCLIB_ERR: nc_call<funcid=%d> return values exceed max size\n", funcid);
			return FALSE;
		}
	}

	return TRUE;
}

/*
 * nc_accept: This is the counterpart of the nc_call mechanism. It
 * accepts a call and returns the parameters in the user specified
 * variables in the variable argument list. All such variables
 * must be passed by reference (like in scanf).
 */
int nc_accept(u_int32_t funcid, ...)
{
	va_list ap;
	int len;
	msghdr_t msghdr;
	rpcaccept_t msgdata;
	msghdr_t rplyhdr;
	char rplydata[sizeof(rpcinfo_t)+MAX_RPCARG_SIZE];
	rpcinfo_t *rpc_data = (rpcinfo_t*)rplydata;
	char *prm_data = rplydata+sizeof(rpcinfo_t);

	msghdr.msg_type = MSG_RPC_ACCEPT;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(rpcaccept_t);
	msghdr.trans_id = TRANSID_NONE;

	msgdata.func_id = funcid;
	msgdata.is_blocking = TRUE;

	hton_rpcaccept(&msgdata);
	msg_write(&msghdr, &msgdata);

	/* Now wait for MSG_RPC_CALL which would contain the
	 * arguments of the RPC call and other info. */
	msg_read(&rplyhdr, rplydata);
	ntoh_rpcinfo(rpc_data);

	if (rpc_data->status != RPC_STATUS_PENDING)
		return FALSE;

	va_start(ap, funcid);
	len = get_param(prm_data, nc_rpcs[funcid].fmt_param, ap);
	if (len < 0)
		return FALSE;
	va_end(ap);

	nc_lastrpc = *rpc_data;

	return TRUE;
}

/*
 * nc_accept_noblock: This is the non-blocking version of the
 * above nc_accept function. If a call is available, it
 * accepts and returns the parameters in the user specified
 * variables in the variable argument list. Otherwise it
 * returns FALSE.
 */
int nc_accept_noblock(u_int32_t funcid, ...)
{
	va_list ap;
	int len;
	msghdr_t msghdr;
	rpcaccept_t msgdata;
	msghdr_t rplyhdr;
	char rplydata[sizeof(rpcinfo_t)+MAX_RPCARG_SIZE];
	rpcinfo_t *rpc_data = (rpcinfo_t*)rplydata;
	char *prm_data = rplydata+sizeof(rpcinfo_t);

	msghdr.msg_type = MSG_RPC_ACCEPT;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(rpcaccept_t);
	msghdr.trans_id = TRANSID_NONE;

	msgdata.func_id = funcid;
	msgdata.is_blocking = FALSE;

	hton_rpcaccept(&msgdata);
	msg_write(&msghdr, &msgdata);

	/* Now wait for MSG_RPC_CALL which would contain the
	 * arguments of the RPC call and other info. */
	msg_read(&rplyhdr, rplydata);
	ntoh_rpcinfo(rpc_data);

	if (rpc_data->status != RPC_STATUS_PENDING)
		return FALSE;

	va_start(ap, funcid);
	len = get_param(prm_data, nc_rpcs[funcid].fmt_param, ap);
	if (len < 0)
		return FALSE;
	va_end(ap);

	nc_lastrpc = *rpc_data;

	return TRUE;
}

/*
 * nc_treturn: Used to end an RPC accept call. The calling
 * task does not resume until this function is called and
 * returns a value, if required by the user function.
 * The dummy parameter takes any value and is only needed
 * for the variable argument list usage.
 */
int nc_treturn(int dummy, ...)
{
	va_list ap;
	int len;
	msghdr_t msghdr;
	char msgdata[sizeof(rpcinfo_t)+MAX_RPCARG_SIZE];
	rpcinfo_t *rpc_data = (rpcinfo_t*)msgdata;
	char *prm_data = msgdata+sizeof(rpcinfo_t);

	if (strcmp(nc_rpcs[nc_lastrpc.func_id].fmt_ret, "%z")) {
		va_start(ap, dummy);
		len = put_param(prm_data, MAX_RPCARG_SIZE,
				nc_rpcs[nc_lastrpc.func_id].fmt_ret, ap);
		if (len < 0)
			return FALSE;
		va_end(ap);
	}
	else
		len = 0;

	msghdr.msg_type = MSG_RPC_RPLY;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(rpcinfo_t) + len;
	msghdr.trans_id = TRANSID_NONE;

	*rpc_data = nc_lastrpc;
	rpc_data->status = RPC_STATUS_SUCCESS;

	hton_rpcinfo(rpc_data);
	msg_write(&msghdr, msgdata);

	return TRUE;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * TASK ENTRY POINT
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * main: This is the hidden entry point of the task. Control
 * is transfered from here to task_main or other subtasks
 * depending on the command line arguments.
 */
int main(int argc, char *argv[])
{
	int i;

	if (argc < 2) {
		fprintf(stderr, "Incorrect invocation of user task. Aborting...\n");	
		return 1;
	}

	if (!strcmp(argv[1], TASK_MASTER_STARTSYM_STR)) {
		TASK_MASTER_STARTSYM(argc-1, &argv[1]);
	}
	else {
		for (i=0;nc_tasks[i].task_name;i++) {
			if (!strcmp(nc_tasks[i].task_name, argv[1])) {
				(*nc_tasks[i].task_sym)();
				break;
			}
		}

		if (nc_tasks[i].task_name==NULL)
			fprintf(stderr, "%s: Undefined task type <%s>. Aborting...\n", argv[0], argv[1]);
	}

	nc_terminate(0);

	return 0;
}
