/*
 * log.c: Handles all logging. Maintains different levels
 * of logging as specified in the config file.
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <stdarg.h>

#include "defs.h"
#include "util.h"
#include "log.h"

void do_log_orig(int type, char *fmt, ...);
void yell_orig(char *fmt, ...);
void (*do_log)(int type, char *fmt, ...) = do_log_orig;
void (*yell)(char *fmt, ...) = yell_orig;

int log_types = DEFAULT_LOGGING;
char log_buffer[MAX_LOG_SIZE];
char *mod_name = NULL;
FILE *log_file = NULL;

/*
 * set_log_types: Start receiving messages of the given types. The
 * 'log' variable must be ORd with LOG_NORMAL, LOG_PANIC etc.
 * Returns TRUE on successfully setting the flag.
 */
int set_log_types(int type)
{
	log_types = type;
	return TRUE;
}

/*
 * set_log_type: Start receiving messages of the given type. The
 * 'log' variable must be one of LOG_NORMAL, LOG_PANIC etc.
 * Returns TRUE on successfully setting the flag.
 */
int set_log_type(int type)
{
	log_types |= type;
	return TRUE;
}

/*
 * clear_log_type: Stop receiving messages of the given type. This
 * function is an opposite of set_log_type. Returns TRUE on success.
 */
int clear_log_type(int type)
{
	log_types &= (~type);
	return TRUE;
}

/*
 * get_log_types: Returns the log_types which holds the bits for the
 * types of messages being logged. Bitwise OR the types to check
 * if a particular kind of message is being logged.
 */
int get_log_types()
{
	return log_types;
}

/*
 * set_module_name: The module name is printed with each log msg.
 * By default the module name is blank.
 */
int set_module_name(char *mname)
{
	mod_name = mname;
#ifndef DEBUG
	openlog(mod_name, LOG_PID, LOG_DAEMON);
#endif
	return TRUE;
}

/*
 * set_log_file: Sets the file to which to log the messages. If not
 * called or called with NULL argument, or if unable to open file,
 * syslog is used.
 */
int set_log_file(char *fname)
{
	char *str;

	if (fname) {
		str = expand_tilde(fname);
		log_file = fopen(str, "at");
		if (str)
			free(str);

		return (log_file) ? TRUE : FALSE;
	}
	return FALSE;
}

/*
 * do_log: Log a message of the specified log type. The message can
 * be specified using the printf style format string and a variable
 * number of arguments.
 */
void do_log_orig(int type, char *fmt, ...)
{
	va_list args;

	/* Ignore if this type of msgs not being logged */
	if (!(log_types & type))
		return;

	va_start(args, fmt);
	vsnprintf(log_buffer, MAX_LOG_SIZE-1, fmt, args);
	va_end(args);

	log_buffer[MAX_LOG_SIZE-1] = '\0';

#ifdef DEBUG
	if (mod_name && *mod_name)
		fprintf(stderr, "%s: ", mod_name);
	fprintf(stderr, "%s\n", log_buffer);
	fflush(stderr);
#else
	if (log_file) {
		if (mod_name && *mod_name)
			fprintf(log_file, "%s: ", mod_name);
		fprintf(log_file, "%s\n", log_buffer);
		fflush(log_file);
	}
	else {
		/* log to syslog */
		syslog(type, "%s", log_buffer);
	}
#endif
}

/*
 * yell: An easy way to log messages using a normal level of logging.
 * Used for normal to slightly critical messages. The message can be
 * specified using printf style format string and variable arguments.
 */
void yell_orig(char *fmt, ...)
{
	va_list args;

	/* Ignore if this type of msgs not being logged */
	if (!(log_types & LOG_ERR))
		return;

	va_start(args, fmt);
	vsnprintf(log_buffer, MAX_LOG_SIZE-1, fmt, args);
	va_end(args);

	log_buffer[MAX_LOG_SIZE-1] = '\0';

#ifdef DEBUG
	if (mod_name && *mod_name)
		fprintf(stderr, "%s: ", mod_name);
	fprintf(stderr, "%s\n", log_buffer);
	fflush(stderr);
#else
	if (log_file) {
		if (mod_name && *mod_name)
			fprintf(log_file, "%s: ", mod_name);
		fprintf(log_file, "%s\n", log_buffer);
	}
	else {
		/* log to syslog */
		syslog(LOG_ERR, "%s", log_buffer);
	}
#endif
}
