/*
 * msghnd.c: Message handlers of both the daemon and the
 * VAS - however, only one of the two are loaded at a time.
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 * VAS Message Handlers written by Syed Ghaus Ali
 * FTP Message Handlers written by Syed Ghaus Ali
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>


#include "defs.h"
#include "sock.h"
#include "io.h"
#include "log.h"
#include "msgs.h"
#include "transact.h"
#include "fileman.h"
#include "util.h"


#include "vas.h"
#include "chash.h"

#ifdef NCDAEMON
#include "task.h"
#endif


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                  M e s s a g e   H a n d l e r s
 *
 * Below are the functions for handling incoming messages; one for each
 * type of message. Check the MSG_HANDLER definition in msgs.h for
 * the parameters that the message handler gets. All these are callback
 * functions. They return the new status of the message, which is
 * MSGSTAT_DEL if it is used and maybe removed from the msg queue.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * COMMONLY USED MSG HANDLERS
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* MSG_INVALID: Invalid msg handler. */
MSG_HANDLER(hnd_invalid)
{
	yell("Msg invalid msg_type<%d> data_size<%d>. Trashed!",
			hdr->msg_type, hdr->data_size);
	return MSGSTAT_DEL;
}

/* MSG_PING: Returns PONG to the pinger. */
MSG_HANDLER(hnd_ping)
{
	msghdr_t msg;

	msg.sender = hdr->receiver;
	msg.receiver = hdr->sender;
	msg.msg_type = MSG_PONG;
	msg.priority = 1;
	msg.data_size = 0;
	
	msg_write(&msg, NULL);
	
	return MSGSTAT_DEL;
}

/* MSG_PONG: Does nothing useful. */
MSG_HANDLER(hnd_pong)
{
	return MSGSTAT_DEL;
}

/* MSG_SYSLOG: Logs the receieved msg. */
MSG_HANDLER(hnd_syslog)
{
	char *msg = data;

	do_log(LOG_NOTICE, msg);
	return MSGSTAT_DEL;
}


#ifdef NCVAS
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * VAS MESSAGE HANDLERS
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

extern chash vas_table;


/*
 * MSG_VAS_PONG: Receives the response of the ping sent by the
 * VAS to poll all the connected machines.
 * Expects: none
 * Type: Transaction-Reply
 */
MSG_HANDLER(hnd_vas_pong)
{
	do_log(LOG_INFO, "PONG received");
	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_REG: This msg is sent by a daemon to register itself
 * with the VAS. It contains the vasreg_req_t struct. A reply with
 * MSG_VAS_REG_RPLY is sent to the sender along with vasreg_rply_t
 * as data of the msg. This struct contains the request status.
 * Expects: vasload_t
 * Type: Transaction
 */
MSG_HANDLER(hnd_vas_reg)
{
	msghdr_t msg;
	vasclient_t vitem;
	vasreg_rply_t vrply;
	vasload_t *load = (vasload_t*)data;

	ntoh_vasload(load);
	
	if (hdr->trans_id == TRANSID_NONE)
		return MSGSTAT_DEL;

	msg.sender = hdr->receiver;
	msg.receiver = hdr->sender;
	msg.msg_type = MSG_VAS_REG_RPLY;
	msg.priority = 1;
	msg.data_size = sizeof(vasreg_rply_t);
	msg.trans_id = hdr->trans_id;

	vrply.status = VASREG_STATUS_SUCCESS;
	vitem.dem_id = hdr->sender;

	if (!chash_find_custom(&vas_table, &vitem, ncvas_item_cmp)) {
		vasclient_t *vitem_ptr = (vasclient_t*)malloc(sizeof(vasclient_t));
		vitem_ptr->dem_id = hdr->sender;
		vitem_ptr->load = *load;

		chash_insert_sorted(&vas_table, vitem_ptr, ncvas_itemload_cmp);

		do_log(LOG_INFO, "Daemon register: Successful");
	}
	else {
		/* XXX: Update the load of the registration with 'load' */
		do_log(LOG_INFO, "Daemon register: Successful (Already present)");
	}

	hton_vasreg_rply(&vrply);
	trans_reply(&msg, &vrply, fd);

	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_UNREG: Msg handler at VAS that is called to unregister
 * a daemon from the client list.
 * Expects: none
 * Type: Async
 */
MSG_HANDLER(hnd_vas_unreg)
{
	vasclient_t vitem;
	
	vitem.dem_id = hdr->sender;

#ifdef DEBUG
	do_log(LOG_INFO, "Daemon unregister: Request to remove ip<%d> port<%d>", ntohl(hdr->sender.ip),
			ntohs(hdr->sender.port));
#endif

	chash_remove_custom(&vas_table, &vitem, ncvas_item_cmp);
	do_log(LOG_INFO, "Daemon unregister: Successful");

	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_ISREG: Responds to this message with a positive of a
 * negative reply to the called. It is used to verify if a client
 * is still registered with the VAS.
 * Expects: none
 * Type: Transaction
 */
MSG_HANDLER(hnd_vas_isreg)
{
	msghdr_t msg;
	vasclient_t vitem;
	vasreg_rply_t vrply;
	clist *find;

	/* Init header for reply message */
	msg.sender = hdr->receiver;
	msg.receiver = hdr->sender;
	msg.msg_type = MSG_VAS_ISREG_RPLY;
	msg.priority = 1;
	msg.data_size = sizeof(vasreg_rply_t);
       	msg.trans_id = hdr->trans_id;

	vitem.dem_id = hdr->sender;

	find = chash_find_custom(&vas_table, &vitem, ncvas_item_cmp);

	if (find) {
		vrply.status = VASREG_STATUS_SUCCESS;
		do_log(LOG_INFO, "Daemon isreg request: Success");
	}
	else {
		vrply.status = VASREG_STATUS_FAILURE;
		do_log(LOG_INFO, "Daemon isreg request: Failure");
	}

	hton_vasreg_rply(&vrply);
	trans_reply(&msg, &vrply, fd);

	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_SETLOAD: (re)Sets the load recorded by the VAS to the
 * specified value. Sends no reply.
 * Expects: vasload_t
 * Type: Async
 */
MSG_HANDLER(hnd_vas_setload)
{
char *yeah;
	vasclient_t vitem;
	clist *find;
	vasload_t *vload = (vasload_t*)data;

	ntoh_vasload(vload);

	vitem.dem_id = hdr->sender;
	vitem.load = *vload;
yeah = naddr_to_str(&vitem.dem_id);
do_log(LOG_INFO, "Setload: from daemon<%s>", yeah);
	find = chash_find_custom(&vas_table, &vitem, ncvas_item_cmp);

	if (find) {
		vasclient_t *vitem_new = malloc(sizeof(vasclient_t));
		*vitem_new = *(vasclient_t*)find->data;
		vitem_new->load = *vload;

		chash_remove(&vas_table, find->data);

		chash_insert_sorted(&vas_table, vitem_new, ncvas_itemload_cmp);
		do_log(LOG_INFO, "Setload: Success");
	}
	else {
		do_log(LOG_INFO, "Setload: Failure. Item not found.");
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_GETLOAD: This message requests the VAS to send the load
 * of the requesting client. Either the load is sent or no reply is
 * sent incase the calling client isnt registered with this VAS.
 * Expects: none
 * Type: Transaction
 */
MSG_HANDLER(hnd_vas_getload)
{
	msghdr_t msg;
	vasclient_t vitem;
	clist *find;

	vitem.dem_id = hdr->sender;
	find = chash_find_custom(&vas_table, &vitem, ncvas_item_cmp);

	if (find) {
		vasclient_t *vitem_ptr = find->data;
		vasload_t vload = vitem_ptr->load;

		msg.sender = hdr->receiver;
		msg.receiver = hdr->sender;
		msg.msg_type = MSG_VAS_GETLOAD_RPLY;
		msg.priority = 1;
		msg.data_size = sizeof(vasload_t);
		msg.trans_id = hdr->trans_id;

		hton_vasload(&vload);
		trans_reply(&msg, &vload, fd);
		do_log(LOG_INFO, "Getload: Successfully sent load");
	}
	else
		do_log(LOG_INFO, "Getload: Request trashed - Sender not registered");

	return MSGSTAT_DEL;
}


/*
 * MSG_VAS_GETMINLOAD: Returns the daemon with the minimum load to
 * the caller. It is possible that the minimum load daemon is the
 * caller itself. No reply if no clients are registered.
 * Expects: none
 * Type: Transaction
 */
MSG_HANDLER(hnd_vas_getminload)
{
	msghdr_t msg;
	clist *find;

	find = chash_find_min_head(&vas_table, ncvas_itemload_cmp);

	if (find) {
		vasclient_t *vitem = find->data;

		msg.sender = hdr->receiver;
		msg.receiver = hdr->sender;
		msg.msg_type = MSG_VAS_GETMINLOAD_RPLY;
		msg.priority = 1;
		msg.data_size = sizeof(daemon_id);
		msg.trans_id = hdr->trans_id;
		
		trans_reply(&msg, &vitem->dem_id, fd);
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_DISPLAY: Display the vas_table, listing all the
 * clients which are currently registered with the VAS.
 * Expects: none
 * Type: Async
 */
MSG_HANDLER(hnd_vas_display)
{
	chash_foreach(&vas_table, ncvas_item_display, NULL);
	return MSGSTAT_DEL;
}
#endif /* NCVAS */



#ifdef NCDAEMON
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * DAEMON MESSAGE HANDLERS
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * MSG_VAS_PING: Handler for the PING sent by the VAS. Only responds
 * if the ping is from the currently connected daemon (dconf.vas_cur).
 * Expects: none
 * Type: Transaction
 */
MSG_HANDLER(hnd_vas_ping)
{
	msghdr_t msg;

	/* Reply only if PING from the current VAS server */
	if (!memcmp(&hdr->sender, &dconf.vas_cur, sizeof(naddr_t))) {
		msg.sender = hdr->receiver;
		msg.receiver = hdr->sender;
		msg.msg_type = MSG_VAS_PONG;
		msg.priority = 1;
		msg.data_size = 0;
		msg.trans_id = hdr->trans_id;

		if (trans_reply(&msg, NULL, fd))
			do_log(LOG_INFO, "PING reply: Sent to VAS");
		else
			do_log(LOG_ERR, "PING reply: Unable to send to VAS");
	}
	else
		do_log(LOG_ERR, "PING reply: Not sent - PING not from current VAS");

	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_REG_RPLY: A reply sent in response to registration with
 * the VAS. The reply could either be success or failure. Sets
 * vconf.vas_cur if registration succeeds.
 * Expects: vasreg_rply_t
 * Type: Transaction-Reply
 */
MSG_HANDLER(hnd_vas_reg_rply)
{
	vasreg_rply_t *vreg_rply = data;

	ntoh_vasreg_rply(vreg_rply);

	if (vreg_rply->status == VASREG_STATUS_SUCCESS) {
		dconf.vas_cur = hdr->sender;
		do_log(LOG_NOTICE, "VAS register: Successful");
	}
	else {
		/* XXX: Add code to call ncdaemon_vas_register */
		do_log(LOG_NOTICE, "VAS register: Failure");
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_ISREG_RPLY: This msg is sent in reply to MSG_VAS_ISREG.
 * It returns the status of the registration of the daemon in
 * vasreg_rply_t.
 * Expects: vasreg_rply_t
 * Type: Transaction-Reply
 */
MSG_HANDLER(hnd_vas_isreg_rply)
{
	vasreg_rply_t *vreg_rply = data;
	ntoh_vasreg_rply(vreg_rply);

	if (vreg_rply->status == VASREG_STATUS_SUCCESS)
		do_log(LOG_INFO, "VAS isreg reply: Success");
	else {
		/* XXX: Add code to call ncdaemon_vas_register */
		do_log(LOG_INFO, "VAS isreg reply: Failure");
	}
	
	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_GETLOAD_RPLY: This reply receives the vasload_t of
 * ourselves which is currently maintained by the VAS.
 * Expects: vasload_t
 * Type: Transaction-Reply
 */
MSG_HANDLER(hnd_vas_getload_rply)
{
	vasload_t *vload = data;
	ntoh_vasload(vload);

	do_log(LOG_INFO, "Getload reply: numprocs<%d>", vload->numprocs);

	return MSGSTAT_DEL;
}

/*
 * MSG_VAS_GETMINLOAD_RPLY: This reply received the daemon_id of
 * the daemon with the least load on the network.
 * Expects: daemon_id
 * Type: Transaction-Reply
 */
MSG_HANDLER(hnd_vas_getminload_rply)
{
	taskreq_t *treq;
	daemon_id *dem_minload = data;

	if (hdr->trans_id == TRANSID_NONE || !trans || !trans->faildata)
		return MSGSTAT_DEL;

	treq = trans->faildata;

	/* Code continued for create_task */
	if (!create_task_ftp(treq, dem_minload)) {
		/* XXX: Add code to create a local task instead */
		yell("Create task: getminload: Failed. Aborting task creation.");
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_CREATE_MASTER: A transaction based request handler. Replies
 * to the transaction with trans_reply. This message is usually
 * sent by ncloader. Only the 'file_name' and 'priority' fields of
 * the taskreq_t are used. The rest are ignored.
 * Expects: taskreq_t taskarg_t [zero_separated_arguments]
 * Type: Transaction (LOCAL)
 */
MSG_HANDLER(hnd_create_master)
{
	msghdr_t rply_hdr;
	taskrply_t rply_data;
	taskreq_t *treq = data;
	taskarg_t *targ = (taskarg_t*)((char*)data+sizeof(taskreq_t));
	char *targdata = (char*)data+sizeof(taskreq_t)+sizeof(taskarg_t);
	task_id *tid = NULL;

	ntoh_taskreq(treq);
	ntoh_taskarg(targ);

	if (fd < 0) {
		do_log(LOG_ERR, "Create master: No reply fd found");
		return MSGSTAT_DEL;
	}
	else {
		/* XXX: Only handles a single create master msg from
		 * a loader. Inefficient for multiple so check msgclass. */

		/* Convert the fd to a local msg socket with no ttl */
		conn_fd_unreg(fd);
		fd_set_type(fd, FDTYPE_LOCALMSG);
	}

	rply_hdr.sender = dconf.self_id;
	rply_hdr.receiver = dconf.self_id;
	rply_hdr.msg_type = MSG_CREATE_TASK_RPLY;
	rply_hdr.priority = 1;
	rply_hdr.data_size = sizeof(taskrply_t);
	rply_hdr.trans_id = hdr->trans_id;

	if (access(treq->file_name, X_OK)) {
		yell("Master task creation failed: File not executable<%s>",
				treq->file_name);
		rply_data.status = TASKRPLY_STATUS_ERR;
	}
	else if (!(tid = create_master(treq->file_name, treq->priority, targ->argc, targdata, fd))) {
		yell("Unable to create master: create_master failed");
		rply_data.status = TASKRPLY_STATUS_ERR;
	}

	if (tid) {
		rply_data.id = *tid;
		rply_data.master = *tid;
		TASK_ID_NULL(rply_data.parent);
		task_inc_load();
		task_update_load();
	}
	rply_data.status = TASKRPLY_STATUS_SUCCESS;
	hton_taskrply(&rply_data);
	trans_lreply(&rply_hdr, &rply_data, fd);

	return MSGSTAT_DEL;
}

/*
 * MSG_CREATE_TASK: Creates a task. This message is sent by
 * a task to a daemon locally and only the 'priority' and
 * 'start_sym' fields are used. The rest are filled implicitly.
 * Expects: taskreq_t
 * Type: Transaction (LOCAL)
 */
MSG_HANDLER(hnd_create_task)
{
	msghdr_t rply_hdr;
	taskrply_t rply_data;
	task_id *tid;
	taskinfo_t *tinf;
	taskreq_t *treq = data;

	ntoh_taskreq(treq);

	/* get task info of the parent task */
	tinf = task_find_msgio(fd);

	if (!tinf) {
		yell("Create task: Non-local message for task creation. Ignored.");
		return MSGSTAT_DEL;
	}

	treq->master = tinf->master;
	treq->parent = tinf->id;
	strcpy(treq->file_name, tinf->file_name);

	rply_hdr.sender = dconf.self_id;
	rply_hdr.receiver = dconf.self_id;
	rply_hdr.msg_type = MSG_CREATE_TASK_RPLY;
	rply_hdr.priority = 1;
	rply_hdr.data_size = sizeof(taskrply_t);
	rply_hdr.trans_id = hdr->trans_id;

	if (access(tinf->file_name, R_OK|X_OK)) {
		yell("Create task: Failed - File not readable/executable<%s>",
				tinf->file_name);
		rply_data.status = TASKRPLY_STATUS_ERR;
		hton_taskrply(&rply_data);
		trans_lreply(&rply_hdr, &rply_data, fd);
	}
	else if (IS_UNKNOWN_NADDR(dconf.vas_cur)) {
		/* No VAS server is set. Create a task locally. */
		tid = create_local_task(treq->file_name, treq->start_sym, &treq->master,
				&treq->parent, treq->priority);

		if (tid == NULL) {
			rply_data.status = TASKRPLY_STATUS_ERR;
			do_log(LOG_ERR, "Create task: Failed (no VAS)");
		}
		else {
			rply_data.id = *tid;
			rply_data.master = treq->master;
			rply_data.parent = treq->parent;
			rply_data.status = TASKRPLY_STATUS_SUCCESS;

			task_inc_load();
			task_update_load();

			do_log(LOG_INFO, "Create task: Success (no VAS)");
		}

		hton_taskrply(&rply_data);
		trans_lreply(&rply_hdr, &rply_data, fd);
	}
	else {
		/* Find the minimum load daemon from VAS and create
		 * a task on it. Minimum load might be on ourselves.
		 * From here on, all task creation steps are async
		 * so theyre a little difficult to trace. */

		if (create_task_minload(treq))
			do_log(LOG_INFO, "Create task: Task async creation initiated");
		else
			do_log(LOG_ERR, "Create task: Failed");
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_CREATE_LOCALTASK: Creates a task locally. This function is
 * called by a daemon after it receives a minimum load daemon
 * address on which to run the local task. All fields of the
 * taskreq_t struct are used. This is a non-local message sent
 * to create a local task on the daemon it is sent to.
 * Expects: taskreq_t
 * Type: Transaction
 */
MSG_HANDLER(hnd_create_localtask)
{
	msghdr_t rply_hdr;
	taskrply_t rply_data;
	task_id *tid;
	taskreq_t *treq;
	char *bfile_name;
	char *file_name;

	treq = data;
	ntoh_taskreq(treq);

	rply_hdr.sender = hdr->receiver;
	rply_hdr.receiver = hdr->sender;
	rply_hdr.msg_type = MSG_CREATE_TASK_RPLY;
	rply_hdr.priority = 1;
	rply_hdr.data_size = sizeof(taskrply_t);
	rply_hdr.trans_id = hdr->trans_id;

	bfile_name = base_name(treq->file_name);
	file_name = malloc_strcat(ftp_dl_path, bfile_name);

	if (access(file_name, R_OK|X_OK)) {
		yell("Create localtask: Failed - File not readable/executable<%s>",
				file_name);
		rply_data.status = TASKRPLY_STATUS_ERR;
	}
	else {
		tid = create_local_task(file_name, treq->start_sym, &treq->master,
				&treq->parent, treq->priority);

		if (tid == NULL) {
			rply_data.status = TASKRPLY_STATUS_ERR;
			do_log(LOG_ERR, "Create localtask: Failed");
		}
		else {
			rply_data.id = *tid;
			rply_data.master = treq->master;
			rply_data.parent = treq->parent;
			rply_data.status = TASKRPLY_STATUS_SUCCESS;

			task_inc_load();
			task_update_load();

			do_log(LOG_INFO, "Create localtask: Success");
		}
	}

	hton_taskrply(&rply_data);
	trans_lreply(&rply_hdr, &rply_data, fd);

	if (bfile_name)
		free(bfile_name);
	if (file_name)
		free(file_name);

	return MSGSTAT_DEL;
}

/*
 * MSG_CREATE_TASK_RPLY: A reply sent after the creation of a
 * master of any other task. It contains the status of the task
 * creation request. The reply may convey failure.
 * Expects: taskrply_t
 * Type: Async
 */
MSG_HANDLER(hnd_create_task_rply)
{
	task_id parent;
	msghdr_t thdr;
	taskrply_t *trply = data;

	ntoh_taskrply(trply);

	parent = trply->parent;

	if (trply->status == TASKRPLY_STATUS_SUCCESS)
		task_add_child(&trply->parent, &trply->id);
	
	thdr.sender = dconf.self_id;
	thdr.receiver = dconf.self_id;
	thdr.msg_type = MSG_CREATE_TASK_RPLY;
	thdr.priority = 1;
	thdr.data_size = sizeof(taskrply_t);
	thdr.trans_id = TRANSID_NONE;

	hton_taskrply(trply);
	if (!msg_lwrite(&thdr, trply, &parent))
		do_log(LOG_CRIT, "Task create reply: Unable to forward reply to parent.");

	return MSGSTAT_DEL;
}

/*
 * MSG_IOREDIR_OUT: Message sent from a daemon running a child
 * task to the master of the child, containing the data which
 * is sent to stdout/stderr by the child. The data is fwded to
 * the ncloader in the same message format.
 * Expects: ioredir_t + stdio_data
 * Type: Async
 */
MSG_HANDLER(hnd_ioredir_out)
{
	taskinfo_t *master_inf;
	/* no need to cast ioredir_t - its in nbo */
	ioredir_t *iohdr = (ioredir_t*)data;

	master_inf = task_find(&iohdr->master);

	if (master_inf) {
		msghdr_t fwdhdr = *hdr;
		fwdhdr.sender = dconf.self_id;
		fwdhdr.trans_id = TRANSID_NONE; /* just in case */

		if (master_inf->loader_fd >= 0)
			msg_fd_write(&fwdhdr, data, master_inf->loader_fd);
	}
	else
		do_log(LOG_WARNING, "IO redir: Master not found - Data trashed!");

	return MSGSTAT_DEL;
}

/*
 * MSG_ISTERMINATE: Message sent by a timed process for each
 * master task running on the local machine. The message is
 * sent to all the children of the master and then the daemon
 * either waits for a certain period after which it assumes
 * that the user application has terminated and sends out
 * MSG_DOTERMINATE, and otherwise, if a MSG_NOTERMINATE is
 * received from any of the other machines within the timeout
 * period, it aborts the terminate procedure. The msg body
 * contains the master's task_id for which the termination is
 * being diagnosed and a list of addresses of machines thru
 * which this msg has passed.
 * Expects: taskpoll_t
 * Type: Async
 */
MSG_HANDLER(hnd_isterminate)
{
	msghdr_t fwdhdr;
	taskpoll_t *tpoll = (taskpoll_t*)data;
	
	ntoh_taskpoll(tpoll);

	/* Check if any local task has the specified
	 * master and if so, stop and send MSG_NOTERMINATE. */
	if (!task_all_terminated(&tpoll->master)) {
		fwdhdr.sender = dconf.self_id;
		fwdhdr.receiver = tpoll->master.cdaemon;
		fwdhdr.msg_type = MSG_NOTERMINATE;
		fwdhdr.priority = 1;
		fwdhdr.data_size = sizeof(task_id);
		fwdhdr.trans_id = TRANSID_NONE;

		msg_write(&fwdhdr, &tpoll->master);
	}
	else {
		taskpoll_t *fwddata;
		childinfo_t *cur;
		childinfo_t *prev;
		int i;

		/* Initialize the header and the data */

		fwdhdr.sender = dconf.self_id;
		fwdhdr.msg_type = MSG_ISTERMINATE;
		fwdhdr.priority = 1;
		fwdhdr.data_size = hdr->data_size + sizeof(daemon_id);
		fwdhdr.trans_id = TRANSID_NONE;

		fwddata = malloc(fwdhdr.data_size);
		*fwddata = *tpoll;
		fwddata->daemons[tpoll->num] = dconf.self_id;
		fwddata->num = fwddata->num + 1;
		hton_taskpoll(fwddata);

		/* Now get the list of all machines having
		 * children of the specified master which are
		 * located remotely. */

		cur = task_remote_child(&tpoll->master);

		if (cur == NULL)
			do_log(LOG_INFO, "MSG_ISTERMINATE: No child tasks found");

		while (cur) {
			for (i=0;i<tpoll->num;i++) {
				if (!memcmp(&cur->child.cdaemon, &tpoll->daemons[i],
							sizeof(daemon_id)))
					break;
			}

			if (i==tpoll->num) {
				fwdhdr.receiver = cur->child.cdaemon;
				msg_write(&fwdhdr, fwddata);

				do_log(LOG_INFO, "MSG_ISTERMINATE: MSG fwded to remote daemon");
			}

			prev = cur;
			cur = cur->next;
			free(prev);
		}

		free(fwddata);
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_NOTERMINATE: Message sent in result to MSG_ISTERMINATE
 * and if a process with the specified master is found
 * running on the local machine, this message is sent back
 * to the master machine to abort the collective termination.
 * Expects: task_id
 * Type: Async
 */
MSG_HANDLER(hnd_noterminate)
{
	task_id *master = (task_id*)data;

	if (taskpoll_cancel(master))
		do_log(LOG_INFO, "MSG_NOTERMINATE: Terminate aborted");
	else
		do_log(LOG_INFO, "MSG_NOTERMINATE: Abortion ignored - done already");

	return MSGSTAT_DEL;
}

/*
 * MSG_DOTERMINATE: On the machine originating a collective
 * terminate poll, if the timeout expires without receiving
 * any MSG_NOTERMINATE for a master task, this msg is sent
 * to all childs of the master and recursively to their
 * children, while at each step all tasks with the specified
 * master are removed from the task list. The recurive send
 * of this message is similar to that of MSG_ISTERMINATE.
 * Expects: taskpoll_t
 * Type: Async
 */
MSG_HANDLER(hnd_doterminate)
{
	msghdr_t fwdhdr;
	taskpoll_t *fwddata;
	taskpoll_t *tpoll = (taskpoll_t*)data;
	childinfo_t *cur;
	childinfo_t *prev;
	int i;
	
	tpoll = (taskpoll_t*)data;
	ntoh_taskpoll(tpoll);

	/* Initialize the header and the data */

	fwdhdr.sender = dconf.self_id;
	fwdhdr.msg_type = MSG_DOTERMINATE;
	fwdhdr.priority = 1;
	fwdhdr.data_size = hdr->data_size + sizeof(daemon_id);
	fwdhdr.trans_id = TRANSID_NONE;

	fwddata = malloc(fwdhdr.data_size);
	*fwddata = *tpoll;
	fwddata->daemons[tpoll->num] = dconf.self_id;
	fwddata->num = fwddata->num + 1;
	hton_taskpoll(fwddata);

	/* Now abort all the local tasks of the specified
	 * master and then traverse the list of all the
	 * remote children of this master. */

	cur = task_abort(&tpoll->master);

	if (cur == NULL)
		do_log(LOG_INFO, "MSG_DOTERMINATE: No child tasks found");

	while (cur) {
		for (i=0;i<tpoll->num;i++) {
			if (!memcmp(&cur->child.cdaemon, &tpoll->daemons[i],
						sizeof(daemon_id)))
				break;
		}

		/* If cur child not already sent the terminate msg */
		if (i==tpoll->num) {
			fwdhdr.receiver = cur->child.cdaemon;
			msg_write(&fwdhdr, fwddata);

			do_log(LOG_INFO, "MSG_DOTERMINATE: MSG fwded to remote daemon");
		}

		prev = cur;
		cur = cur->next;
		free(prev);
	}

	free(fwddata);

	return MSGSTAT_DEL;
}

/*
 * DAEMON FTP MESSAGE HANDLERS
 */

/*
 * MSG_FTP_INIT: The MSG_FTP_INIT is sent from the daemon which
 * needs to send the file to the receiving daemon. The receiver
 * could either ignore this request (it would timeout at the
 * sender), or he could reply with a MSG_FTP_REQ, requesting a
 * file data transfer.
 * Expects: ftpinit_t + filename(asciiz)
 * Type: Async
 */
MSG_HANDLER(hnd_ftp_init)
{
	int ftpstat;
	msghdr_t msghdr;
	ftpid_t ftp_id;
	ftpinit_t *finit = (ftpinit_t*)data;
	char *fname = ((char*)data)+sizeof(ftpinit_t);

	ntoh_ftpinit(finit);

	if (!fname || !(*fname)) {
		yell("FTP init: Invalid file name in request"); 
		return MSGSTAT_DEL;
	}

	msghdr.sender = hdr->receiver;
	msghdr.receiver = hdr->sender;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(ftpid_t);
	msghdr.trans_id = TRANSID_NONE;
	ftp_id = finit->ftp_id;

	ftpstat = ftp_recv(&finit->ftp_id, fname, finit->file_size, finit->md5,
			NULL, NULL);
	if (ftpstat < 0) {
		yell("FTP init: Unable to create operation");
		return MSGSTAT_DEL;
	}
	else if (ftpstat == 0) {
		msghdr.msg_type = MSG_FTP_DEINIT;
		hton_ftpid(&ftp_id);
		msg_reply(&msghdr, &ftp_id, fd);
		do_log(LOG_INFO, "FTP init: Success. File already found. Send ignored<%s>",
				fname);
	}
	else {
		msghdr.msg_type = MSG_FTP_REQ;
		hton_ftpid(&ftp_id);
		msg_reply(&msghdr, &ftp_id, fd);
		do_log(LOG_INFO, "FTP init: Success. Ready to receieve file<%s>",
				fname);

	}

	return MSGSTAT_DEL;
}

/*
 * MSG_FTP_REQ: This is sent in response to MSG_FTP_INIT and MSG_FTP_DATA
 * to request data from the sender.
 * Expects: ftpid_t
 * Type: Async
 */
MSG_HANDLER(hnd_ftp_req)
{
	ftpinfo_t *finf;
	ftpid_t *ftp_id = (ftpid_t*)data;

	ntoh_ftpid(ftp_id);

	finf = ftp_get_send(ftp_id);
	if (!finf) {
		yell("FTP req: No record of ftp operation");
		return MSGSTAT_DEL;
	}

	/* the ftp_send_block function sends a MSG_FTP_DATA msg */
	switch (ftp_send_block(ftp_id, fd)) {
		case -1:
			/* Call the event handler for failure */
			finf->status = FTPSTAT_FAIL;
			if (finf && finf->hnd_event)
				(*finf->hnd_event)(finf, finf->user_data);

			ftp_remove_send(ftp_id);
			yell("FTP req: Unable to send data.");
			break;
		case 0:
			/* The file transfer is complete but the
			 * handler would be called for the COMPLETE
			 * event only after the confirmation is
			 * received thru MSG_FTP_DEINIT. */
			break;
		default:
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_FTP_DATA: This is sent in response to MSG_FTP_REQ and it includes
 * the file data. This handler receives the data and writes it to the
 * local file.
 * Expects: ftpid_t + raw_binary_file_data
 * Type: Async
 */
MSG_HANDLER(hnd_ftp_data)
{
	msghdr_t msghdr;
	ftpinfo_t *finf;
	ftpid_t ftp_id;
	ftpid_t *msg_ftp_id = (ftpid_t*)data;
	char *msg_data = (char*)data + sizeof(ftpid_t);
	unsigned long msg_data_size = hdr->data_size - sizeof(ftpid_t);

	ntoh_ftpid(msg_ftp_id);

	/* Prepare the reply msghdr */
	msghdr.sender = hdr->receiver;
	msghdr.receiver = hdr->sender;
	msghdr.priority = 1;
	msghdr.data_size = sizeof(ftpid_t);
	msghdr.trans_id = TRANSID_NONE;

	finf = ftp_get_recv(msg_ftp_id);
	if (!finf) {
		yell("FTP data: No record of ftp operation");
		return MSGSTAT_DEL;
	}

	switch (ftp_recv_block(msg_ftp_id, msg_data, msg_data_size)) {
		case -1:
			/* Call the event handler for failure */
			finf->status = FTPSTAT_FAIL;
			if (finf && finf->hnd_event)
				(*finf->hnd_event)(finf, finf->user_data);

			ftp_remove_recv(msg_ftp_id);
			yell("FTP data: Unable to receive data.");
			return MSGSTAT_DEL;
		case 0:
			finf->status = FTPSTAT_COMPLETE;
			if (finf && finf->hnd_event)
				(*finf->hnd_event)(finf, finf->user_data);
			ftp_remove_recv(msg_ftp_id);
			do_log(LOG_INFO, "FTP data: File recv completed");

			msghdr.msg_type = MSG_FTP_DEINIT;
			break;
		default:
			msghdr.msg_type = MSG_FTP_REQ;
			break;
	}

	ftp_id = *msg_ftp_id;
	hton_ftpid(&ftp_id);
	msg_reply(&msghdr, &ftp_id, fd);

	return MSGSTAT_DEL;
}

/*
 * MSG_FTP_DEINIT: This msg is sent by the file receiver to the sender
 * and it signifies that the file send is complete. Its a confirmation
 * received after sending the last block of bytes.
 * In some cases, this msg would be sent directly in response to
 * MSG_FTP_INIT, if the file is already present with the receiver.
 * Expects: ftpid_t
 * Type: Async
 */
MSG_HANDLER(hnd_ftp_deinit)
{
	ftpinfo_t *finf;
	ftpid_t *msg_ftp_id = (ftpid_t*)data;

	ntoh_ftpid(msg_ftp_id);

	finf = ftp_get_send(msg_ftp_id);

	if (!finf) {
		yell("FTP deinit: No record of ftp operation");
		return MSGSTAT_DEL;
	}

	/* Call the event handler for completion */
	finf->status = FTPSTAT_COMPLETE;
	if (finf && finf->hnd_event)
		(*finf->hnd_event)(finf, finf->user_data);

	/* Remove the operation from the send list */
	ftp_remove_send(msg_ftp_id);
	do_log(LOG_INFO, "FTP deinit: File send completed");

	return MSGSTAT_DEL;
}

/*
 * MSG_RPC_CALL: This is the message sent by the RPC caller task
 * to its controlling daemon to the callee's controlling daemon
 * (where it is queued until an RPC accept is aknowledged by the
 * callee) and finally to the callee. The initial send to the
 * controlling daemon need not specify rpcinfo_t.src, which is
 * filled in here before forwarding the msg.
 * Expects: rpcinfo_t callret_arguments
 * Type: Async
 */
MSG_HANDLER(hnd_rpc_call)
{
	msghdr_t fwdhdr;
	rpcinfo_t *rpcinf = data;
	taskinfo_t *tinf;

	ntoh_rpcinfo(rpcinf);

	fwdhdr.sender = dconf.self_id;
	fwdhdr.receiver = rpcinf->dest.cdaemon;
	fwdhdr.msg_type = MSG_RPC_CALL;
	fwdhdr.priority = 1;
	fwdhdr.data_size = hdr->data_size;
	fwdhdr.trans_id = TRANSID_NONE;

	/* get task info of the caller task */
	tinf = task_find_msgio(fd);
	if (tinf) {
		/* Fill in unfilled fields of local msg */
		rpcinf->status = RPC_STATUS_PENDING;
		rpcinf->src = tinf->id;
	}

	if (!memcmp(&rpcinf->dest.cdaemon, &dconf.self_id, sizeof(daemon_id))) {
		/* The local or non-local msg has arrived on
		 * the destination machine. Now check if any
		 * task waiting on MSG_RPC_CALL/funcid */

		taskinfo_t *cur = task_list;

		while (cur) {
			tstate_wait_t *waitinf;

			if (cur->state != TSTATE_WAIT || !cur->state_data) {
				cur = cur->next;
				continue;
			}

			waitinf = cur->state_data;

			if (waitinf->msg_type == MSG_RPC_CALL) {
				rpcaccept_t *rpcacc;
				rpcacc = (rpcaccept_t*)waitinf->data;

				if (rpcacc->func_id == rpcinf->func_id) {
					/* Found a task waiting for this RPC */
					hton_rpcinfo(rpcinf);
					msg_lwrite(&fwdhdr, data, &cur->id);

					task_setstate_execute(cur);
					return MSGSTAT_DEL;
				}
			}
			
			cur = cur->next;
		}

		/* No matching rpc_accept found. Queue msg. */
		return MSGSTAT_WAIT;
	}
	else {
		/* Local msg on src machine. Fwd it to dest. */

		if (!tinf) {
			do_log(LOG_CRIT, "RPC call: Unable to retrieve taskinfo of caller");
			return MSGSTAT_DEL;
		}

		hton_rpcinfo(rpcinf);
		msg_write(&fwdhdr, data);
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_RPC_RPLY: This is the reply to an RPC call. The caller
 * blocks until it received this message by the callee. It
 * contains the return type data of the user function/procedure
 * if any, in the same format as the one used in MSG_RPC_CALL
 * for passing function arguments.
 * Expects: rpcinfo_t callret_arguments
 * Type: Async
 */
MSG_HANDLER(hnd_rpc_rply)
{
	task_id src_task;
	taskinfo_t *tinf;
	msghdr_t fwdhdr;
	rpcinfo_t *rpcinf = data;

	ntoh_rpcinfo(rpcinf);

	if (!memcmp(&rpcinf->src.cdaemon, &dconf.self_id, sizeof(daemon_id))) {
		/* Msg has arrived back on the src machine
		 * so should now be fwded to the local task. */

		src_task = rpcinf->src;

		fwdhdr.sender = dconf.self_id;
		fwdhdr.receiver = dconf.self_id;
		fwdhdr.msg_type = MSG_RPC_RPLY;
		fwdhdr.priority = 1;
		fwdhdr.data_size = hdr->data_size;
		fwdhdr.trans_id = TRANSID_NONE;

		hton_rpcinfo(rpcinf);
		msg_lwrite(&fwdhdr, data, &src_task);
	}
	else {
		/* Msg is on the dest callee machine so fwd
		 * it to the caller machine. */

		fwdhdr.sender = dconf.self_id;
		fwdhdr.receiver = rpcinf->src.cdaemon;
		fwdhdr.msg_type = MSG_RPC_RPLY;
		fwdhdr.priority = 1;
		fwdhdr.data_size = hdr->data_size;
		fwdhdr.trans_id = TRANSID_NONE;

		hton_rpcinfo(rpcinf);
		msg_write(&fwdhdr, data);
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_RPC_ACCEPT: This is the aknowledgment of an RPC call.
 * It first checks whether the requested RPC function ID call
 * it already available in the message queue. If found, the
 * call is fwded to the task, otherwise the status of the
 * task is changed to waiting and it blocks until a call of
 * the specified type arrives.
 * Expects: rpcaccept_t
 * Type: Async
 */
MSG_HANDLER(hnd_rpc_accept)
{
	taskinfo_t *tinf;
	tstate_wait_t waitinf;
	rpcaccept_t *new_rpcacc;
	rpcaccept_t *rpcacc = data;
	clist *cur_msg = in_msgs;

	ntoh_rpcaccept(rpcacc);

	tinf = task_find_msgio(fd);

	if (!tinf) {
		do_log(LOG_CRIT, "RPC accept: Non-local accept msg. Ignored...");
		return MSGSTAT_DEL;
	}

	/* Find rpc call msg and fwd it to tinf->id */
	while (cur_msg) {
		msginfo_t *msginf = (msginfo_t*) cur_msg->data;
		msghdr_t *msghdr = (msghdr_t*) msginf->hdr;

		if (msginf->status == MSGSTAT_WAIT &&
				msghdr->msg_type == MSG_RPC_CALL)
		{
			/* rpcinf is already in host byte order */
			rpcinfo_t *rpcinf = (rpcinfo_t*)msginf->data;

			if (rpcinf->func_id == rpcacc->func_id) {
				msghdr_t fwdhdr;

				fwdhdr.sender = dconf.self_id;
				fwdhdr.receiver = dconf.self_id;
				fwdhdr.msg_type = MSG_RPC_CALL;
				fwdhdr.priority = 1;
				fwdhdr.data_size = msghdr->data_size;
				fwdhdr.trans_id = TRANSID_NONE;

				hton_rpcinfo(rpcinf);
				msg_lwrite(msghdr, msginf->data, &tinf->id);

				msginf->status = MSGSTAT_DEL;
				return MSGSTAT_DEL;
			}
		}

		cur_msg = cur_msg->next;
	}

	/* No matching RPC call found. Change task state
	 * state to wait until required call arrives. But
	 * only if accept is blocking. Otherwise reply with
	 * RPC_STATUS_PENDING. */

	if (rpcacc->is_blocking) {
		new_rpcacc = (rpcaccept_t*)malloc(sizeof(rpcaccept_t));
		*new_rpcacc = *rpcacc;
		waitinf.msg_type = MSG_RPC_CALL;
		waitinf.data = new_rpcacc;

		task_setstate_wait(tinf, &waitinf);
	}
	else {
		msghdr_t rplyhdr;
		rpcinfo_t rplyrpc;

		rplyhdr.sender = dconf.self_id;
		rplyhdr.receiver = dconf.self_id;
		rplyhdr.msg_type = MSG_RPC_CALL;
		rplyhdr.priority = 1;
		rplyhdr.data_size = sizeof(rplyrpc);
		rplyhdr.trans_id = TRANSID_NONE;

		rplyrpc.status = RPC_STATUS_NOBLOCK;
		TASK_ID_NULL(rplyrpc.src);
		TASK_ID_NULL(rplyrpc.dest);
		rplyrpc.func_id = rpcacc->func_id;

		hton_rpcinfo(&rplyrpc);
		msg_lwrite(&rplyhdr, &rplyrpc, &tinf->id);
	}

	return MSGSTAT_DEL;
}

/*
 * MSG_MY_TASKID: This msg is sent by a local task to its
 * controlling daemon. The daemon replys back with the same
 * msg type along with the task_id of the calling task as
 * the msg data.
 * Expects: none
 * Type: Async
 */
MSG_HANDLER(hnd_my_taskid)
{
	msghdr_t fwdhdr;
	taskinfo_t *tinf;

	/* get task info of the caller task */
	tinf = task_find_msgio(fd);

	if (!tinf) {
		do_log(LOG_CRIT, "MSG_MY_TASKID from non-local task");
		return MSGSTAT_DEL;
	}

	fwdhdr.sender = dconf.self_id;
	fwdhdr.receiver = dconf.self_id;
	fwdhdr.msg_type = MSG_MY_TASKID;
	fwdhdr.priority = 1;
	fwdhdr.data_size = sizeof(task_id);
	fwdhdr.trans_id = TRANSID_NONE;

	msg_lwrite(&fwdhdr, &tinf->id, &tinf->id);

	return MSGSTAT_DEL;
}

/*
 * MSG_TERMINATE: This msg is sent locally by a task to the
 * controlling daemon. The task ID of the exitinfo struct can
 * be skipped optionally. The task is removed from the daemon
 * task list, and terminated if its still running, with the
 * KILL signal. This message can only be sent locally.
 * Expects: terminate_t
 * Type: Async
 */
MSG_HANDLER(hnd_terminate)
{
	msghdr_t fwdhdr;
	taskinfo_t *tinf;
	exitinfo_t *einf = data;

	ntoh_exitinfo(einf);

	/* get task info of the caller task */
	tinf = task_find_msgio(fd);

	if (!tinf) {
		do_log(LOG_CRIT, "MSG_TERMINATE from non-local task");
		return MSGSTAT_DEL;
	}

	if (!task_terminate(&tinf->id, einf->status))
		do_log(LOG_CRIT, "MSG_TERMINATE: Unable to terminate task");

	return MSGSTAT_DEL;
}

/*
 * MSG_ABORT: Sending this message results in a complete
 * abort. The msg is forwarded to the master which then
 * terminates itself and its direct decendents, whom inturn
 * terminate their decendents and so on. The tid field
 * contains the ID of the master. Sender expects no reply.
 * Expects: exitinfo_t
 * Type: Async
 */
MSG_HANDLER(hnd_abort)
{
	childinfo_t *cur;
	childinfo_t *prev;
	msghdr_t fwdhdr;
	exitinfo_t fwddata;
	taskinfo_t *tinf;
	exitinfo_t *einf = data;

	ntoh_exitinfo(einf);

	fwdhdr.sender = dconf.self_id;
	fwdhdr.msg_type = MSG_ABORT;
	fwdhdr.priority = 1;
	fwdhdr.data_size = sizeof(exitinfo_t);
	fwdhdr.trans_id = TRANSID_NONE;

	/* get task info of the caller task */
	tinf = task_find_msgio(fd);

	if (tinf) {
		einf->tid = tinf->master;
		fwddata = *einf;
		hton_exitinfo(&fwddata);

		if (memcmp(&einf->tid.cdaemon, &dconf.self_id, sizeof(daemon_id))) {
			fwdhdr.receiver = tinf->master.cdaemon;
			msg_write(&fwdhdr, &fwddata);
		}
	}
	else {
		fwddata = *einf;
		hton_exitinfo(&fwddata);
	}

	/* First terminate all tasks with einf->tid as master
	 * and then forward this message to all remote children
	 * of the tasks which were locally terminated. */

	cur = task_abort(&einf->tid);

	if (cur == NULL)
		yell("MSG_ABORT: No child tasks found");

	while (cur) {
		fwdhdr.receiver = cur->child.cdaemon;
		msg_write(&fwdhdr, &fwddata);

		yell("MSG_ABORT: MSG fwded to remote daemon");

		prev = cur;
		cur = cur->next;
		free(prev);
	}

	return MSGSTAT_DEL;
}

#endif /* NCDAEMON */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * VECTOR FOR MESSAGE HANDLERS
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Initialize the message handlers to the above functions */
msghandler_t msg_handlers[NUMBER_OF_HANDLERS] = {
	{ hnd_invalid },		/* msg_type = 0 */
	{ hnd_ping },			/* msg_type = 1 */
	{ hnd_pong },			/* msg_type = 2 */
	{ hnd_syslog },			/* msg_type = 3 */
	{ NULL },			/* msg_type = 4 */
	{ NULL },			/* msg_type = 5 */
	{ NULL },			/* msg_type = 6 */
	{ NULL },			/* msg_type = 7 */
	{ NULL },			/* msg_type = 8 */
	{ NULL },			/* msg_type = 9 */
	{ NULL },			/* msg_type = 10 */
	{ NULL },			/* msg_type = 11 */
	{ NULL },			/* msg_type = 12 */
	{ NULL },			/* msg_type = 13 */
	{ NULL },			/* msg_type = 14 */
	{ NULL },			/* msg_type = 15 */
	{ NULL },			/* msg_type = 16 */
	{ NULL },			/* msg_type = 17 */
	{ NULL },			/* msg_type = 18 */
	{ NULL },			/* msg_type = 19 */

#ifdef NCVAS
	{ hnd_vas_pong },		/* msg_type = 20 */
	{ hnd_vas_reg },		/* msg_type = 21 */
	{ hnd_vas_unreg },		/* msg_type = 22 */
	{ hnd_vas_isreg },		/* msg_type = 23 */
	{ hnd_vas_setload },		/* msg_type = 24 */
	{ hnd_vas_getload },		/* msg_type = 25 */
	{ hnd_vas_getminload },		/* msg_type = 26 */
	{ hnd_vas_display },		/* msg_type = 27 */
	{ NULL },			/* msg_type = 28 */
	{ NULL },			/* msg_type = 29 */
#else
	{ NULL },			/* msg_type = 20 */
	{ NULL },			/* msg_type = 21 */
	{ NULL },			/* msg_type = 22 */
	{ NULL },			/* msg_type = 23 */
	{ NULL },			/* msg_type = 24 */
	{ NULL },			/* msg_type = 25 */
	{ NULL },			/* msg_type = 26 */
	{ NULL },			/* msg_type = 27 */
	{ NULL },			/* msg_type = 28 */
	{ NULL },			/* msg_type = 29 */
#endif

#ifdef NCDAEMON
	{ hnd_vas_ping },		/* msg_type = 30 */
	{ hnd_vas_reg_rply },		/* msg_type = 31 */
	{ hnd_vas_isreg_rply },		/* msg_type = 32 */
	{ hnd_vas_getload_rply },	/* msg_type = 33 */
	{ hnd_vas_getminload_rply },	/* msg_type = 34 */
	{ hnd_create_master },		/* msg_type = 35 */
	{ hnd_create_task },		/* msg_type = 36 */
	{ hnd_create_localtask },	/* msg_type = 37 */
	{ hnd_create_task_rply },	/* msg_type = 38 */
	{ hnd_ioredir_out },		/* msg_type = 39 */
	{ hnd_isterminate },		/* msg_type = 40 */
	{ hnd_noterminate },		/* msg_type = 41 */
	{ hnd_doterminate },		/* msg_type = 42 */
	{ NULL },			/* msg_type = 43 */
	{ NULL },			/* msg_type = 44 */
	{ NULL },			/* msg_type = 45 */
	{ hnd_ftp_init },		/* msg_type = 46 */
	{ hnd_ftp_req },		/* msg_type = 47 */
	{ hnd_ftp_data },		/* msg_type = 48 */
	{ hnd_ftp_deinit },		/* msg_type = 49 */
	{ hnd_rpc_call },		/* msg_type = 50 */
	{ hnd_rpc_rply },		/* msg_type = 51 */
	{ hnd_rpc_accept },		/* msg_type = 52 */
	{ hnd_my_taskid },		/* msg_type = 53 */
	{ hnd_terminate },		/* msg_type = 54 */
	{ hnd_abort },			/* msg_type = 55 */
	{ NULL },			/* msg_type = 56 */
	{ NULL },			/* msg_type = 57 */
	{ NULL },			/* msg_type = 58 */
	{ NULL }			/* msg_type = 59 */
#else
	{ NULL },			/* msg_type = 30 */
	{ NULL },			/* msg_type = 31 */
	{ NULL },			/* msg_type = 32 */
	{ NULL },			/* msg_type = 33 */
	{ NULL },			/* msg_type = 34 */
	{ NULL },			/* msg_type = 35 */
	{ NULL },			/* msg_type = 36 */
	{ NULL },			/* msg_type = 37 */
	{ NULL },			/* msg_type = 38 */
	{ NULL },			/* msg_type = 39 */
	{ NULL },			/* msg_type = 40 */
	{ NULL },			/* msg_type = 41 */
	{ NULL },			/* msg_type = 42 */
	{ NULL },			/* msg_type = 43 */
	{ NULL },			/* msg_type = 44 */
	{ NULL },			/* msg_type = 45 */
	{ NULL },			/* msg_type = 46 */
	{ NULL },			/* msg_type = 47 */
	{ NULL },			/* msg_type = 48 */
	{ NULL },			/* msg_type = 49 */
	{ NULL },			/* msg_type = 50 */
	{ NULL },			/* msg_type = 51 */
	{ NULL },			/* msg_type = 52 */
	{ NULL },			/* msg_type = 53 */
	{ NULL },			/* msg_type = 54 */
	{ NULL },			/* msg_type = 55 */
	{ NULL },			/* msg_type = 56 */
	{ NULL },			/* msg_type = 57 */
	{ NULL },			/* msg_type = 58 */
	{ NULL }			/* msg_type = 59 */
#endif
};
