/*
 * msgs.c: Message queuing and invocation of handlers
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>

#include "defs.h"
#include "sock.h"
#include "io.h"
#include "log.h"
#include "msgs.h"
#include "transact.h"



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                M e s s a g i n g   I n t e r f a c e
 *
 * The following are functions for message handling which includes
 * message queuing and dispatching. Also includes other utility functions.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * msg_free_deletable: Free the messages which have their status
 * set to MSGSTAT_DEL. Returns the number of messages free'd.
 */
int msg_free_deletable()
{
	msginfo_t *msginf;
	int ndel = 0;
	clist *cur_msg = in_msgs;
	clist *tmp_msg;

	while (cur_msg) {
		tmp_msg = cur_msg;
		cur_msg = cur_msg->next;
		msginf = (msginfo_t *)tmp_msg->data;

		if (msginf->status == MSGSTAT_DEL) {
			free(msginf->hdr);
			free(msginf);

			/* Remove message from the msg queue */
			in_msgs = clist_remove_link(in_msgs, tmp_msg);
			ndel++;
		}
	}

	return ndel;
}

/*
 * msg_enqueue: Add a message to the incoming messages queue. Returns
 * TRUE on success and FALSE on failure. The ptr to the msg passed
 * must be dynamically allocated. The msgs module is responsible for
 * freeing this memory. The class is used to differentiate between
 * local and remote messages.
 */
int msg_enqueue(msghdr_t *msg, int fd, int mclass)
{
	msginfo_t *msginf;

	msginf = (msginfo_t *)malloc(sizeof(msginfo_t));

	msginf->hdr = msg;
	msginf->data = (void*) MSG_DATA(msg);
	msginf->ttl = 0;
	msginf->fd = fd;
	msginf->mclass = mclass;
	msginf->status = MSGSTAT_UNDELIV;

	in_msgs = clist_append(in_msgs, msginf);

	return TRUE;
}

/*
 * msg_dispatch: Dispatches messages to their corresponding handler
 * functions and once handled, removes them from the queue. Returns
 * the number of messages successfully delivered. A handler may choose
 * to keep the message queued for later handling. In such a case it
 * would be redelivered. All messages with unknown type and sent to
 * the invalid msg handler.
 *
 * Note that each message handler is passed an 'fd' from which the
 * message arrived. This may only be valid for the first time that
 * a msg handler is called coz by the next time, the connection
 * manager might have closed this fd. Therefore it is set to -1 after
 * the first call.
 */
int msg_dispatch()
{
	msginfo_t *msginf;
	msghdr_t *hdr;
	msghandler_t *mhand;
	int msg_type;
	int newstat;
	int msgsent = 0;
	clist *cur_msg = in_msgs;
	transact_t *trans;

	while (cur_msg) {
		msginf = (msginfo_t *) cur_msg->data;
		hdr = msginf->hdr;
		msg_type = hdr->msg_type;
		trans = NULL;

		if (msginf->status != MSGSTAT_UNDELIV) {
			cur_msg = cur_msg->next;
			continue;
		}

		if (msg_type < 0 || msg_type >= NUMBER_OF_HANDLERS ||
				msg_handlers[msg_type].func==NULL)
			msg_type = MSG_INVALID;
		else if (hdr->trans_id != TRANSID_NONE && hdr->trans_isrply) {
			
			/* if the msg is valid and is a transaction's reply
			 * then fill in 'trans' to be passed to the handler.
			 * and if it is not found then it must have timedout
			 * so give a msg and take appropriate action. */
			trans = trans_get(hdr->trans_id);

			if (!trans) {
				do_log(LOG_NOTICE, "Received transaction reply after timeout: trans_id<%d> - invalidated!",
						hdr->trans_id);
				msg_type = MSG_INVALID;
			}
			else
				trans->status = TRANS_STATUS_SUCCESS;
		}

		mhand = &msg_handlers[msg_type];

		/* Invoke the message handler for this message */
		if (mhand->func) {
			newstat = mhand->func(msginf->hdr, msginf->data,
					msginf->fd, trans);
			msginf->status = newstat;
			msginf->fd = -1; /* expire the fd after 1st call */
		}

		/* Remove the transaction if completed successfully */
		if (trans)
			trans_remove(trans->trans_id);
		
		cur_msg = cur_msg->next;
	}

	/* Free the messages which have been marked as MSGSTAT_DEL */
	msg_free_deletable();

	return TRUE;
}
