/*
 * sock.c: Generic sockets interface
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>

#include "defs.h"
#include "sock.h"

/* global fd_set for readable and writable sockets */
fd_set fd_ractive;
fd_set fd_wactive;
int global_max_fd = -1;

/*
 * make_sockaddr_in: Returns a sockaddr for AF_INET to the specified
 * host:port. If unable to resolve the hostname, terminates the program.
 */
struct sockaddr_in make_sockaddr_in(char *hostname, int port)
{
	struct sockaddr_in saddr;
	struct hostent *hostinfo;

	hostinfo = gethostbyname(hostname);
	if (hostinfo == NULL) {
		fprintf(stderr, "Unable to resolve hostname %s\n", hostname);
		exit(EXIT_FAILURE);
	}
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(port);
	saddr.sin_addr = *(struct in_addr *)hostinfo->h_addr;
	return saddr;
}

/*
 * make_socket: Makes a connection to the specified host:port and
 * returns the socket of the connection.
 * Returns -1 on error.
 */ 
int make_socket(char *hostname, int port)
{
	int sock;
	struct sockaddr_in saddr;

	sock = socket(PF_INET, SOCK_STREAM, 0);
	saddr = make_sockaddr_in(hostname, port);
	
	if (connect(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
#ifdef DEBUG
		perror("Error connecting");
#endif
		return -1;
	}

	return sock;
}

/*
 * make_listen_socket: Makes a socket for listening on the specified
 * port. This i used by servers for creating a listening socket.
 * Returns -1 on error.
 */
int make_listen_socket(int port)
{
	int sock;
	struct sockaddr_in saddr;

	sock = socket(PF_INET, SOCK_STREAM, 0);

	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(port);
	saddr.sin_addr.s_addr = htonl(INADDR_ANY);

	if (bind(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
#ifdef DEBUG
		perror("Error binding socket");
#endif
		return -1;
	}

	if (listen(sock, 1) < 0) {
#ifdef DEBUG
		perror("Listen");
#endif
		close(sock);
		return -1;
	}

	return sock;
}

/*
 * make_socketpair: Makes a socketpair and stores it in the given
 * argument. Returns TRUE on success and FALSE on failure.
 */
int make_socketpair(int sv[2])
{
	if (socketpair(PF_LOCAL, SOCK_STREAM, 0, sv))
		return FALSE;

	return TRUE;
}

/*
 * make_udp_socket: Makes a UDP socket which sends packages to the
 * specified host:port by default. On error, prints and error message
 * and terminates the program.
 */
int make_udp_socket(char *hostname, int port)
{
	int sock;
	struct sockaddr_in saddr;

	sock = socket(PF_INET, SOCK_DGRAM, 0);
	saddr = make_sockaddr_in(hostname, port);
	
	if (connect(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
#ifdef DEBUG
		perror("Error connecting socket");
#endif
		return -1;
	}

	return sock;
}

/*
 * make_udp_listen_socket: Makes a socket for listening on the
 * specified port for UDP packets. Returns -1 on error.
 */
int make_udp_listen_socket(int port)
{
	int sock;
	struct sockaddr_in saddr;

	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(port);
	saddr.sin_addr.s_addr = htonl(INADDR_ANY);

	sock = socket(PF_INET, SOCK_DGRAM, 0);

	if (bind(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
#ifdef DEBUG
		perror("Error binding socket");
#endif
		return -1;
	}

	if (listen(sock, 1) < 0) {
#ifdef DEBUG
		perror("Listen");
#endif
		close(sock);
		return -1;
	}

	return sock;
}
