/*
 * task.c: handles the task list and other task related operations
 * (C) Copyright 2003, netcluster
 *
 * Written by Kashif Jamal
 * Task management support by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include "defs.h"
#include "structs.h"
#include "io.h"
#include "msgs.h"
#include "transact.h"
#include "log.h"
#include "task.h"
#include "util.h"
#include "ioredir.h"
#include "fileman.h"


/* the list of all local tasks */
taskinfo_t *task_list;


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * TASK LIST MANAGEMENT
 * Below are the functions used to handle the task list. It includes
 * adding, removing and searching the list. The child task list is also
 * managed by this functions.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* task_free_item: Utility function. Frees a single given element
 * of the task list. Frees all malloc'ed elements in the item.  */
void task_free_item( taskinfo_t *tsk ) {
	if (tsk) {
		childinfo_t *tmp;
		childinfo_t *chd = tsk->child_list;

		/* free the child list */
		while (chd) {
			tmp = chd;
			chd = chd->next;
			free(tmp);
		}

		if (tsk->sym_name)
			free(tsk->sym_name);
		if (tsk->start_sym)
			free(tsk->start_sym);
		if (tsk->file_name)
			free(tsk->file_name);
		free(tsk);
		tsk = NULL;
	}
}

/* task_copy_item: A utility function to make a copy of the task */
taskinfo_t *task_copy_item( taskinfo_t *dest, taskinfo_t *src ) {
	memcpy(dest, src, sizeof(taskinfo_t));
	if (src->sym_name)
		dest->sym_name = (char*) strdup(src->sym_name);
	if (src->start_sym)
		dest->start_sym = (char*) strdup(src->start_sym);
	if (src->file_name)
		dest->file_name = (char*) strdup(src->file_name);
	return dest;
}

/* task_free_list: Free the whole task list */
int task_free_list() {
	taskinfo_t *cur = task_list;
	while( cur != NULL ) {
		taskinfo_t *prev = cur;
		cur = cur ->next;
		task_free_item( prev );
	}
	task_list = NULL;
	return TRUE;
}

/* task_add: Returns TRUE on success FALSE on failure. Takes the
 * given new_task and appends it to the list as is. The new_task
 * must be dynamically allocated by the caller. */
int task_add( taskinfo_t *new_task ) {
	new_task->next = NULL;

	if( task_list == NULL ) {
		task_list = new_task;
	}
	else {
		taskinfo_t *cur = task_list;
		while( cur->next != NULL )
			cur = cur->next;
		cur->next = new_task;
	}

	return TRUE;
}

/* task_remove: Returns TRUE if successfully removed or FALSE
 * if the given task_id was not found in the list. */
int task_remove( task_id *tid ) {
	taskinfo_t *cur;
	taskinfo_t *prev;

	if (!task_list)
		return FALSE;

	if( !memcmp(tid, &task_list->id, sizeof(task_id)) ) { /* head is to be removed */
		prev = task_list;
		task_list = task_list->next;
		task_free_item( prev );
		return TRUE;
	}

	cur = task_list;
	while( cur->next != NULL ) {
		prev = cur;
		cur = cur->next;
		if( !memcmp(&cur->id, tid, sizeof(task_id)) ) {
			prev->next = cur->next;
			task_free_item( cur );
			return TRUE;
		}
	}
	
	return FALSE;
}

/* task_find: Returns NULL if no task is found */
taskinfo_t *task_find( task_id *tid ) {
	taskinfo_t *cur = task_list;
	while( cur != NULL ) {
		if( !memcmp(&cur->id,tid,sizeof(task_id)) )
			break;
		else
			cur = cur->next;
	}
	return cur;
}

/* task_find_name: Returns NULL if no task is found */
taskinfo_t *task_find_name( char *name ) {
	taskinfo_t *cur = task_list;
	while( cur != NULL ) {
		int result = strcmp( name, cur->sym_name );
		if( result == 0 )
			break;
		else
			cur = cur->next;
	}
	return cur;
}

/* task_find_stdio: Searches the tasklist for a task who's stdio,
 * both daemon and task, fds match the given fd. Returns the task
 * which matches or NULL if not found. */
taskinfo_t *task_find_stdio( int fd )
{
	taskinfo_t *cur = task_list;
	while ( cur != NULL ) {
		if (fd == cur->std_io[0] || fd == cur->std_io[1])
			break;
		else
			cur = cur->next;
	}
	return cur;
}

/* task_find_stdio: Searches the tasklist for a task who's stdio,
 * both daemon and task, fds match the given fd. Returns the task
 * which matches or NULL if not found. */
taskinfo_t *task_find_msgio( int fd )
{
	taskinfo_t *cur = task_list;
	while ( cur != NULL ) {
		if (fd == cur->msg_io[0] || fd == cur->msg_io[1])
			break;
		else
			cur = cur->next;
	}
	return cur;
}

/* task_find_child: Returns the parent to whom this child belongs.
 * Returns NULL if no task is found. */
taskinfo_t *task_find_child( task_id *child )
{
	taskinfo_t *cur = task_list;

	while (cur) {
		childinfo_t *curchd = cur->child_list;

		while (curchd) {
			if (!memcmp(&curchd->child, child, sizeof(task_id)))
				return cur;
			curchd = curchd->next;
		}

		cur = cur->next;
	}

	return cur;
}

/* task_all_terminated: Searches the task list and if any task
 * with the specified master isnt TERMINATED yet, it returns
 * FALSE and if all local tasks of this master have terminated
 * TRUE is returned. */
int task_all_terminated(task_id *master)
{
	taskinfo_t *cur = task_list;
	while( cur != NULL ) {
		if ( !memcmp(&cur->master,master,sizeof(task_id)) &&
				cur->state != TSTATE_TERMINATE )
			return FALSE;
		else
			cur = cur->next;
	}
	return TRUE;
}

/* task_remove_terminated: Removes all tasks from this machine
 * with the given master task and which have all terminated.
 * Unterminated tasks are also sent the kill signal before
 * removal from the task list. */
int task_remove_terminated( task_id *master )
{
	taskinfo_t *prev;
	taskinfo_t *cur = task_list;

	while (cur) {
		if (memcmp(&cur->master, master, sizeof(task_id))) {
			cur = cur->next;
			continue;
		}

		if (cur->state == TSTATE_EXECUTE || cur->state == TSTATE_WAIT ||
				cur->state == TSTATE_SUSPEND) {

			if (cur->std_io[0] >= 0)
				fd_close(cur->std_io[0]);
			if (cur->msg_io[0] >= 0)
				fd_close(cur->msg_io[0]);

			/* Send the process a KILL signal */
			kill(ntohl(cur->id.pid), SIGKILL);
		}

#if 0
		if (!memcmp(&cur->id, &cur->master, sizeof(task_id))) {
			if (cur->loader_fd >= 0) {
				conn_fd_open(cur->loader_fd);
			}
		}
#endif

		prev = cur;
		cur = cur->next;

		do_log(LOG_INFO, "task_remove_terminated: Removed pid<%d>",
				ntohl(prev->id.pid));
		task_remove(&prev->id);
	}

	return TRUE;
}

/* task_add_child: Sets the specified child as a child of the
 * parent. Returns TRUE on success or FALSE if the parent process
 * was not found on this machine. */
int task_add_child(task_id *parent, task_id *child)
{
	taskinfo_t *tsk;
	childinfo_t *newchild;

	if (!(tsk = task_find(parent)))
		return FALSE;

	newchild = malloc(sizeof(childinfo_t));
	newchild->child = *child;
	newchild->next = tsk->child_list;
	tsk->child_list = newchild;

	return TRUE;
}

/* task_remove_child: Removes the specified child from the child
 * list of the parent. Returns TRUE on success and FALSE otherwise. */
int task_remove_child(task_id *parent, task_id *child)
{
	taskinfo_t *tsk;

	tsk = task_find(parent);

	if (tsk) {
		childinfo_t *curchd = tsk->child_list;
		childinfo_t *prevchd = NULL;

		while (curchd) {
			if (!memcmp(&curchd->child, child, sizeof(task_id))) {
				if (prevchd == NULL)
					tsk->child_list = curchd->next;
				else
					prevchd->next = curchd->next;

				free(curchd);
				return TRUE;
			}

			prevchd = curchd;
			curchd = curchd->next;
		}
	}

	return FALSE;
}

/* task_remove_all_child: Removes all children from the list of
 * a task. Returns no. of children free'd on success and o/w FALSE, */
int task_remove_all_child(task_id *parent)
{
	taskinfo_t *tsk;
	int numchild = 0;

	tsk = task_find(parent);

	if (tsk) {
		childinfo_t *curchd = tsk->child_list;
		childinfo_t *prevchd;

		while (curchd) {
			prevchd = curchd;
			curchd = curchd->next;
			free(prevchd);

			numchild++;
		}

		tsk->child_list = NULL;
	}

	return numchild;
}

/* task_display_all: Print the task list. For debug. */
void task_display_all()
{
	taskinfo_t *cur = task_list;
	childinfo_t *cld;

	while (cur) {
		printf("pid:%d parent:%d master:%d prior:%d sym:%s start:%s file:%s stat:%d msg:%d:%d std:%d:%d\n",
				cur->id.pid, cur->parent.pid,
				cur->master.pid, cur->priority,
				(cur->sym_name)?cur->sym_name:"none",
				(cur->start_sym)?cur->start_sym:"none",
				(cur->file_name)?cur->file_name:"none",
				cur->state,
				cur->msg_io[0], cur->msg_io[1],
				cur->std_io[0], cur->std_io[1]);
		cld = cur->child_list;

		while (cld) {
			printf("\tchild:%d\n", cld->child.pid);
			cld = cld->next;
		}

		cur = cur->next;
	}
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * TASK MANAGEMENT
 * Below are the functions for managing the tasks themselves. They
 * communicate with the operating system to create new processes to
 * host the tasks and also setup the initial environment for the task
 * to execute in.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * task_get_msgsock: Returns the message fd of the specified task for
 * a local task. Is useful for sending LOCAL messages to tasks. Returns
 * the fd thru which messages can be sent to the task, or -1 on error.
 */
int task_get_msgio(task_id *tid)
{
	taskinfo_t *cur = task_list;

	if (!tid)
		return -1;

	while (cur) {
		if (!memcmp(&cur->id, tid, sizeof(task_id)))
			return cur->msg_io[0];

		cur = cur->next;
	}
	
	return -1;
}

/*
 * task_inc_load: Increments the daemon load and updates it at the VAS.
 */
int task_inc_load()
{
	dconf.load.numprocs += 1;
	return TRUE;
}

/*
 * task_dec_load: Decrements the daemon load and updates it at the VAS.
 */
int task_dec_load()
{
	dconf.load.numprocs -= 1;
	return TRUE;
}

/*
 * task_calc_load: Calculates the current load.
 * Returns TRUE if there is a change in the machine load.
 */
int task_calc_load()
{
	vasload_t load = { 0 };
	taskinfo_t *cur = task_list;

	while(cur != NULL) {
		if (cur->state == TSTATE_EXECUTE ||
				cur->state == TSTATE_WAIT ||
				cur->state == TSTATE_SUSPEND)
			load.numprocs++;

		cur = cur->next;
	}

	if (dconf.load.numprocs != load.numprocs) {
		dconf.load = load;
		return TRUE;
	}

	return FALSE;
}

/*
 * task_update_load: Informs the VAS of the current load.
 */
int task_update_load()
{
	msghdr_t req_hdr;
	vasload_t req_data;

	if (IS_UNKNOWN_NADDR(dconf.vas_cur))
		return FALSE;

	req_hdr.sender = dconf.self_id;
	req_hdr.receiver = dconf.vas_cur;
	req_hdr.msg_type = MSG_VAS_SETLOAD;
	req_hdr.priority = 1;
	req_hdr.data_size = sizeof(vasload_t);
	req_hdr.trans_id = TRANSID_NONE;

	req_data = dconf.load;
	hton_vasload(&req_data);

	if (!msg_write(&req_hdr, &req_data)) {
		do_log(LOG_WARNING, "task_set_load: Unable to update load at VAS");
		return FALSE;
	}

	return TRUE;
}

/*
 * task_terminate: Terminates the specified task. It performs
 * the following functions:
 *   1. Closes the local msg and stdio descriptors (not loader).
 *   2. Sends the task a KILL signal if its running.
 *   3. Sets the task state to TSTATE_TERMINATE.
 * The record of the task is maintained even after termination
 * until every task of this user application terminates.
 * Returns TRUE on success and FALSE otherwise.
 */
int task_terminate(task_id *tid, u_int32_t status)
{
	taskinfo_t *tinf;

	tinf = task_find(tid);

	if (!tinf)
		return FALSE;

	task_setstate_terminate(tinf, status);

	/* Close the msg and stdio descriptors to the task. But
	 * keep the loader descriptor open if this task is the
	 * master coz the rest of the childs might be running. */
	if (tinf->std_io[0] >= 0) {
		fd_close(tinf->std_io[0]);
		tinf->std_io[0] = -1;
	}
	if (tinf->msg_io[0] >= 0) {
		fd_close(tinf->msg_io[0]);
		tinf->msg_io[0] = -1;
	}

	if (waitpid(ntohl(tinf->id.pid), NULL, WNOHANG|WUNTRACED) == 0) {
		kill(ntohl(tinf->id.pid), SIGKILL);
		waitpid(ntohl(tinf->id.pid), NULL, 0);
	}

	if (task_calc_load())
		task_update_load();

	return TRUE;
}

/*
 * task_child_extract: A utility function used by task_abort.
 * It extracts tasks from the child list which are not
 * already in the uniq_list and which are remotely located
 * and adds those to the uniq_list. Returns TRUE on success.
 */
int task_child_extract(childinfo_t **uniq_list, childinfo_t *cld_list)
{
	childinfo_t *cur_uniq;
	childinfo_t *cur_cld = cld_list;

	while (cur_cld) {
		if (!memcmp(&cur_cld->child.cdaemon, &dconf.self_id, sizeof(daemon_id))) {
			cur_cld = cur_cld->next;
			continue;
		}

		cur_uniq = *uniq_list;

		while (cur_uniq) {
			if (!memcmp(&cur_cld->child, &cur_uniq->child, sizeof(task_id)))
				break;
			cur_uniq = cur_uniq->next;
		}

		if (!cur_uniq) {
			cur_uniq = malloc(sizeof(childinfo_t));
			*cur_uniq = *cur_cld;
			cur_uniq->next = *uniq_list;
			*uniq_list = cur_uniq;
		}

		cur_cld = cur_cld->next;
	}

	return TRUE;
}

/*
 * task_remote_child: Extracts all tasks with the given master
 * which are remotely located. The list only contains the
 * unique children atmost one per machine. The caller is
 * responsible for free'ing the childinto_t list returned.
 */
childinfo_t *task_remote_child(task_id *master)
{
	taskinfo_t *cur = task_list;
	childinfo_t *cld_list = NULL;

	while (cur) {
		if (!memcmp(&cur->master, master, sizeof(task_id)))
			task_child_extract(&cld_list, cur->child_list);

		cur = cur->next;
	}

	return cld_list;
}

/*
 * task_abort: Aborts all tasks with the specified master.
 * The function traverses the task list and:
 *   1. Closes the loader, local msg and stdio descriptors.
 *   2. Sends the task a KILL signal if its running.
 *   3. Stores the childs of the task in child (unique ones).
 *   4. Removes the task and its children from tasklist.
 * The unique list of children of all tasks running on this
 * machine with children running remotely, is returned by
 * the function. The caller is responsible for freeing it.
 */
childinfo_t *task_abort(task_id *master)
{
	taskinfo_t *prev;
	taskinfo_t *cur = task_list;
	childinfo_t *cld_list = NULL;

	while (cur) {
		if (memcmp(&cur->master, master, sizeof(task_id))) {
			cur = cur->next;
			continue;
		}

		if (cur->state == TSTATE_EXECUTE || cur->state == TSTATE_WAIT ||
				cur->state == TSTATE_SUSPEND) {

			/* if (cur->loader_fd >= 0)
				fd_close(cur->loader_fd); */
			if (cur->std_io[0] >= 0)
				fd_close(cur->std_io[0]);
			if (cur->msg_io[0] >= 0)
				fd_close(cur->msg_io[0]);

			/* Send the process a KILL signal */
			kill(ntohl(cur->id.pid), SIGKILL);
		}
		/*
		else if (cur->state == TSTATE_TERMINATE)
			if (cur->loader_fd >= 0) fd_close(cur->loader_fd);
		*/

		/* Now extract unique remote child tasks */
		task_child_extract(&cld_list, cur->child_list);

		prev = cur;
		cur = cur->next;
		task_remove(&prev->id);
	}

	if (task_calc_load())
		task_update_load();

	return cld_list;
}

/*
 * create_master: Create a new master (aka task), link the daemon
 * to it thru a pipe and execute it. Returns the task_id* of the
 * newly created process or NULL on failure.
 */
task_id *create_master(char *binary, int priority, int argc, char *argvs,
		int loader_fd)
{
	task_id tid;
	pid_t pid;
	int msg_io[2];
	int std_io[2];
		char **argv;

	TASK_ID_NULL(tid);

	if (access(binary, R_OK|X_OK)) {
		yell("Task create_master: Binary does not have read/execute permission: path<%s>",
				binary);
		return NULL;
	}
	
	if (!make_socketpair(msg_io) || !make_socketpair(std_io)) {
		yell("Task create_master: Unable to create socketpair: %s", strerror(errno));
		return NULL;
	}
	fd_open(msg_io[0], FDTYPE_LOCALMSG);
	fd_open(std_io[0], FDTYPE_LOCALDATA);

	/* now setup the command line arguments */
	argv = get_argv(argvs, argc, 2);
	argv[0] = strdup(binary);
	argv[1] = strdup(TASK_MASTER_STARTSYM);

	pid = fork();

	if (pid < 0) {
		/* an error has occured */
		yell("Task create_master: Unable to fork new task: binary<%s>", binary);
	}
	else if (pid == 0) {
		/* new task successfully forked */

		dup2(std_io[1], 0);	/* redir stdin */
		dup2(std_io[1], 1);	/* redir stdout */
		dup2(std_io[1], 2);	/* redir stderr */
		dup2(msg_io[1], 3);	/* fd 3 for msg read/write */

		execv(binary, argv);
		do_log(LOG_CRIT, "Task create_master: Unable to exec task: binary<%s>", binary);
		exit(EXIT_FAILURE);
	}
	else {
		/* the parent daemon */
		taskinfo_t *tsk = (taskinfo_t*) malloc(sizeof(taskinfo_t));

		/* generate task id using controlling daemon's id */
		make_task_id(&tid, &dconf.self_id, pid);

		/* add to tasks list */
		tsk->id = tid;
		TASK_ID_NULL(tsk->parent);
		tsk->master = tid;
		tsk->priority = 1;
		tsk->sym_name = base_name(binary);
		tsk->start_sym = (char*) strdup(TASK_MASTER_STARTSYM);
		tsk->file_name = (char*) strdup(binary);
		tsk->state = TSTATE_EXECUTE;
		tsk->state_data = NULL;
		tsk->loader_fd = loader_fd;
		tsk->child_list = NULL;

		tsk->msg_io[0] = msg_io[0];
		tsk->msg_io[1] = msg_io[1];
		tsk->std_io[0] = std_io[0];
		tsk->std_io[1] = std_io[1];

		task_add(tsk);
		return &tsk->id;
	}

	return NULL;
}

/*
 * create_local_task: Create a new task locally and link it to the
 * daemon thru a pipe and execute it. Returns the task_id* of the
 * newly created process or NULL on failure.
 */
task_id *create_local_task(char *binary, char *start_sym, task_id *master,
		task_id *parent, int priority)
{
	task_id tid;
	pid_t pid;
	int msg_io[2];
	int std_io[2];

	TASK_ID_NULL(tid);

	if (access(binary, R_OK|X_OK)) {
		yell("Task create_local: Binary does not have read/execute permission: path<%s>",
				binary);
		return NULL;
	}
	
	if (!make_socketpair(msg_io) || !make_socketpair(std_io)) {
		yell("Task create_local: Unable to create socketpair: %s", strerror(errno));
		return NULL;
	}
	fd_open(msg_io[0], FDTYPE_LOCALMSG);
	fd_open(std_io[0], FDTYPE_LOCALDATA);

	pid = fork();

	if (pid < 0) {
		/* an error has occured */
		yell("Task create_local: Unable to fork new task: binary<%s>", binary);
	}
	else if (pid == 0) {
		/* new task successfully forked */
		dup2(std_io[1], 0);	/* redir stdin */
		dup2(std_io[1], 1);	/* redir stdout */
		dup2(std_io[1], 2);	/* redir stderr */
		dup2(msg_io[1], 3);	/* fd 3 for msg read/write */

		execl(binary, binary, start_sym, NULL);
		do_log(LOG_CRIT, "Task create_local: Unable to exec task: binary<%s>", binary);
		exit(EXIT_FAILURE);
	}
	else {
		/* the parent daemon */
		taskinfo_t *tsk = (taskinfo_t*) malloc(sizeof(taskinfo_t));

		/* generate task id using controlling daemon's id */
		make_task_id(&tid, &dconf.self_id, pid);

		/* add to tasks list */
		tsk->id = tid;
		tsk->parent = *parent;
		tsk->master = *master;
		tsk->priority = 1;
		tsk->sym_name = base_name(binary);
		tsk->start_sym = (char*) strdup(start_sym);
		tsk->file_name = (char*) strdup(binary);
		tsk->state = TSTATE_EXECUTE;
		tsk->state_data = NULL;
		tsk->loader_fd = -1;
		tsk->child_list = NULL;

		tsk->msg_io[0] = msg_io[0];
		tsk->msg_io[1] = msg_io[1];
		tsk->std_io[0] = std_io[0];
		tsk->std_io[1] = std_io[1];

		task_add(tsk);
		return &tsk->id;
	}

	return NULL;
}

/* The following functions make up the remote create task */

/*
 * failhnd_getminload: Called if no reply to get min load
 * request to the VAS. For now simple fails the task
 * creation.
 * XXX: Create a local task instead.
 */
TRANSFAIL_HANDLER(failhnd_getminload)
{
	taskreq_t *treq = data;

	if (treq)
		yell("Create task: getminload timed out. Aborting task creation<>",
				treq->file_name);
	else
		yell("Create task: getminload timed out. Operation not found.");
}

/*
 * create_task_minload: A pretty complicated function used to create a new
 * task by contacting the VAS and getting the machine with the
 * smallest load. The load is received in the VAS_GETLOAD_RPLY
 * transaction reply. treq is in host byte order (default).
 */
int create_task_minload(taskreq_t *treq)
{
	msghdr_t req_hdr;
	taskreq_t *mtreq;

	mtreq = (taskreq_t*) malloc(sizeof(taskreq_t));
	*mtreq = *treq;

	req_hdr.sender = dconf.self_id;
	req_hdr.receiver = dconf.vas_cur;
	req_hdr.msg_type = MSG_VAS_GETMINLOAD;
	req_hdr.priority = 1;
	req_hdr.data_size = 0;

	if (!trans_write(&req_hdr, NULL, failhnd_getminload, mtreq, TRANS_TIMEOUT)) {
		yell("Create task: Unable to send getminload request to VAS");
		return FALSE;
	}

	return TRUE;
}

/*
 * failhnd_create_localtask: Handles a timeout on the last step
 * of task creation, i.e. sending the local task creation msg to
 * the daemon with the minimum load.
 */
TRANSFAIL_HANDLER(failhnd_create_localtask)
{
	yell("Create task: failhnd_create_localtask: Request timeout. Task creation aborted.");
}

/*
 * ftphnd_taskbin: Handler for the events associated with the
 * transfer of the task binary to the minimum load daemon. The
 * transfer is initiated in create_task_ftp (below).
 */
FTP_HND_EVENT(ftphnd_taskbin)
{
	msghdr_t req_hdr;
	taskreq_t *treq = (taskreq_t*) data;
	
	if (!treq) {
		yell("Create task: ftphnd_taskbin: Operation not found");
		return;
	}

	switch (ftpitem->status) {
		case FTPSTAT_COMPLETE:
			/* Now we are ready to call the local task
			 * creation on the min load daemon. No need to
			 * strip binary path coz crt local task does
			 * that itself. */
			req_hdr.sender = dconf.self_id;
			req_hdr.receiver = ftpitem->remote_dem;
			req_hdr.msg_type = MSG_CREATE_LOCALTASK;
			req_hdr.priority = 1;
			req_hdr.data_size = sizeof(taskreq_t);

			hton_taskreq(treq);

			if (!trans_write(&req_hdr, treq, failhnd_create_localtask,
						NULL, TRANS_TIMEOUT)) {
				yell("Create task: ftphnd_taskbin: Unable to send task creation request");
				/* XXX: Create task locally instead */
			}
			break;

		/* XXX: In all of the cases below, create a local task instead */

		case FTPSTAT_TIMEOUT:
			yell("Create task: ftphnd_taskbin: Transfer timedout. Task creation aborted.");
			break;
		case FTPSTAT_CANCEL:
			yell("Create task: ftphnd_taskbin: Transfer canceled. Task creation aborted.");
			break;
		case FTPSTAT_FAIL:
			yell("Create task: ftphnd_taskbin: Transfer failed. Task creation aborted.");
			break;
	}
}

/*
 * create_task_ftp: Next step after getting the minimum load
 * daemon is to send the file to it. This initiates the file
 * transfer to the daemon. Control goes to either the trans
 * timeout handler or to FTP_DEINIT handler.
 */
int create_task_ftp(taskreq_t *treq, daemon_id *min_loaded)
{
	taskreq_t *mtreq;

	mtreq = (taskreq_t*) malloc(sizeof(taskreq_t));
	*mtreq = *treq;

	if (!ftp_send(min_loaded, treq->file_name, ftphnd_taskbin, mtreq))
		return FALSE;

	return TRUE;
}

/* Functions below are for managing task states */

/*
 * task_state_datafree: A utility function that Frees the
 * data accociated with the current state of the task.
 * This must be called each time the state of the task is
 * being changed. Returns TRUE always.
 */
int task_state_datafree(taskinfo_t *tinf)
{
	tstate_wait_t *ts_wait;

	if (!tinf->state_data)
		return TRUE;

	switch (tinf->state) {
	case TSTATE_WAIT:
		ts_wait = tinf->state_data;
		if (ts_wait->data)
			free(ts_wait->data);
		break;

	case TSTATE_NOTFOUND:
	case TSTATE_EXECUTE:
	case TSTATE_COMPLETE:
	case TSTATE_SUSPEND:
	case TSTATE_TERMINATE:
	case TSTATE_UNDETERMINED:
		break;
	}

	free(tinf->state_data);
	tinf->state_data = NULL;
	return TRUE;
}

/*
 * task_setstate_wait: Sets the task state to TSTATE_WAIT.
 * The sdata parameter is copied into an allocated buffer.
 * Returns TRUE on success state change (always).
 */
int task_setstate_wait(taskinfo_t *tinf, tstate_wait_t *sdata)
{
	tstate_wait_t *smdata;

	task_state_datafree(tinf);
	tinf->state = TSTATE_WAIT;
	smdata = (tstate_wait_t*)malloc(sizeof(tstate_wait_t));
	*smdata = *sdata;
	tinf->state_data = smdata;

	return TRUE;
}

/*
 * task_setstate_execute: Sets the task state to TSTATE_EXECUTE.
 * Returns TRUE on successful state change, or FALSE otherwise.
 */
int task_setstate_execute(taskinfo_t *tinf)
{
	task_state_datafree(tinf);
	tinf->state = TSTATE_EXECUTE;

	return TRUE;
}

/*
 * task_setstate_terminate: Sets the task state to TSTATE_TERMINATE.
 * The status parameter is the exit status of this task.
 * Returns TRUE on successful state change, or FALSE otherwise.
 */
int task_setstate_terminate(taskinfo_t *tinf, u_int32_t status)
{
	int *smdata;

	task_state_datafree(tinf);
	tinf->state = TSTATE_TERMINATE;
	
	smdata = (u_int32_t*)malloc(sizeof(u_int32_t));
	*smdata = status;
	tinf->state_data = smdata;

	return TRUE;
}
