/*
 * clist.c: A generic linked list data structure
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "defs.h"
#include "clist.h"
#include "io.h"
#include "msgs.h"
#include "transact.h"


/* the list of all registered transactions */
clist *trans_list = NULL;

/* 
 * trans_cmp: A utility function which compares two transactions
 * and is used for sorted insertions into clist.
 * Returns 1 if a>b, -1 if a<b and 0 if a==b.
 */
int trans_cmp(void *a, void *b)
{
	transact_t *x = (transact_t*)a;
	transact_t *y = (transact_t*)b;

	if (x->trans_id > y->trans_id)
		return 1;
	else if (x->trans_id < y->trans_id)
		return -1;
	else
		return 0;
}

/*
 * trans_add: A utility function which generates the next
 * available transaction id and then adds the given transaction
 * to trans_list by then ID. Only the trans_id field is filled
 * in by this function. The item must be malloced by the caller
 * and the transaction module is responsible for freeing it.
 * Returns the trans_id of the newly added transaction.
 */
u_int16_t trans_add(transact_t *item)
{
	u_int16_t trans_id = 1;
	clist *cur = trans_list;

	while (cur != NULL) {
		transact_t *titem = (transact_t*) cur->data;
		if (titem->trans_id != trans_id)
			break;
		trans_id++;
		cur = cur->next;
	}

	item->trans_id = trans_id;
	trans_list = clist_insert_sorted(trans_list, item, trans_cmp);

	return item->trans_id;
}

/*
 * trans_remove: Deletes the specified transaction from trans_list.
 * Also frees the faildata if it is non-NULL.
 * Returns TRUE on success and FALSE on failure.
 */
int trans_remove(u_int16_t trans_id)
{
	clist *cur = trans_list;

	while (cur != NULL) {
		transact_t *item = (transact_t*) cur->data;
		if (item->trans_id == trans_id) {
			if (item->faildata)
				free(item->faildata);
			free(item);

			trans_list = clist_remove_link(trans_list, cur);
			break;
		}
		cur = cur->next;
	}

	return (cur) ? TRUE : FALSE;
}

/*
 * trans_remove_all: Deletes all transactions and frees the
 * transaction list. This is used as a cleanup before exit.
 */
int trans_remove_all()
{
	clist *tmp;
	clist *cur = trans_list;

	while (cur != NULL) {
		transact_t *item = (transact_t*) cur->data;
		tmp = cur;
		cur = cur->next;

		if (item->faildata)
			free(item->faildata);
		free(item);

		trans_list = clist_remove_link(trans_list, tmp);
	}

	return TRUE;
}

/*
 * trans_write: Creates a new transaction. From the user's view this
 * is the same as msg_write but instead in this function, a new
 * transaction is created and then msg_write is called as usual.
 * msghdr.trans_id and msghdr.trans_isrply are filled out by this
 * function before calling msg_write.
 * The faildata is stored in the 'data' field of this transaction
 * and would be passed to the failure handler incase the transaction
 * timeouts. And if not, the transaction would be removed normally.
 * The faildata is free'ed by the transaction module if it is
 * non-NULL so it must be malloc'ed by the caller.
 * Returns what msg_write returns.
 */
int trans_write(msghdr_t *msghdr, void *msgdata,
		TRANSFAIL_HANDLER_PTR(hnd_fail),
		void *faildata, int timeout)
{
	transact_t *trans;

	trans = (transact_t*) malloc(sizeof(transact_t));
	trans->trans_id = TRANSID_NONE;
	trans->status = TRANS_STATUS_PENDING;
	trans->ttl = time(NULL)+timeout;
	trans->faildata = faildata;
	trans->hnd_fail = hnd_fail;

	msghdr->trans_id = trans_add(trans);
	msghdr->trans_isrply = FALSE;

	if (!msg_write(msghdr, msgdata)) {
		trans_remove(msghdr->trans_id);
		return FALSE;
	}

	return TRUE;
}

/*
 * trans_reply: This function is used to send a reply to a transactional
 * message. Unlike trans_write, msghdr.trans_id must be filled in by
 * the caller. This function only sets msghdr.trans_isrply to TRUE
 * before calling msg_reply.
 * Returns whatever msg_reply returns.
 */
int trans_reply(msghdr_t *msghdr, void *msgdata, int fd)
{
	msghdr->trans_isrply = TRUE;
	return msg_reply(msghdr, msgdata, fd);
}

/*
 * trans_get: Searches the transaction list for the given trans_id.
 * If found, returns the transact_t item of the transaction and
 * on failure, returns NULL.
 */
transact_t *trans_get(u_int16_t trans_id)
{
	clist *cur = trans_list;

	while (cur) {
		transact_t *item = (transact_t*) cur->data;
		if (item->trans_id == trans_id)
			return item;
		cur = cur->next;
	}

	return NULL;
}

/*
 * trans_run: This function is called every second and it goes through
 * the trans_list checking for transaction which have expired. If any
 * is found, it calls its failure handler and removes it from the list.
 * Returns nothing.
 */
CTIMER_FUNC(trans_run)
{
	clist *cur = trans_list;
	time_t cur_time = time(NULL);

	while (cur) {
		transact_t *item = (transact_t*) cur->data;
		
		if (cur_time >= item->ttl) {
			item->status = TRANS_STATUS_TIMEOUT;

			if (item->hnd_fail)
				item->hnd_fail(item, item->faildata);
			free(item);
			
			trans_list = clist_remove_link(trans_list, cur);
		}
		
		cur = cur->next;
	}
}

#ifdef NCDAEMON
/*
 * trans_lwrite: Use this function in the daemon for writing
 * a transaction to a local task. And to reply a task's
 * transaction use trans_lreply (which is a synonym to trans_reply.
 * The faildata is free'ed by the transaction module if it is
 * non-NULL so it must be malloc'ed by the caller.
 * Returns whatever msg_lreply returns.
 */
int trans_lwrite(msghdr_t *msghdr, void *msgdata, task_id *tid,
		TRANSFAIL_HANDLER_PTR(hnd_fail),
		void *faildata, int timeout)
{
	transact_t *trans;

	trans = (transact_t*) malloc(sizeof(transact_t));
	trans->trans_id = TRANSID_NONE;
	trans->status = TRANS_STATUS_PENDING;
	trans->ttl = time(NULL)+timeout;
	trans->faildata = faildata;
	trans->hnd_fail = hnd_fail;

	msghdr->trans_id = trans_add(trans);
	msghdr->trans_isrply = FALSE;

	if (!msg_lwrite(msghdr, msgdata, tid)) {
		trans_remove(msghdr->trans_id);
		return FALSE;
	}
	
	return TRUE;
}
#endif
