/*
 * util.c: Utility functions which dont fit anywhere else
 * (C) Copyright 2003, netcluster
 *
 * Written by Fakhir Ansari
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#include "defs.h"
#include "sock.h"
#include "util.h"


/*
 * make_daemon_id: Fills in the daemon_id struct in the correct
 * byte order and returns TRUE on success and FALSE on failure.
 */
int make_daemon_id(daemon_id *id, char *ip, int port)
{
	if ((id->ip = inet_addr(ip)) < 0)
		return FALSE;
	id->port = htons(port);
	return TRUE;
}

/*
 * make_self_daemon_id: Makes a daemon ID which points to the
 * same machine this function is executing on.
 */
int make_self_daemon_id(daemon_id *id, int port)
{
	int i;
	struct hostent *hostinfo;
	char hname[MAX_HOSTNAME_SIZE+1];

	if (gethostname(hname, MAX_HOSTNAME_SIZE) < 0)
		return FALSE;

	hostinfo = gethostbyname(hname);
	if (hostinfo == NULL)
		return FALSE;

	id->ip = ((struct in_addr *)hostinfo->h_addr_list[0])->s_addr;
	id->port = htons(port);

#if 0
	for (i=0;;i++) {
		if (hostinfo->h_addr_list[i] == 0)
			break;
		fprintf(stderr, "Self IP: %s\n", inet_ntoa(*((struct in_addr *)hostinfo->h_addr_list[i])));
	}
#endif

	return TRUE;
}

/*
 * make_task_id: Fills in the task_id in the correct byte order
 * and returns TRUE on success and FALSE on failure.
 */
int make_task_id(task_id *tid, daemon_id *did, int pid)
{
	tid->cdaemon = *did;
	tid->pid = htonl(pid);
	return TRUE;
}

/*
 * sockaddr_to_naddr: Convert a sockaddr_in to the locally used data
 * type naddr_t. The two types are almost equivalent and the local
 * one is used to save a few bytes. It is assumed that the sockaddr
 * is in network byte order. Returns TRUE on success (always).
 */
int sockaddr_to_naddr(struct sockaddr_in *sa, naddr_t *na)
{
	na->ip = sa->sin_addr.s_addr;
	na->port = sa->sin_port;
	return TRUE;
}

/* str_to_naddr: Convert a string to an naddr_t. */
int str_to_naddr(char *str, naddr_t *na)
{
	char *serv;
	char *port;
	int pnum;
	struct sockaddr_in sa;
	
        if ((serv = next_arg_custom(str, &str, ": \t")) == NULL)
		return FALSE;
	port = next_arg_custom(str, &str, ": \t");

	/* The string must now be empty after removing serv and port */
	if (*str)
		return FALSE;

	if (!port)
		pnum = DPORT;
	else if (!str_isinteger(port))
		return FALSE;
	else
		pnum = atoi(port);

	sa = make_sockaddr_in(serv, pnum);
	return sockaddr_to_naddr(&sa, na);
}

/*
 * naddr_to_str: Converts naddr to a string in a static buffer.
 * Subsequent calls will overwrite it.
 */
char *naddr_to_str(naddr_t *na)
{
	struct in_addr in;
	static char str[MAX_STRADDR_SIZE];
	
	in.s_addr = na->ip;
	strcpy(str, inet_ntoa(in));
	return str;
}

/* DATA TYPE CONVERSION FUNCTIONS */

/*
 * ntoh_msg: Convert the network byte ordered fields in a message
 * to the host byte order so they would be used locally. All msgs
 * sent and received from the network are in network byte order.
 * The sender and receiver fields are not affected bcoz the xIDs
 * are always in network byte order whereas the msg is translated
 * to host byte order when being processed locally.
 * Returns TRUE on success (always).
 */
int ntoh_msg(msghdr_t *msg)
{
	msg->msg_type = ntohs(msg->msg_type);
	msg->priority = ntohs(msg->priority);
	msg->data_size = ntohl(msg->data_size);
	msg->trans_id = ntohs(msg->trans_id);
	return TRUE;
}

/*
 * hton_msg: Convert the host ordered fields in the message to the
 * network order. This must be called before a message is sent on
 * the network. The sender and receiver fields are left unaffected.
 * Returns TRUE on success (always)
 */
int hton_msg(msghdr_t *msg)
{
	msg->msg_type = htons(msg->msg_type);
	msg->priority = htons(msg->priority);
	msg->data_size = htonl(msg->data_size);
	msg->trans_id = htons(msg->trans_id);
	return TRUE;
}

/* ntoh_taskreq: Bring taskreq_t into host byte order */
int ntoh_taskreq(taskreq_t *tr)
{
	tr->priority = ntohs(tr->priority);
	return TRUE;
}

/* ntoh_taskreq: Bring taskreq_t into network byte order */
int hton_taskreq(taskreq_t *tr)
{
	tr->priority = htons(tr->priority);
	return TRUE;
}

/* ntoh_taskarg: Bring taskarg_t into host byte order */
int ntoh_taskarg(taskarg_t *tr)
{
	tr->argc = ntohs(tr->argc);
	return TRUE;
}

/* ntoh_taskarg: Bring taskarg_t into network byte order */
int hton_taskarg(taskarg_t *tr)
{
	tr->argc = htons(tr->argc);
	return TRUE;
}

/* hton_taskrply: Bring taskrply_t into host byte order */
int ntoh_taskrply(taskrply_t *tr)
{
	tr->status = ntohs(tr->status);
	return TRUE;
}

/* hton_taskrply: Bring taskrply_t into network byte order */
int hton_taskrply(taskrply_t *tr)
{
	tr->status = htons(tr->status);
	return TRUE;
}

/* ntoh_vasreg_rply: Bring vasreg_rply_t into host byte order */
int ntoh_vasreg_rply(vasreg_rply_t *tr)
{
	tr->status = ntohl(tr->status);
	return TRUE;
}

/* hton_vasreg_rply: Bring vasreg_rply_t into network byte order */
int hton_vasreg_rply(vasreg_rply_t *tr)
{
	tr->status = htonl(tr->status);
	return TRUE;
}

/* ntoh_vasload: Bring vasload_t into host byte order */
int ntoh_vasload(vasload_t *tr)
{
	tr->numprocs = ntohl(tr->numprocs);
	return TRUE;
}

/* hton_vasload: Bring vasload_t into network byte order */
int hton_vasload(vasload_t *tr)
{
	tr->numprocs = htonl(tr->numprocs);
	return TRUE;
}

/* ntoh_ftpid: Bring ftpid_t into host byte order */
int ntoh_ftpid(ftpid_t *tr)
{
	tr->ftpind = ntohl(tr->ftpind);
	return TRUE;
}

/* hton_ftpid: Bring ftpid_t into network byte order */
int hton_ftpid(ftpid_t *tr)
{
	tr->ftpind = htonl(tr->ftpind);
	return TRUE;
}

/* ntoh_ftpinit: Bring ftpinit_t into host byte order */
int ntoh_ftpinit(ftpinit_t *tr)
{
	ntoh_ftpid(&tr->ftp_id);
	tr->file_size = ntohl(tr->file_size);
	return TRUE;
}

/* hton_ftpinit: Bring ftpinit_t into network byte order */
int hton_ftpinit(ftpinit_t *tr)
{
	hton_ftpid(&tr->ftp_id);
	tr->file_size = htonl(tr->file_size);
	return TRUE;
}

/* ntoh_rpcinfo: Bring rpcinfo_t into host byte order */
int ntoh_rpcinfo(rpcinfo_t *tr)
{
	tr->status = ntohl(tr->status);
	tr->func_id = ntohl(tr->func_id);
	return TRUE;
}

/* hton_rpcinfo: Bring rpcinfo_t into network byte order */
int hton_rpcinfo(rpcinfo_t *tr)
{
	tr->status = htonl(tr->status);
	tr->func_id = htonl(tr->func_id);
	return TRUE;
}

/* ntoh_rpcaccept: Bring rpcaccept_t into host byte order */
int ntoh_rpcaccept(rpcaccept_t *tr)
{
	tr->func_id = ntohl(tr->func_id);
	tr->is_blocking = ntohl(tr->is_blocking);
	return TRUE;
}

/* hton_rpcaccept: Bring rpcaccept_t into network byte order */
int hton_rpcaccept(rpcaccept_t *tr)
{
	tr->func_id = htonl(tr->func_id);
	tr->is_blocking = htonl(tr->is_blocking);
	return TRUE;
}

/* ntoh_exitinfo: Bring exitinfo_t into network byte order */
int ntoh_exitinfo(exitinfo_t *tr)
{
	tr->status = ntohs(tr->status);
	return TRUE;
}

/* hton_exitinfo: Bring exitinfo_t into network byte order */
int hton_exitinfo(exitinfo_t *tr)
{
	tr->status = htons(tr->status);
	return TRUE;
}

/* ntoh_taskpoll: Bring taskpoll_t into network byte order */
int ntoh_taskpoll(taskpoll_t *tr)
{
	tr->num = ntohs(tr->num);
	return TRUE;
}

/* hton_taskpoll: Bring taskpoll_t into network byte order */
int hton_taskpoll(taskpoll_t *tr)
{
	tr->num = htons(tr->num);
	return TRUE;
}

/*
 * malloc_strcpy: Copy a string into another malloc'd buffer.
 */
char *malloc_strcpy(char *str)
{
	return strdup(str);
}

/*
 * malloc_strcat: Concatenate one string to another into a malloc'd
 * buffer and returns a ptr to the buffer.
 */
char *malloc_strcat(char *str1, char *str2)
{
	char *new_str = (char*)malloc(strlen(str1)+strlen(str2)+1);
	strcpy(new_str, str1);
	strcat(new_str, str2);
	return new_str;
}

/*
 * strmcpy: Similar to the strncpy function but also always adds a
 * trailing NULL to the copied string.
 */
char *strmcpy(char *dest, char *src, int n)
{
	strncpy(dest, src, n);
	dest[n-1] = '\0';
	return dest;
}

/*
 * strmcat: Similar to strncat but always adds a trailing NULL to
 * the concatenated string.
 */
char *strmcat(char *dest, char *src, int n)
{
	strncat(dest, src, n);
	dest[n-1] = '\0';
	return dest;
}

/*
 * trim_leading_spaces: Removes the spaces from the start of the string.
 * The original string is modified.
 */
char *trim_leading_spaces(char *str)
{
	char *real = str;

	while (*real && isspace(*real))
		real++;

	return (char*) memmove(str, real, strlen(real)+1);
}
 
/*
 * trim_trailing_spaces: Removes spaces from the end of the string.
 */
char *trim_trailing_spaces(char *str)
{
	int len;
	char *ptr;

	if ((len = strlen(str)) == 0)
		*str = '\0';
	else {
		ptr = str + len - 1;
		while (*ptr && isspace(*ptr))
			ptr--;
		ptr[1] = '\0';
	}

	return str;
}

/*
 * trim_spaces: Removes whitespaces from both the start and
 * the end of the specified string. Original string is modified.
 */
char *trim_spaces(char *str)
{
	trim_leading_spaces(str);
	trim_trailing_spaces(str);
	return str;
}

/*
 * str_isempty: Returns nonzero if the string is of zero length or
 * if it only contains spaces and zero otherwise.
 */
int str_isempty(char *str)
{
	if (!str)
		return FALSE;

	while (*str && isspace(*str))
		str++;
	return (*str) ? FALSE : TRUE;
}

/*
 * str_isinteger: Returns true if the string is a valid integer.
 * Leading and trailing spaces are ignored.
 */
int str_isinteger(char *str)
{
	if (!str)
		return FALSE;

	while (*str && isspace(*str)) str++;
	if (!*str) return FALSE;
	if (*str == '-' || *str == '+') str++;
	while (*str && isdigit(*str)) str++;
	while (*str && isspace(*str)) str++;
	return (*str) ? FALSE : TRUE;
}

/*
 * str_isfloat: Returns true if the given string is a float. Leading
 * and trailing spaces are ignored.
 */
int str_isfloat(char *str)
{
	int has_dec = 0;

	if (!str)
		return FALSE;

	while (*str && isspace(*str)) str++;
	if (!*str) return FALSE;
	if (*str == '-' || *str == '+') str++;
	while (*str && (isdigit(*str) || *str=='.')) {
		if (*str == '.') {
			if (!has_dec)
				has_dec = TRUE;
			else
				break;
		}
		str++;
	}
	while (*str && isspace(*str)) str++;
	return (*str) ? FALSE : TRUE;
}

/*
 * str_ischarset: Returns true if the given string is only composed
 * of the character set passed to this function. If the first char
 * of the character set is '^', then the set is complimented.
 */
int str_ischarset(char *str, char *charset)
{
	if (!str || !charset)
		return FALSE;

	if (*charset == '^') {
		charset++;
		while (*str) {
			if (strchr(charset, *str) != NULL)
				return FALSE;
			str++;
		}
	}
	else {
		while (*str) {
			if (strchr(charset, *str) == NULL)
				return FALSE;
			str++;
		}
	}
	return TRUE;
}

/*
 * str_validquotes: Returns true if the quotes in the string are
 * balanced.
 * XXXX: Need to handle escaped quotes.
 */
int str_validquotes(char *str)
{
	int nquote = 0;

	if (!str)
		return FALSE;
	
	for (;*str;str++)
		if (*str == '"') nquote++;

	return !(nquote%2);
}

/*
 * sindex: Returns a ptr to the first occurance of a character in
 * cset. Works similar to index() but if the first character of
 * cset is '^' then returns the first occurance of a character
 * _not_ in the cset.
 */
char *sindex(char *str, char *cset)
{
	if (*cset == '^') {
		cset++;
		for (;*str;str++) {
			if (strchr(cset, *str) == NULL)
				break;
		}
	}
	else {
		for (;*str;str++) {
			if (strchr(cset, *str) != NULL)
				break;
		}
	}

	if (*str)
		return str;

	return NULL;
}

/*
 * rsindex: Similar to rindex but inverts the character set being
 * searched if cset's first character is '^'.
 */
char *rsindex(char *str, char *cset)
{
	int ind = strlen(str) - 1;

	if (!(*str))
		return NULL;

	if (*cset == '^') {
		cset++;
		for (;ind>=0;ind--) {
			if (strchr(cset, str[ind]) == NULL)
				break;
		}
	}
	else {
		for (;ind>=0;ind--) {
			if (strchr(cset, str[ind]) != NULL)
				break;
		}
	}

	if (ind>=0)
		return &str[ind];

	return NULL;
}

/*
 * next_arg: Extracts the next word from the string by ignoring
 * all white spaces and returns that word. The original string
 * is modified and the new_str parameter is updated to point
 * beyond the returned arg. Returns a ptr to the extracted arg.
 * new_ptr can be NULL if the string position is not to be
 * updated. Treats words inside double quotes as a single word.
 * XXXX: Need to handle escaped quotes.
 */
char *next_arg(char *str, char **new_str)
{
	char *ptr;

	if (!str || str_isempty(str))
		return NULL;

	if ((ptr = sindex(str, "^ \t")) != NULL) {
		if (*ptr == '"') {
			ptr++;
			if ((str = sindex(ptr, "\"")) != NULL)
				*str++ = '\0';
			else {
				ptr = NULL;
				str = empty_string;
			}
		}
		else {
			if ((str = sindex(ptr, " \t")) != NULL)
				*str++ = '\0';
			else
				str = empty_string;
		}
	}
	else
		str = empty_string;

	if (new_str)
		*new_str = str;
	return ptr;
}

/*
 * next_arg_custom: Similar to next_arg but does not assume
 * whitespace to be the token separater. Takes an extra argument
 * for the token separater(s). Make sure the first character of
 * 'sep' is not '^' coz its used internally. Doesnt handle quotes.
 */
char *next_arg_custom(char *str, char **new_str, char *sep)
{
	char *ptr;
	char *neg = malloc_strcat("^", sep);

	if ((ptr = sindex(str, neg)) != NULL) {
		if ((str = sindex(ptr, sep)) != NULL)
			*str++ = '\0';
		else
			str = empty_string;
	}
	else
		str = empty_string;

	free(neg);

	if (new_str)
		*new_str = str;
	return ptr;
}

/*
 * last_arg: Is the opposite of next_arg. Returns the last word
 * in the string and the original string is modified. Treats
 * words inside double quotes as single words.
 * XXXX: Need to handle escaped quotes.
 */
char *last_arg(char *str, char **new_str)
{
	char *ptr;
	char *qend;
	int cur;

	if (!str || str_isempty(str))
		return NULL;

	if (new_str)
		*new_str = str;

	trim_spaces(str);
	cur = strlen(str) - 1;

	/* Handle quoted strings */
	if (str[cur] == '"') {
		str[cur] = '\0';

		for (cur--;cur>=0;cur--) {
			if (str[cur] == '"') {
				str[cur] = '\0';
				return &str[cur+1];
			}
		}

		return NULL;
	}
	else if ((ptr = rsindex(str, " \t")) != NULL) {
		*ptr++ = '\0';
	}
	else {
		ptr = str;
		if (new_str)
		       *new_str = empty_string;
	}

	return ptr;
}

/*
 * word_count: Returns the number of words in the string. Handles
 * double quoted strings ("...") as single words. Use str_validquotes
 * to check if the quotes are balanced.
 * XXXXX: Need to handle escaped quotes.
 */
int word_count(char *str)
{
	int cnt = 0;

	do {
		while (isspace(*str)) str++;
		if (!(*str))
			break;
		if (*str == '"') {
			str++;
			while (*str && *str!='"') str++;
			if (*str) str++;
		}
		else
			while (*str && !isspace(*str)) str++;
		cnt++;
	} while (*str);

	return cnt;
}

/*
 * base_name: Returns the filename stripped of the directory name
 * The returned string is malloc'ed and the caller is responsible
 * for freeing it. Returns NULL on error.
 */
char *base_name(char *path)
{
	char *sptr;

	if (!path)
		return NULL;

	sptr = strrchr(path, DIR_SEPARATOR);

	if (sptr) {
		if (*(++sptr))
			return strdup(sptr);
	}
	
	return strdup(path);
}

/*
 * get_homedir: Returns the effective user's home directory in a
 * malloc'ed buffer. The caller is responsible for freeing it.
 * Returns NULL if unable to get the directory.
 */
char *get_homedir()
{
	struct passwd *entry;
	char *buf = NULL;
	int len;

	if ((entry = getpwuid(getuid()))) {
		len = strlen(entry->pw_dir);
		if (!len)
			buf=NULL;
		buf = (char*)malloc(len+1);
		strcpy(buf, entry->pw_dir);
	}
	else
		buf=NULL;

	return buf;
}

/*
 * expand_tilde: Expand the tilde character to the home directory of
 * the user. The expanded string is returned in a malloc'd buffer
 * and the caller is responsible for free'ing it.
 */
char *expand_tilde(char *ostr)
{
	struct passwd *entry;
	char buf[MAX_PATH_BUFFER];
	char *rest;
	char *tstr;
	char *nstr;

	if (!ostr)
		return NULL;

	nstr = strdup(ostr);
	tstr = nstr;

	if (*tstr == '~') {
		if (*++tstr == '/') {
			rest = tstr+1;
			if ((entry = getpwuid(getuid())))
				strmcpy(buf, entry->pw_dir, MAX_PATH_BUFFER);
			else
				*buf='\0';
		}
		else {

			if ((rest = strchr(tstr, '/')) != NULL)
				*rest++ = '\0';

			if ((entry = getpwnam(tstr)) != NULL)
				strmcpy(buf, entry->pw_dir, MAX_PATH_BUFFER);
			else {
				free(tstr);
				return NULL;
			}
		}

		if (rest) {
			strmcat(buf, "/", MAX_PATH_BUFFER);
			strmcat(buf, rest, MAX_PATH_BUFFER);
		}
	}
	else
		strmcpy(buf, tstr, MAX_PATH_BUFFER);

	if (nstr)
		free(nstr);
	return malloc_strcpy(buf);
}

/*
 * put_param: A utility function that stores the ap parameters
 * into the buffer according to the given format. Also performs
 * byte order conversion to network byte order. Only a few format
 * specifiers are supported currently. Returns the size of the
 * buffer used or -1 if buffer not big enough.
 */
int put_param(char *buf, size_t n, char *fmt, va_list ap)
{
	int num;
	char ch;
	char *str;
	task_id tid;
	rawdata_t raw;
	int len;
	char *curfmt = fmt;
	int curlen = 0;

	while (*curfmt == '%') {
		switch (*(++curfmt)) {
		case 's':
			str = va_arg(ap, char *);
			len = strlen(str)+1;
			if (curlen + len > n)
				return -1;
			memcpy(buf, str, len);
			buf += len;
			curlen += len;
			break;
		case 'c':
			ch = va_arg(ap, char);
			if (curlen + 1 > n)
				return -1;
			*buf++ = ch;
			curlen++;
			break;
		case 'd':
			num = htonl(va_arg(ap, u_int32_t));
			len = sizeof(u_int32_t);
			if (curlen + len > n)
				return -1;
			memcpy(buf, &num, len);
			buf += len;
			curlen += len;
			break;
		case 't':
			tid = va_arg(ap, task_id);
			len = sizeof(task_id);
			if (curlen + len > n)
				return -1;
			memcpy(buf, &tid, len);
			buf += len;
			curlen += len;
			break;
		case 'r':
			raw = va_arg(ap, rawdata_t);

			/* Copy both size and data to buffer */
			len = sizeof(u_int32_t) + raw.size;
			if (curlen + len > n)
				return -1;
			memcpy(buf, &raw.size, sizeof(u_int32_t));
			memcpy(buf+sizeof(u_int32_t), raw.data, raw.size);
			buf += len;
			curlen += len;
			break;
		}

		if (*curfmt)
			curfmt++;
	}

	return curlen;
}

/*
 * get_param: A utility function that restores values from the
 * buffer into the ap parameters according to the given format.
 * Also performs byte order conversion to host byte order.
 * Only a few format specifiers are supported currently.
 * NOTE: The data for %r raw data type is malloc'ed and the
 * caller is responsible for freeing it.
 * Returns the size of buffer read.
 */
int get_param(char *buf, char *fmt, va_list ap)
{
	int *num;
	char *ch;
	char *str;
	task_id *tid;
	rawdata_t *raw;
	int len;
	char *curfmt = fmt;
	int curlen = 0;

	while (*curfmt == '%') {
		switch (*(++curfmt)) {
		case 's':
			str = va_arg(ap, char *);
			len = strlen(buf)+1;
			memcpy(str, buf, len);
			buf += len;
			curlen += len;
			break;
		case 'c':
			ch = va_arg(ap, char *);
			*ch = *buf++;
			curlen++;
			break;
		case 'd':
			num = va_arg(ap, u_int32_t *);
			len = sizeof(u_int32_t);
			memcpy(num, buf, len);
			*num = ntohl(*num);
			buf += len;
			curlen += len;
			break;
		case 't':
			tid = va_arg(ap, task_id *);
			len = sizeof(task_id);
			memcpy(tid, buf, len);
			buf += len;
			curlen += len;
			break;
		case 'r':
			raw = va_arg(ap, rawdata_t *);

			/* Copy the size of the data */
			len = sizeof(u_int32_t);
			memcpy(&raw->size, buf, len);
			buf += len;
			curlen += len;

			/* Copy the actual data */
			len = raw->size;
			str = malloc(len);
			memcpy(str, buf, len);
			raw->data = str;
			buf += len;
			curlen += len;
			break;
		}

		if (*curfmt)
			curfmt++;
	}

	return curlen;
}

/*
 * get_sparam: A utility function. This is the same as get_param
 * but this version first skips the parameters in fmt_skip
 * before processing fmt.
 * Returns the size of buffer read.
 */
int get_sparam(char *buf, char *fmt_skip, char *fmt, va_list ap)
{
	int *num;
	char *ch;
	char *str;
	task_id *tid;
	rawdata_t *raw;
	int len;
	char *curfmt;
	int curlen = 0;

	curfmt = fmt_skip;

	while (*curfmt == '%') {
		switch (*(++curfmt)) {
		case 's':
			va_arg(ap, char *);
			break;
		case 'c':
			va_arg(ap, char);
			break;
		case 'd':
			va_arg(ap, u_int32_t);
			break;
		case 't':
			va_arg(ap, task_id);
			break;
		case 'r':
			va_arg(ap, rawdata_t);
			break;
		}

		if (*curfmt)
			curfmt++;
	}

	curfmt = fmt;

	while (*curfmt == '%') {
		switch (*(++curfmt)) {
		case 's':
			str = va_arg(ap, char *);
			len = strlen(buf)+1;
			memcpy(str, buf, len);
			buf += len;
			curlen += len;
			break;
		case 'c':
			ch = va_arg(ap, char *);
			*ch = *buf++;
			curlen++;
			break;
		case 'd':
			num = va_arg(ap, u_int32_t *);
			len = sizeof(u_int32_t);
			memcpy(num, buf, len);
			*num = ntohl(*num);
			buf += len;
			curlen += len;
			break;
		case 't':
			tid = va_arg(ap, task_id *);
			len = sizeof(task_id);
			memcpy(tid, buf, len);
			buf += len;
			curlen += len;
			break;
		case 'r':
			raw = va_arg(ap, rawdata_t *);

			/* Copy the size of the data */
			len = sizeof(u_int32_t);
			memcpy(&raw->size, buf, len);
			buf += len;
			curlen += len;

			/* Copy the actual data */
			len = raw->size;
			str = malloc(len);
			memcpy(str, buf, len);
			raw->data = str;
			buf += len;
			curlen += len;
			break;
		}

		if (*curfmt)
			curfmt++;
	}

	return curlen;
}


/*
 * put_argv: A utility function which puts all the argv
 * command line arguments to the buffer. Each argument is
 * seperated by a NULL byte. Returns the size of the
 * buffer used for storing the arguments or -1 if the
 * buffer is not big enough.
 */
int put_argv(char *buf, size_t n, int argc, char *argv[])
{
	int i;
	int curlen;
	int len;

	*buf = '\0';
	curlen = 0;

	for (i=0;i<argc;i++) {
		len = strlen(argv[i])+1;

		if (len + curlen > n)
			return FALSE;

		memcpy(buf+curlen, argv[i], len);
		curlen += len;
	}

	return curlen;
}

/*
 * get_argv: A utility function which is the counterpart of
 * the put_argv function. It restores the NULL seperated
 * arguments from the buffer into a string vector and
 * returns the malloc'ed vector (char *argv[]). The blank
 * argument specifies the number of initial arguments
 * which are left blank for the caller to fill in.
 * Returns NULL on failure.
 */
char **get_argv(char *buf, int argc, int blank)
{
	int i;
	char **argv;

	argv = (char**) malloc(sizeof(char*)*(blank+argc+1));

	for (i=0;i<blank;i++)
		argv[i] = NULL;

	for (;i<blank+argc;i++) {
		argv[i] = strdup(buf);
		buf += strlen(buf) + 1;
	}

	argv[i] = NULL;
	return argv;
}

/*
 * free_argv: Another utility function to free the structure
 * allocated by get_argv. Always returns NULL.
 */
char *free_argv(int argc, char **argv)
{
	int i;
	
	if (argv) {
		for (i=0;i<argc;i++)
			if (argv[i]) free((void*)argv[i]);
	}

	free((void*)argv);
	return NULL;
}
