/*
 * ncgpp.c: Netcluster Preprocessor.
 * This preprocessor is responsible for carrying out
 * the following functionality:
 * 1.	For all the RPC_FUNC(...):
 *		- Defines the function IDs (Starting from 0),
 *		- Generate the table for the parm, indexed
 *		  by function ID
 * 2.	For all the subtasks
 *		- Generate the subtask table
 *
 * (C) Copyright 2003, netcluster
 *
 * Written by Syed Ghaus Ali
 *
 * See the COPYING file for more info
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <malloc.h>
#include <ctype.h>


#define fbsize 1024
#define maxtokensize 256
#define maxtypes 6

/* Type and type map must be in the same index order*/
char *type[maxtypes] = {"void","int","task_id","rawdata_t","char","cptr"};
char *typemap[maxtypes] = {"%z","%d","%t","%r","%c","%s"};
char *seperators =" \n\t(),-*";
char *ignore =" \n\t-,(";


/* Table format for subtask table defintion */
char *stasktableln1 = "\n/* Preprocessor generated user task definitions not in any particular order */\nnc_taskdef_t nc_tasks[] = {\n";
char *stasktableln2 = "\t{ NULL, NULL }\n};\n";
char *stasktablecol1 ="\t{ \"";
char *stasktablecol2 = "\", ";
char *stasktablecol3 = " },\n";


/* Table format for rpc function ids, param, return types */
char *rpctableln1 = "\n/* Preprocessor generated user function definitions indexed by function ID */\nnc_rpcdef_t nc_rpcs[] = {\n";
char *rpctableln2 = "\t{ NULL, NULL, NULL }\n};\n";
char *rpctablecol1 = "\t{ \"";
char *rpctablecol2 = "\", \"";
char *rpctablecol3 = "\", \"";
char *rpctablecol4 = "\" },\n";


/* Comments before rpc function id definitions */
char *rpc_def_comments = "/* Preprocessor generated RPC Function IDs (Zero based indexes) */\n";


int ncpp_parse(char *);
int getnexttoken(int,char *);
int is_type(char *);
int is_seperator(char);
int is_ignore(char);
void parse_rpc(int);
void parse_subtask(int);
void define_rpc_id(int);
void write_rpc_table(int);
void write_stask_table(int);
void ncpp_cleanup();
void ncpp_usage(char *);
void ncpp_writeoutput(char *);
void out(char *,int);


typedef struct parmlist_t{
	char type[3];				/* Stores the type eg. %d */
	struct parmlist_t *next;
}parmlist;


typedef struct funcdeflist_t{
	char *funcid;
	parmlist *parm;				/* The parameter list */
	parmlist *ret;				/* The return list */
	struct funcdeflist_t *next;
}funcdeflist;



/*
 * parmlist_free: Frees the given list recursively.
 */
void parmlist_free(parmlist *list){
	if(list){
		parmlist_free(list->next);
		free(list);
		list = NULL;
	}
}


/*
 * funcdeflist_free: Frees the given list recursively.
 */
void funcdeflist_free(funcdeflist * list)
{
	if(list){
		funcdeflist_free(list->next);
		free(list->funcid);
		parmlist_free(list->ret);
		parmlist_free(list->parm);
		free(list);
		list = NULL;
	}
}


/* rpc_list stores the RPC_FUNC, their parameters and returns */
funcdeflist *rpc_list = NULL;
/* stask_list stores the function Id of the subtasks*/
funcdeflist *stask_list = NULL;


/*
 * ncpp_parse: Read the file token by token and then
 * if the token match that of RPC or a SUBTASK
 * parse them accordingly calling parse_rpc &
 * parse_subtask.
 */
int ncpp_parse(char *filename)
{
	int handle;
	char token[maxtokensize];

	handle = open(filename, O_RDONLY );
	if(handle <= 0){
		fprintf(stderr, "File not found.\n");
		return 0;
	}

	while(getnexttoken(handle,token))
	{
		if(strcmp(token,"RPC_FUNC") == 0)
			parse_rpc(handle);
		else if(strcmp(token,"SUBTASK") == 0)
			parse_subtask(handle);
	}
	close(handle);
	return 1;
}


/*
 * getnexttoken: Get the handle to the file as parm1 and
 * the buffer as parm2 to save the token.
 * Reads token separated by the seperators and
 * returns 1 if buffer has a token
 * returns 0 otherwise
 */
int getnexttoken(int handle,char *buffer)
{
	int index;
	char charbuffer;
	static char last_token = 0;
	int flag = 1;

	index = 0;
	if(last_token != 0){
		buffer[0] = last_token;
		buffer[1] = 0;
		last_token = 0;
		return 1;
	}

	/*
	if(eof(handle))
		return 0;
	*/

	while ((index<(maxtokensize-1))) {
		if (read(handle, &charbuffer, sizeof(char)) == 0) {
			if (flag)
				return 0;
			else
				return 1;
		}

		flag = 0;

		if(is_seperator( charbuffer )){
			if( !is_ignore( charbuffer ) ){
				if(index != 0){
					buffer[index] = 0;
					last_token = charbuffer;
					return 1;
				}
				else{
					buffer[0] = charbuffer;
					buffer[1] = 0;
					return 1;
				}
			}
			else{
				if(index != 0){
					buffer[index] = 0;
					return 1;
				}
			}
		}
		else{
			buffer[index] = charbuffer;
			index++;
		}
	}
	buffer[index] = 0;
	return 1;
}


/*
 * is_seperator:  checks if the given char is a
 * seperator char
 */
int is_seperator(char c)
{
	int i;
	for(i=0;i<strlen(seperators);i++){
		if(c == seperators[i] ){
			return 1;
		}
	}
	return 0;
}


/*
 * is_ignore:  checks if the given char is a
 * ignore char
 */
int is_ignore(char c)
{
	int i;
	for(i=0;i<strlen(ignore);i++){
		if(c == ignore[i] ){
			return 1;
		}
	}
	return 0;
}


/*
 * is_type: The parm is checked if its a type in nc
 * If its a type then val 1-5 is retured otherwise 0
 * is returned
 */
int is_type(char * id)
{
	int i;
	for(i=0;i<maxtypes;i++){
		if(strcmp(type[i], id)==0)
			return (i+1);
	}
	return 0;
}


/*
 * parse_rpc: This function is called when the RPC token
 * is matched. All the return types, function ID and the
 * parameters ofthis function are save in rpc_list.
 */
void parse_rpc(int handle)
{
	int func_id_found = 0;
	char token[maxtokensize];
	static funcdeflist *rpc = NULL;
	parmlist *currentparam = NULL;

	if(rpc_list == NULL)
		rpc_list = rpc = (funcdeflist *)malloc( sizeof(funcdeflist));
	else{
		rpc->next = (funcdeflist *)malloc( sizeof(funcdeflist));
		rpc = rpc->next;
	}
	rpc->ret = NULL;
	rpc->parm = NULL;

	while( getnexttoken(handle,token) )
	{
		if(strcmp(token,")") == 0)
			break;
		if(strcmp(token,"*") == 0){
			strcpy(currentparam->type,typemap[is_type("cptr")-1] );
			continue;
		}

		if(is_type(token)>0 && is_type(token)<(maxtypes+1)){
			if(currentparam == NULL)
				if( !func_id_found )
					rpc->ret = currentparam = (parmlist *)malloc(sizeof(parmlist));
				else
					rpc->parm = currentparam = (parmlist *)malloc(sizeof(parmlist));
			else{
				currentparam->next = (parmlist *)malloc(sizeof(parmlist));
				currentparam = currentparam->next;
			}
			strcpy(currentparam->type,typemap[is_type(token)-1] );
			currentparam->next = NULL;
		}
		else{
			if( !func_id_found ){
				rpc->funcid = (char*)malloc(strlen(token)+1);
				strcpy(rpc->funcid,token);
				if(rpc->ret == NULL){
					rpc->ret = (parmlist *)malloc(sizeof(parmlist));
					strcpy( rpc->ret->type, typemap[is_type("void")-1] );
				}
				currentparam = NULL;
				func_id_found = 1;
			}
		}
	}
	rpc->next = NULL;
}


/*
 * parse_subtask: This function is called when the SUBTASK token
 * is matched. Only function ID is saved in stask_list.
 */
void parse_subtask(int handle)
{
	char token[maxtokensize];
	static funcdeflist *stask = NULL;

	if(stask_list == NULL)
		stask_list = stask = (funcdeflist *)malloc( sizeof(funcdeflist));
	else{
		stask->next = (funcdeflist *)malloc( sizeof(funcdeflist));
		stask = stask->next;
	}
	getnexttoken(handle,token);
	stask->funcid = (char*)malloc(strlen(token)+1);
	strcpy(stask->funcid,token);
	stask->parm = NULL;
	stask->ret = NULL;
	stask->next = NULL;
}


/*
 * write_stask_table: Read from the stask_list and writes
 * it in the file in the format given by stasktablecol &
 * stasktableln.
 */
void write_stask_table(int handle)
{
	funcdeflist *slist = stask_list;

	out(stasktableln1,handle);
	while(slist){
		out(stasktablecol1,handle);
		out(slist->funcid,handle);
		out(stasktablecol2,handle);
		out(slist->funcid,handle);
		out(stasktablecol3,handle);
		slist = slist->next;
	}
	out(stasktableln2,handle);
}


/*
 * write_rpc_table: Read from the rpc_list and writes
 * it in the file in the format given by rpctablecol &
 * rpctableln.
 */
void write_rpc_table(int handle)
{
	funcdeflist *flist = rpc_list;
	parmlist *plist;

	out(rpctableln1,handle);
	while(flist){
		out(rpctablecol1,handle);
		out(flist->funcid,handle);
		out(rpctablecol2,handle);
		plist = flist->parm;
		while(plist){
			out(plist->type,handle);
			plist = plist->next;
		}
		out(rpctablecol3,handle);
		plist = flist->ret;
		while(plist){
			out(plist->type,handle);
			plist = plist->next;
		}
		flist = flist->next;
		out(rpctablecol4,handle);
	}
	out(rpctableln2,handle);
}


/*
 * define_rpc_id: Read from the rpc_list and defines
 * the function IDs in the file starting from 0
 *
 */
void define_rpc_id(int handle)
{
	int count = 0;
	char cntstr[10];
	funcdeflist *flist = rpc_list;

	out(rpc_def_comments,handle);
	while(flist){
		out("#define ",handle);
		out(flist->funcid,handle);
		out(" ",handle);
		sprintf(cntstr, "%d", count);
		out(cntstr,handle);
		out("\n",handle);
		flist = flist->next;
		count++;
	}
	out("\n",handle);
}


/*
 * ncpp_cleanup: For the list generated in ncPP_parse
 * this function frees those lists
 */
void ncpp_cleanup()
{
	funcdeflist_free(rpc_list);
	funcdeflist_free(stask_list);
}


/*
 * ncpp_usage: Provides the proper usage for this file
 * Gives parameter information incase of wrong arg to program
 */
void ncpp_usage(char *av)
{
	printf("Netcluster Preprocessor\n");
	printf("Syntax: %s <filename.n*>\n",av);
	printf("Input file name <src.n*>, output file name <src.*>\n",av);
	exit(1);
}


/*
 * ncpp_writeoutput: Generates the file. It writes the
 * RPC_FUNC definitions at the top of the file. Then all
 * the file is writtten and then the table for RPC funs and
 * subtasks are generated.
 */
void ncpp_writeoutput(char * srcname)
{
	int i;
	int j;
	int first_dot=1;
	int ihandle;
	int ohandle;
	int bytesread;
	char buffer[fbsize];
	char *destname;

	ihandle = open(srcname, O_RDONLY );
	if(ihandle <= 0){
		fprintf(stderr, "File not found.\n");
		return;
	}

	destname = (char *)malloc(strlen(srcname));
	j=0;
	for(i=0;i<=strlen(srcname);i++,j++){
		destname[j] = srcname[i];
		if(srcname[i] == '.'  && first_dot==1 ){
			first_dot = 0;
			i++;
		}
	}

	remove(destname);
	ohandle = open(destname, O_WRONLY|O_CREAT|O_TRUNC,
			S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if(ohandle <= 0){
		fprintf(stderr, "Cannot create output file.\n");
		return;
	}
	/* Write RPC definitions */
	define_rpc_id(ohandle);
	/* Copy the entire source file */
	while(1){
		bytesread = read(ihandle,buffer,fbsize);
		if (bytesread == 0)
			break;
		write(ohandle,buffer,bytesread );
	}
	close(ihandle);
	/* Write RPC table */
	write_rpc_table(ohandle);
	/* Write SUBTASK table */
	write_stask_table(ohandle);
	close(ohandle);
	free(destname);
}


/*
 * out: The buffer is written to file identified by the
 * handle.
 */
void out(char *buffer, int handle)
{
	write(handle, buffer,strlen(buffer));
}

/*
 * main: Entry point for parser. Checks the no of args 
 * if dont match, usage message is given. Calls the
 * parser, writes the new file and then cleans up for
 * the lists made during parsing
 */
int main(int argc, char *argv[])
{
	if(argc != 2)
		ncpp_usage("ncgpp");
	if(!ncpp_parse(argv[1]))
		return 0;
	ncpp_writeoutput(argv[1]);
	ncpp_cleanup();
	return 0;
}
