/*
 * vasmain.c: The main vas file.
 * (C) Copyright 2003, netcluster
 *
 * Written by Syed Ghaus Ali
 * Added config file support (Fakhir Ansari)
 * Added a timed pinger (Fakhir Ansari)
 *
 * See the COPYING file for more info
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#include "defs.h"
#include "clist.h"
#include "chash.h"
#include "sock.h"
#include "io.h"
#include "msgs.h"
#include "transact.h"
#include "connman.h"
#include "util.h"
#include "conf.h"
#include "log.h"
#include "vas.h"


chash vas_table;

/* Places to search for the VAS config files */
char *vconf_paths[] = {
	"ncvas.conf",
	"../ncvas.conf",
	"/etc/ncvas.conf",
	"~/ncvas.conf"
};

/* Tag specification for daemon config. To rename a tag, modify
 * it in this struct and also in the ncdaemon_readconf function
 * where it is interpretted. */
taginfo_t vconf_tags[] = {
	{ "LISTEN_PORT", TAGCSET_INT, TAGTYPE_INT, 0, NULL },
	{ "SELF_ADDR", TAGCSET_ALL, TAGTYPE_NADDR, 0, NULL },
	{ "CONN_TIMEOUT", TAGCSET_INT, TAGTYPE_INT, 0, NULL },
	{ "PING_INTERVAL", TAGCSET_INT, TAGTYPE_INT, 0, NULL },
	{ "PING_TIMEOUT", TAGCSET_INT, TAGTYPE_INT, 0, NULL },
	{ "LOGGING_FILE", TAGCSET_ALL, TAGTYPE_QOTDSTR, 0, NULL },
	{ "LOG_EMERG", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_ALERT", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_CRIT", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_ERR", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_WARNING", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_NOTICE", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_INFO", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ "LOG_DEBUG", TAGCSET_NULL, TAGTYPE_NULL, 0, NULL },
	{ NULL, NULL, 0, 0, NULL }
};

/* VAS config struct */
vconf_t vconf;

/* ncvas_readconf: Reads the config files if found and inits
 * the vconf data to sane values. Returns FALSE on error. */
int ncvas_readconf()
{
	int i;
	clist *itm;
	int *iptr;
	naddr_t *nptr;

	/* Try all config files one by one */
	for (i=0;i<sizeof(vconf_paths)/sizeof(char*);i++) {
		if (!access(vconf_paths[i], R_OK)) {
			if (!conf_read(vconf_paths[i], vconf_tags))
				return FALSE;
			break;
		}
	}

	if (i==sizeof(vconf_paths)/sizeof(char*))
		do_log(LOG_NOTICE,"Initialization: no config file found");

	vconf.log_types = 0;

	/* Store the values read from the config file into vconf */
	for (i=0;vconf_tags[i].tag;i++) {
		itm = vconf_tags[i].values;

		if (!strcmp(vconf_tags[i].tag, "LISTEN_PORT")) {
			if (itm && itm->data)
				vconf.listen_port = *((int*)itm->data);
			else
				vconf.listen_port = DPORT;
		}
		if (!strcmp(vconf_tags[i].tag, "SELF_ADDR")) {
			if (itm && itm->data) {
				vconf.self_id = *((naddr_t*)itm->data);
			}
			else if (!make_self_daemon_id(&(vconf.self_id),
							vconf.listen_port)) {
				do_log(LOG_ERR, "Unable to get local hostname. Specify"
						"SELF_ADDR in the config file.");
				return FALSE;
			}
		}
		else if (!strcmp(vconf_tags[i].tag, "CONN_TIMEOUT")) {
			if (itm && itm->data)
				vconf.conn_ttl = *((int*)itm->data);
			else
				vconf.conn_ttl = DEFAULT_CONN_TTL;
		}
		else if (!strcmp(vconf_tags[i].tag, "PING_INTERVAL")) {
			if (itm && itm->data)
				vconf.ping_interval = *((int*)itm->data);
			else
				vconf.ping_interval = VAS_PING_INTERVAL;
		}
		else if (!strcmp(vconf_tags[i].tag, "PING_TIMEOUT")) {
			if (itm && itm->data)
				vconf.ping_timeout = *((int*)itm->data);
			else
				vconf.ping_timeout = VAS_PING_TIMEOUT_INTERVAL;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOGGING_FILE")) {
			if (itm && itm->data)
				vconf.log_file = strdup((char*)itm->data);
			else
				vconf.log_file = NULL;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOG_EMERG")) {
			if (itm && itm->data)
				vconf.log_types |= LOG_EMERG;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOG_ALERT")) {
			if (itm && itm->data)
				vconf.log_types |= LOG_ALERT;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOG_CRIT")) {
			if (itm && itm->data)
				vconf.log_types |= LOG_CRIT;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOG_ERR")) {
			if (itm && itm->data)
				vconf.log_types |= LOG_ERR;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOG_WARNING")) {
			if (itm && itm->data)
				vconf.log_types |= LOG_WARNING;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOG_NOTICE")) {
			if (itm && itm->data)
				vconf.log_types |= LOG_NOTICE;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOG_INFO")) {
			if (itm && itm->data)
				vconf.log_types |= LOG_INFO;
		}
		else if (!strcmp(vconf_tags[i].tag, "LOG_DEBUG")) {
			if (itm && itm->data)
				vconf.log_types |= LOG_DEBUG;
		}

		/* Free the data read from the config file. Set 'values'
		 * to NULL if it isnt to be free'd. */
		if (vconf_tags[i].values) {
			vconf_tags[i].values = clist_custfree(
					vconf_tags[i].values, free);
		}
	}

	/* Listen port takes precedence over specified port */
	vconf.self_id.port = htons(vconf.listen_port);

#ifdef DEBUG
	fprintf(stderr, "Setting SELF_ADDR to: %s\n", naddr_to_str(&vconf.self_id));
#endif

	vconf.homedir = get_homedir();

	return TRUE;
}

/* 
 * ncvas_calc_hash: This function calculates the hash value for
 * the vas_list hash table. The value is calculated as a sum of
 * the ip and port numbers. The load is ignored.
 */ 
int ncvas_calc_hash(void *a, int size)
{
	vasclient_t *x = (vasclient_t*)a;
	int value;

	value = x->dem_id.ip + x->dem_id.port;
	if (value < 0)
		value = -(value);

	return (value % size);
}

/* 
 * ncvas_item_cmp: The two parameters are compared on the basis
 * of vasclient_t.dem_id only and returns a value:
 *    0 if equal
 *    1 if a > b
 *    -1 otherwise 
 * Is used by hash table find functions.
 */
int ncvas_item_cmp(void *a, void *b)
{
	vasclient_t *x = (vasclient_t*)a;
	vasclient_t *y = (vasclient_t*)b;

	return memcmp(&x->dem_id, &y->dem_id, sizeof(daemon_id));
}

/* 
 * ncvas_itemload_cmp: The two parameters are compared on the
 * basis of vasclient_t.load and returns a value:
 *    0 if equal
 *    1 if a > b
 *    -1 otherwise 
 * Is used by hash table insert_sorted functions.
 */
int ncvas_itemload_cmp(void *a, void *b)
{
	vasclient_t *x = (vasclient_t*)a;
	vasclient_t *y = (vasclient_t*)b;

	if (x->load.numprocs > y->load.numprocs)
		return 1;
	else if (x->load.numprocs < y->load.numprocs)
		return -1;
	else
		return 0;
}

/*
 * ncvas_item_display: Display the daemon's information for
 * debugging purposes only.
 */
void ncvas_item_display(void *vclient, void *dummy)
{
	vasclient_t *item = vclient;
	do_log(LOG_INFO, "Display Item: ip<%d> port<%d> load<%d>", ntohl(item->dem_id.ip),
			ntohs(item->dem_id.port),
			item->load.numprocs);
}

/*
 * hndfail_pinger: This transaction failure handler is called if
 * no ping reply is receieved from a daemon. It removes the daemon
 * from the vas_table. The 'data' parameter is the vas_item for
 * the daemon.
 */
TRANSFAIL_HANDLER(hndfail_pinger)
{
	vasclient_t *vitem = (vasclient_t*)data;

	chash_remove_custom(&vas_table, vitem, ncvas_item_cmp);
	do_log(LOG_INFO, "Daemon PING timeout: Unregistered");
}

/*
 * ncvas_item_ping: Pings the item for which it is called.
 */
void ncvas_item_ping(void *vclient, void *dummy)
{
	msghdr_t hdr;
	vasclient_t *cp_vitem;
	vasclient_t *vitem = vclient;

	hdr.sender = vconf.self_id;
	hdr.receiver = vitem->dem_id;
	hdr.msg_type = MSG_VAS_PING;
	hdr.priority = 1;
	hdr.data_size = 0;

	/* Allocate a vitem copy for the transaction */
	cp_vitem = malloc(sizeof(vasclient_t));
	*cp_vitem = *vitem;

	if (!trans_write(&hdr, NULL, hndfail_pinger,
			cp_vitem, vconf.ping_timeout))
	{
		/* Daemon is down already. Unable to
		 * deliver the ping so remove it. */
		chash_remove_custom(&vas_table, vitem, ncvas_item_cmp);
		do_log(LOG_INFO, "Daemon PING timeout: Unregistered (unable to send PING)");
	}
	else
		do_log(LOG_INFO, "PING sent");
}

/*
 * ncvas_pinger: This ctimer function is called after every
 * VAS_PING_TIMEOUT_INTERVAL secs and it pings all the connected
 * daemons in vas_table.
 */
CTIMER_FUNC(ncvas_pinger)
{
	chash_foreach(&vas_table, ncvas_item_ping, NULL);
}

/* terminate_handler: Termination handler for SIGINT, SIGHUP, SIGTERM */
void terminate_handler(int signum)
{
	/* Close all connections */
	fd_close_all();

	/* Close the log file */
	if (log_file)
		fclose(log_file);

	exit(signum);
}

/* ncvas_init: Performs all the initialization of the VAS */
int ncvas_init()
{
	FILE *lf;
	struct sigaction new_action;

	/* Set up the signal handler */
	new_action.sa_handler = terminate_handler;
	sigemptyset(&new_action.sa_mask);
	new_action.sa_flags = 0;
	sigaction(SIGINT, &new_action, NULL);
	sigaction(SIGHUP, &new_action, NULL);
	sigaction(SIGTERM, &new_action, NULL);

	/* Initialize the open fd to zero */
	FD_ZERO(&fd_ractive);
	FD_ZERO(&fd_wactive);

	/* Read and parse the config file */
	if (!ncvas_readconf()) {
		fprintf(stderr, "Initialization: failed\n");
		return FALSE;
	}

	set_log_types(vconf.log_types);
	set_module_name("ncvas");
	conn_set_ttl(vconf.conn_ttl);

#ifndef DEBUG
	if (vconf.log_file) {
		set_log_file(vconf.log_file);
	}
#endif

	/* Register trans_run to be called every second */
	ctimer_add(1, 0, trans_run, NULL);
	ctimer_add(vconf.ping_interval, 0, ncvas_pinger, NULL);
	
	chash_init(&vas_table, VAS_HASHTABLE_SIZE, ncvas_calc_hash, free);

	do_log(LOG_INFO,"Initialization: Success");

	return TRUE;
}

/* 
 * ncvas_deinit: Performs the deinitialization of the vas before exit.
 */
int ncvas_deinit()
{
	/* Close all connections */
	conn_close_all();

	/* Delete all timers */
	ctimer_remove_all();

	return TRUE;
}

/*
 * ncvas_idle: This function is continuously called in the idle loop.
 * It performs background tasks such as timing out connections and
 * running timers.
 */
void ncvas_idle()
{
        static time_t lasttime;
	
	if (lasttime != time(NULL)) {
		lasttime = time(NULL);

		/* Timeout the expired connections */
		conn_dec_ttl();

		/* Call the timer handler */
		ctimer_run();
	}
}


/*
 * ncvas_main: The main control of the vas. Its tasks include listening
 * for incoming messages and and handling them accordingly by calling
 * msg_dispatch function
 */
int ncvas_main()
{
	int i;
	struct timeval tv;
	fd_set fd_read;
	struct sockaddr_in clientaddr;
	int sock;
	int nmsgs;
	int size;
	int new;
	
	if ((sock = make_listen_socket(vconf.listen_port)) < 0) {
		yell("Unable to open port<%d> for listening. Aborting...",
				vconf.listen_port);
		return FALSE;
	}
	
	/* Dont register as connection coz its only for listening */
	fd_open(sock, FDTYPE_LISTEN);

	while (1) {
		ncvas_idle();

		fd_read = fd_ractive;
		tv.tv_sec = 0;
		tv.tv_usec = 500;

		if (!select(global_max_fd+1, &fd_read, NULL, NULL, &tv))
			continue;

		for (i=0;i<=global_max_fd;i++) {
			if (!FD_ISSET(i, &fd_read))
				continue;

			switch (FDTYPE(i)) {
				case FDTYPE_LISTEN:
					/* new connection */
					if (i != sock)
						continue;

					size = sizeof(clientaddr);
					new = accept(sock, &clientaddr, &size);
					if (new < 0) {
						yell("Unable to accept:%s", strerror(errno));
						continue;
					}

					conn_fd_open(new);

					do_log(LOG_NOTICE,"Connect from addr<%s:%d> sock<%d>",
							inet_ntoa(clientaddr.sin_addr),
							ntohs(clientaddr.sin_port), new);
					break;

				case FDTYPE_MSG:
					/* data on one of connected descriptors */

					/* if nmsgs < 0 then either an error
					 * has occured or the connection has
					 * been closed. if zero then a partial
					 * msg has been read and is buffered
					 * otherwise number of msg read is
					 * returned */
					nmsgs = msg_read(i);
					conn_reset_ttl(i);

					if (nmsgs < 0) {
						conn_fd_close(i);
						do_log(LOG_NOTICE,"Session closed: sock<%d>",i);

					}
					else if (nmsgs > 0)
						msg_dispatch();
					break;

				default:
					do_log(LOG_ERR, "Unknown file des: fd<%d> type<%d>",
							i, FDTYPE(i));
			};
		}
	}
	
	return TRUE;
}

/* 
 * main: VAS main program 
 */
int main(int argc, char *argv)
{
	pid_t child_pid;

	if (!ncvas_init(argc, argv)) {
		fprintf(stderr, "ncvas:: Initialization failed. Aborting...");
		exit(EXIT_FAILURE);
	}

#ifndef DEBUG
	child_pid = fork();

	if (child_pid == -1) {
		fprintf(stderr, "ncvas: Cannot startup. Error forking child process!\n");
		exit(EXIT_FAILURE);
	}
	else if (child_pid != 0) {
		fprintf(stderr, "netcluster VAS: Started up successfully\n");
		fprintf(stderr, "pid=%d\n", child_pid);
		exit(EXIT_SUCCESS);
	}

	close(0);
	close(1);
	setsid();
#endif /* DEBUG */

	ncvas_main();
	ncvas_deinit();

	return 0;
}
